#!/usr/bin/env python3
"""
geoms_qa.py

"""

import os
import sys
import argparse
import copy
import datetime
from pathlib import Path
import geoms_vntools
import geoms_tctools
import geoms_tools


def geoms_qa(tav_file, template_file, orig_file):
    """
    Wrapper for AVDC/EVDC QA tool



     error codes
     0 QA pass, 	no TC
     1 QA limit, 	no TC
     2 QA fail, 	no TC
     3 wrong parameters
     4 QA pass, 	TC pass
     5 QA limit, 	TC pass
     6 QA fail, 	TC pass
     7 QA pass,	TC fail
     8 QA limit,	TC fail
     9 QA fail, 	TC fail

    """


    qa_error_code = 3             # default QA error code
    tc_error_code = 1             # default TC error code
    tc_error_code_2 = 1           # default VN error code


    if tav_file is None:
        qa_error_code = 2   # can still do template checks if HDF and template files are present
        print('WARNING: no TAV file given.')
    else:
        if os.path.isfile(tav_file):
            qa_error_code = 0
        else:
            qa_error_code = 2   # can still do template checks if HDF and template files are present
            print('ERROR: TAV file not found.')

    if orig_file is None:
        qa_error_code = 3
        print('ERROR: no GEOMS file given.')
    else:
        if os.path.isfile(orig_file):
            if qa_error_code != 2:
                qa_error_code = 0
        else:
            qa_error_code = 3
            print('ERROR: GEOMS file not found.')

            if __name__ == "__main__":
                # Return QA error code to system if run as main program
                sys.exit(qa_error_code)
            else:
                # Return value to calling program
                return qa_error_code

    if template_file is None:
        # print('no GEOMS-TE file given.')
        have_template_file=False
    else:
        if os.path.isfile(template_file):
            have_template_file=True
            if qa_error_code != 3:
                tc_error_code = 0 # i.e. HDF file exists
        else:
            have_template_file=False
            print('ERROR: GEOMS-TE file not found.')

    if qa_error_code != 3:
        tc_error_code_2 = 0  # i.e. HDF file exists

    if qa_error_code == 0:
        print(tav_file)
        print(orig_file)

        # set log file output dir to working dir
        qa_error_code = geoms_tools.geoms_tools(0, tav_file, orig_file, outdir=os.path.abspath('.') )

    elif tc_error_code == 0 or tc_error_code_2 == 0:
        # Can't do QA but can do Template and/or Version Name checks so create logfile

        log_file = Path(os.path.abspath(orig_file)).with_suffix('.log')
        with open(log_file, 'a') as f:
            f.write('GEOMS Quality Assurance Tool\n')
            f.write('Maintained by the Aura Validation Data Center, NASA/GSFC\n')
            f.write('http://avdc.gsfc.nasa.gov/GEOMS\n\n')
            f.write('Log started on {}\n'.format(datetime.datetime.utcnow().strftime("%a %b %d %H:%M:%S %Y")))
            f.write('Result of performing Quality Assurance on ' + os.path.abspath(orig_file) + '\n')
            f.write('(0=PASS, 1=GEOMS compliant with non-standard components, 2=FAIL)\n')
            f.write('ERROR: TAV file not found\n')
            if template_file is not None:
                f.write('INFORMATION: Required Metadata Template: ' + os.path.basename(template_file))
            f.write('\n\n')
            f.write('2 (Failed QA)')


    if qa_error_code != -2:
        # Determine working directory for the version names list file
        # First try the template directory, but otherwise use the TAV file directory
        if template_file is not None:
            chDirGEOMSvn = os.path.dirname(os.path.abspath(template_file))
        elif tav_file is not None:
            chDirGEOMSvn = os.path.dirname(os.path.abspath(tav_file))
        else:
            chDirGEOMSvn = ''     # can't find directory holding the version names list file

        if tc_error_code == 0:
            tc_error_code = geoms_tctools.geoms_tctools(template_file, orig_file, qa_error_code)
            tc_error_code_hold = tc_error_code*3 + 4
        elif have_template_file and qa_error_code != 3:
            log_file = Path(os.path.abspath(orig_file)).with_suffix('.log')
            with open(log_file, 'a') as f:
                f.write('\nERROR: GEOMS Metadata Template file not found')
            qa_error_code = 3
            tc_error_code = 0
            tc_error_code_hold = 0
        else:
            tc_error_code = 0
            tc_error_code_hold = 0

        if tc_error_code_2 == 0:
            tc_error_code_2 = geoms_vntools.geoms_vntools(chDirGEOMSvn, orig_file, qa_error_code, tc_error_code)
            # Note tc_error_code_2 will return the highest of tc_error_code or the Version Name error check value
            if tc_error_code_2 == -1 or qa_error_code == 3:
                # File input error as data_version_name list file can't be found and there is a version name in the GEOMS file
                tc_error_code = 0
                qa_error_code = 3
            else:
                tc_error_code = tc_error_code_2*3 + 4
        else:
            tc_error_code = tc_error_code_hold

    # Total error
    qatc_error_code = abs(qa_error_code) + tc_error_code

    if __name__ == "__main__":
        # Exit with system error code if run as main program
        sys.exit(qatc_error_code)
    else:
        # return value to calling program
        return qatc_error_code



if __name__ == "__main__":

    qaver = 'version 2.11, 2022-08-09'

    parser = argparse.ArgumentParser(
        prog="geoms_qa.py",
        description="GEOMS quality assurance (geoms_qa)",

        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-t', '--tav_file',
                        dest='tav_file',
                        help="GEOMS TAV file",
                        default=None)

    parser.add_argument('-m', '--template_file',
                        dest='template_file',
                        help="GEOMS template file",
                        default=None)

    parser.add_argument('-d', '--orig_file',
                        dest='orig_file',
                        help="GEOMS HDF file",
                        default=None)


    args = parser.parse_args()
    template_file = args.template_file
    orig_file = args.orig_file
    tav_file =  args.tav_file

    # Check args

    # Need at least HDF and TAV file
    if None in (orig_file, tav_file):
        parser.print_help()
        txt_message = ("\nGEOMS quality assurance (geoms_qa)\n"
                       "Release {}\n"
                       "Does GEOMS quality assurance on HDF4/5 and netCDF files\n\n"
                       "1: TAVFILE is the path of the TAV file, you want to validate against\n"
                       "2: GEOMSTEFILE is the path of the GEOMS template file, you want to test\n"
                       "      (optional argument; template checks are only performed if GEOMS-TE is provided)\n"
                       "3: ORIGFILE is the path of the original HDF file, you want to test\n"
                       "------------------------------------------------------------------------------------------------").format(qaver)
        print(txt_message)

        qa_error_code = 3
        sys.exit(qa_error_code)

    geoms_qa(tav_file, template_file, orig_file)
