"""
geoms_tools.py

"""

import os
import sys
import re
from pathlib import Path
import datetime
import glob
import pycr8ascii
import pycr8hdf


def geoms_version_check(geoms_iteration, ga, newdfv, nodfvinc):
    """
    Procedure to determine a new value for the DATA_FILE_VERSION if a new HDF file is to be
    created when performing Harmonization

    """

    if geoms_iteration == 1:
        print("ERROR: not implemented yet. Exiting.")
        sys.exit(0)

def geoms_one_off_check(tav_file,ga,one_off_fix):
    """

    Procedure to check the TAV file for individual corrections required for harmonization. Fields start
    with the title 'GEOMS_HARMON'.

    """

def read_tav_dates(tav_dir, min_ver):
    """
    Create arrays of all available TAV files and their creation dates

    Inputs: tav_dir - Directory containing all the TAV files
            min_ver - Minimum version number for valid TAV files

    Output: tav_ver_d - dictionary of TAV file names: TAV file creation dates, TAV version values


    """

    path_spec=['tableattrvalue_??R???.dat','tableattrvalue_??R???_idl.dat']

    tav_mjd2k = []
    tav_all = []
    tav_files = []
    tav_ver = []
    tav_ver_d = {}
    for i in range(len(path_spec)):
        tav_all += glob.glob(os.path.join(tav_dir, path_spec[i]))

    tcnt = len(tav_all)
    if tcnt > 0:
        tav_vr_vals =[]
        for i, tav_file in enumerate(tav_all):
            tav_name = os.path.basename(tav_file)
            tav_chk = re.split('_|\.', tav_name)
            if len(tav_chk[1]) == 6:
                tav_vr = [tav_chk[1][0:2], tav_chk[1][3:]]
                if all([j.isdecimal() for j in tav_vr]) and int(tav_vr[0]) >= min_ver:
                    tav_ver.append(tav_chk[1])
                    tav_ver_d.update({tav_file : tav_chk[1] } )
                    #tav_files.append(tav_file)

        # Create list of all valid TAV files
        for f_key, val in tav_ver_d.items():
        #for f_key tav_files:
            with open(f_key, 'r') as f:
                for line in f:
                    dum = line.rstrip()
                    if 'BASED' in dum.upper():
                        break
                res = dum.split()
                if len(res) >= 2:
                    t_ymd, t_hms = res[-2:]
                    try:
                        t_iso = datetime.datetime.strptime(t_ymd + t_hms, "%Y-%m-%d%H:%M:%S")
                        tav_ver_d[f_key] = [val, t_iso]
                    except Exception:
                        # Invalid date format, remove from list
                        _ = tav_ver_d.pop(f_key, None)


    return tav_ver_d


def tav_files_for_qa(tav_file, ga_d, log_file):
    """
    Procedure to determine TAV file versions needed for definition checks when doing QA,
    based on FILE_GENERATION_DATE (FGD) and FILE_META_VERSION (FMV) values


    Inputs: tav_file - TAV file used in call to geoms_qa
            ga_d - Global attributes read in by pycr8ascii
            log_file - File for writing messages

    Output: tav_list - list containing the file names of the TAV files needed for testing.
           The first index is the TAV determined from FGD if determined, or the input TAV
           The second index is the TAV determined from FMV, if different to the first TAV
           gf_err - returns True if an error occurs when determining TAV file versions
           cur_tav = reutrns True if using the input TAV rather than the one determined by FGD

    """

    min_ver=4                     # minimum allowable TAV version

    gf_err=False
    cur_tav=False
    tav_list=['']*2
    gf_tav=''

    # Read in all the TAV filenames and definition creation dates
    tav_dir = os.path.dirname(os.path.abspath(tav_file))
    tav_all = read_tav_dates(tav_dir, min_ver)

    # Sorted by date:
    # sorted(tav_list.items(), key=lambda x: x[1][1])
    tav_sorted = sorted(tav_all.items(), key=lambda x: x[1][1])

    #  Identify input TAV file version and revision number
    tav_vr_vals = []
    tav_name = os.path.basename(tav_file)
    tav_chk = re.split('_|\.', tav_name)
    if len(tav_chk) >= 2:
        tav_ver=tav_chk[1].upper()
        if len(tav_chk[1]) == 6:
            tav_vr = [tav_chk[1][0:2], tav_chk[1][3:]]
            if all([j.isdecimal() for j in tav_vr]):
                tav_vr_vals.append(tav_vr)

    # Check FILE_GENERATION_DATE and FILE_META_VERSION to identify correct TAV
    # file to check metadata against
    gf_chk=['FILE_GENERATION_DATE', 'FILE_META_VERSION']
    # n_spl=[1,2]
    for i in (0,1):
        if i == 0:
            gf_val = ga_d[gf_chk[i]]
        else:
            gf_val = re.split(' |;', ga_d['FILE_META_VERSION'])[0]

        try:
            fgd_chk = datetime.datetime.strptime(gf_val, "%Y%m%dT%H%M%SZ")
        except ValueError:
            # FILE_GENERATION_DATE not in correct format
            fgd_chk = datetime.datetime(1,1,1)

        invalid_gf = True
        test2 = False

        if i == 0:
            # current MJD2K + 1 day
            systmjd2k = (datetime.datetime.utcnow() - datetime.datetime(1999,12,31)) # current MJD2K + 1 day
            if fgd_chk > datetime.datetime(1,1,1)  and fgd_chk - datetime.datetime.utcnow():
                invalid_gf=False
                t_tav = [j[1][1] - fgd_chk for j in tav_sorted]
                tav_idx = [j for j, x in enumerate(t_tav) if x.days < 0]
                if len(tav_idx) > 0:
                    t_idx = tav_idx[-1]
                    gf_tav = tav_sorted[t_idx][0]
                    #gf_tav = tav_sorted[tav_idx][0]
                    #gf_tav = tav_sorted[t_idx]
                    test2 = True

        else:
            invalid_gf=False
            if len(gf_val) == 6:
                test2=True       # test is redundant because tavver version and revision values already checked
                if gf_val.upper() != tav_ver.upper():
                    gf_tav = tav_sorted[[i[1][0] for i in tav_sorted].index(gf_val)][0]


        test1=gf_tav != ''         # i.e. tableattrvalue file identified is available

        if test1 and test2:
            tav_list[i] = gf_tav
        elif invalid_gf or not test2 or not test1:
            if i == 0:
                itxtfgd='. Metadata checks based on ' + tav_ver + ' definitions\n'
                tav_list[i]=tav_file
                cur_tav=True
            else:
                itxtfgd=''


            #with open(log_file, 'a') as f:
            f = open(log_file, 'a')
            if invalid_gf:
                if i == 0:

                    f.write('\nERROR: TAV file version containing metadata definitions cannot be identified due to missing or\n')
                    f.write('  invalid ' + gf_chk[i] + ' global attribute' + itxtfgd + '\n')
                else:
                    itxtfgd=''
                    f.write('ERROR: TAV file version cannot be identified due to missing or invalid' + gf_chk[i] +  'global attribute')

            elif not test1:
                if i == 0:
                    itxt='corresponding to'
                else:
                    itxt='matching'

                f.write('ERROR: TAV file '+itxt+' the ' + gf_chk[i] +
                        ' entry ('+gf_val+') not found'+itxtfgd + '\n')
            else:
                # Can only occur with FILE_META_VERSION value
                if min_ver < 10:
                    mvs='0' + str(min_ver)
                else:
                    mvs=str(min_ver)
                f.write('ERROR: '+ 'FILE_GENERATION_DATE'+' metadata definitions must be version '+mvs+
                            ' or greater\n')
            gf_err=True

            f.close()

    if tav_list[0] ==  tav_list[1]:
        tav_list[1] = ''

    return tav_list, gf_err, cur_tav


def geoms_tools(optqh, tav_file, hdf_qh, outdir=None, dataformat='H5', date_time=False, notc=False, nodfvinc=False):
    """
    Program to perform QA or Harmonization on GEOMS compliant HDF and netCDF files.

      Input: optqh - Integer defining type of testing 0=QA, 1=Harmonization
             tav_file - Table Attribute Values file used to test Metadata
             hdfqh - Either a) a single HDF file, or
                            b) a filespec, or
                            c) a string array containing a list of HDF files
             Note: Full path must be included as part of the filenames
             outdir - optional parameter which defines an output directory for the HDF files
                      (if optqh EQ 1) and the log files. When optqh EQ 1, if this parameter
                      is included a new HDF file will be created. If it is not included a new
                      HDF will not be created
             dataformat - optional parameter which defines the output format of the file
                          (currently accepts 'H4'/'H5'/'NC3'). Applies to harmonization option only,
                          it is ignored for QA, and if no outdir parameter is given. If not
                          included any new file is in the same format as the input file.
             date_time - if included, will relax checks on DATETIME in
                  idlcr8hdf.
             notc - if included, changes the Required DATA_TEMPLATE comment
                  for harmonized files so that Template Checking does not take place during
                  QA/TC checks
             nodfvinc - if included, does NOT increment the DATA_FILE_VERSION

      Output: A log file for each HDF file with the same filename as the HDF file except
              with the extension '.log', as well as an exit error code. Currently the exit
              error code will only apply to the last file tested.
              Exit Codes:
                0: Passed QA/Harmonization
                1: QA - GEOMS compliant with non-standard components;
                   Harmonization - Failed
                2: QA - Failed
                3: Incorrect input parameters


    """


    gtver='geoms_tools version 2.11, 2022-08-08'
    gt_error_code=3               # Integer to identify return error code (3 = Invalid input)
    # par1chk=0
    # parchk=0
    qhs1error=False

    # if date_time: # relaxes checks on DATETIME values in idlcr8hdf call
    #     o4='DTFVOK'
    # else:
    #     o4='0'

    # TODO: Check input parameters
    gt_error_code=0

    hdf_dir =  os.path.dirname(os.path.abspath(hdf_qh))

    if outdir is not None:
        if os.path.isdir(outdir):
            intype3=1
            qh_dir = outdir
    else:
        qh_dir = hdf_dir

    if optqh == 0:    # QA Option
        logname='pycr8qa.log'
        lgtxt='Quality Assurance'
        gtcode=['0 (Passed QA)', '1 (GEOMS compliant with non-standard components)', '2 (Failed QA)']
        gttxt='(0=PASS, 1=GEOMS compliant with non-standard components, 2=FAIL)'
        n_lc=1                  # number of times to call idlcr8hdf
    else:           # Harmonization Option
        logname='pycr8hdf.log'
        lgtxt='Harmonization'
        gtcode=['0 (Passed Harmonization)','1 (Failed Harmonization)','1 (Failed Harmonization)']
        gttxt='(0=SUCCESS, 1=FAIL)'
        # dontuse='Reading input Global'
        # Will call idlcr8hdf twice if wanting to make an HDF file, and file requires modification
        # (note - if file meets all GEOMS criteria, a new file will not be generated)
        if intype3 == -1:
            n_lc=1
        else:
            n_lc=2

    h_typeall=dataformat

    qhlog = Path(os.path.abspath(hdf_qh)).with_suffix('.log')

    newlname = qhlog          # log file name may change to match any new HDF file created

    # NB: Make sure this is where the log file is first written to (will overwrite old file)
    with open(qhlog, 'w') as f:
        f.write(gtver + '\n')
        f.write('GEOMS '+lgtxt+' Tool\n')
        f.write('Maintained by the Aura Validation Data Center, NASA/GSFC\n')
        f.write('http://avdc.gsfc.nasa.gov/GEOMS\n\n')

        f.write('Log started on ' + datetime.datetime.utcnow().strftime("%a %b %d %H:%M:%S %Y") + '\n')
        f.write('Result of performing '+lgtxt+' on '+ os.path.abspath(hdf_qh) + '\n')
        f.write(gttxt + '\n')

        
    # Read contents of HDF file into session memory
    geoms_obj = pycr8ascii.Geoms(hdf_qh, hdf_dir)
    # Write output to log file
    tmp_log_file = os.path.join(hdf_dir, 'pycr8ascii.log')
    if os.path.isfile(tmp_log_file):
        os.remove(tmp_log_file)
    Path(tmp_log_file).touch()

    #geoms_obj.set_global_log_file_name(os.path.join(hdf_dir, 'pycr8ascii.log'))
    geoms_obj.set_global_log_file_name(tmp_log_file)
    geoms_obj.get_data()

    # Check for read error etc

    if optqh == 1:
        geoms_iteration=1 # identifies which check to perform in GEOMS_VERSION_CHECK
        newdfv=['']*3  # returns DFV index in ga, new DFV value and label, and boolean to
        # indicate whether DFV is present in the original Global Attributes or not
        geoms_version_check(geoms_iteration, geoms_obj.ga_d, newdfv, nodfvinc)

        # Also need to check for one-off corrections concatenated to the TAV file
        geoms_one_off_check(tav_file, geoms_obj.ga_d, list_changes)

        # # Also need to perform corrections based on revised templates
        # GEOMS_TEMPLATE_CHANGE,ga,sds,list_changes
        # # Also need to check that dependent variables are not list before the variables they are dependent on
        # DEPENDENCY_ORDER,ga,sds,list_changes

    # Read pycr8ascii.log contents
    with open(tmp_log_file, 'r') as f:
        pycr8ascii_log = f.readlines() # Note: qh_s1 in IDL code

    # ESO 4/07/2024 temporarily removing log file (should not be left on disk
    # when running qa).
    os.remove(tmp_log_file)
    

    # NB: optqh == 1 (harmonisation) mostly not implemented yet

    if optqh == 0:  # QA option so identify FILE_META_VERSION
        # Identify TAV files needed for checks based on FILE_GENERATION_DATE and FILE_META_VERSION values
        tav_list, gf_err, cur_tav = tav_files_for_qa(tav_file, geoms_obj.ga_d, qhlog)

        n_tav_chks = 2 if all([i != '' for i in tav_list]) and tav_list[0] != tav_list[1] else 1

        # Change INFORMATION text in pycr8ascii.log entries to ERROR
        # pdb.set_trace()
        for i_dx, l_line in enumerate(pycr8ascii_log):
            if 'INFORMATION' in l_line:
                qhs1error=True
                pycr8ascii_log[i_dx] = l_line.rstrip().replace('INFORMATION', '  ERROR')


        # TODO: Check for 'Calibrated Data' messages etc

        # Delete log file if it is present in the output directory (a new file will be created by idlcr8hdf)
        log_file = os.path.join(qh_dir, logname)
        if os.path.isfile(log_file):
            os.remove(log_file)

        for j in range(n_tav_chks):
            pycr8hdf.run_qa(tav_list[j], geoms_obj, qh_dir, global_log_file=log_file)

        # Read contents of idlcr8hdf logfile into a string array
        with open(log_file, 'r') as f:
            pycr8hdf_log = f.readlines() # Note: qh_s2 in IDL code

        # Delete log file
        os.remove(log_file)


        tavi = [s for s in pycr8hdf_log if 'Input TAV File' in s]
        tavi_idx, _ = pycr8hdf.where_eq([i[:14] for i in pycr8hdf_log], 'Input TAV File')

        tav_name = []
        for j in range(len(tavi)):
            tav_name.append(tavi[j][15:])

        tav_inputs=['Input TAV file calculated from FILE_GENERATION_DATE:',
                    'Input TAV file calculated from FILE_META_VERSION:',
                    'Input TAV file calculated from FILE_GENERATION_DATE and FILE_META_VERSION:']

        if not gf_err and not cur_tav and n_tav_chks == 1:
            # TAV files determined from FGD and FMV are the same
            pycr8hdf_log[tavi_idx[0]] =  tav_inputs[2] + tav_name[0]

        elif not gf_err and not cur_tav and n_tav_chks == 2:
            # TAV files determined from FGD and FMV are different
            for j in (0,1):
                pycr8hdf_log[tavi_idx[j]] =  tav_inputs[j] + tav_name[j]
        elif n_tav_chks == 2:
            pycr8hdf_log[tavi_idx[0]] =  tav_inputs[1] + tav_name[1]

        # ELSE: Only check is with input TAV file so no change needed to qh_s2

        if n_tav_chks == 2:
            tnh = [ti.replace('_idl.dat', '') for ti in tav_name]
            tavrv = [os.path.basename(ff).strip()[-10:-4] for ff in tnh]
            tav_err_chk = [False]*2
            # Check QA using FGD or input TAV for errors
            for j in range(tavi_idx[1]):
                if pycr8hdf_log[j].strip()[:5] == 'ERROR':
                    tav_err_chk[0]=True
            # Check QA using FMV TAV for errors
            for j in range(tavi_idx[1], len(pycr8hdf_log)):
                if pycr8hdf_log[j].strip()[:5] == 'ERROR':
                    tav_err_chk[1]=True

            #pdb.set_trace()
            if tav_err_chk[1]:
                if tav_err_chk[0]:
                    pycr8hdf_log[tavi_idx[1]+1]='  ERROR: QA errors also generated using '+tavrv[1]+' definitions'
                else:
                    pycr8hdf_log[tavi_idx[1]+1]='  ERROR: QA errors generated using '+tavrv[1]+' definitions.'
                    pycr8hdf_log[tavi_idx[1]+1]=pycr8hdf_log[tavi_idx[1]+1]+' Please update first FILE_META_VERSION value to ' + tav_list[0]
            else:
                if tav_err_chk[0]:
                    pycr8hdf_log[tavi_idx[1]+1]='  ERROR: Definitions based on FILE_META_VERSION='+tavrv[1]+' are out-of-date.'
                    pycr8hdf_log[tavi_idx[1]+1]=pycr8hdf_log[tavi_idx[1]+1]+' Please update first FILE_META_VERSION value to '+ tavrv[0]+' and use these TAV definitions'
                else:
                    if qhs1error:
                        pycr8hdf_log[tavi_idx[1]+1]='  ERROR: QA errors also generated using '+tavrv[1]+' definitions'
                    else:
                        pycr8hdf_log[tavi_idx[1]+1]='  INFORMATION: GEOMS file passed using '+tavrv[1]+' definitions'
            # Don't want to keep additional mesasages from 2nd TAV check
            pycr8hdf_log=pycr8hdf_log[0:tavi_idx[1]+2]


        # Move Template file identifier line if present
        # ei = [s for s in pycr8hdf_log if 'Required Metadata Template' in s]

        # TODO: additional processing of the log strings.


        # Combine the log strings
        if len(pycr8ascii_log) > 0:
            qh_str = pycr8hdf_log[0:3] + pycr8ascii_log
            qh_str = qh_str + pycr8hdf_log[3:]
        else:
            qh_str = pycr8hdf_log

        pycr8ascii_log = qh_str

        # Search contents of log file to check for success or failure (transfer to qhlog)
        acnt = len([s for s in pycr8ascii_log if '  ERROR ' in s])
        bcnt = len([s for s in pycr8ascii_log if '  ERROR:' in s])
        ccnt = len([s for s in pycr8ascii_log if '  NON-ST' in s])

        ei=0
        if acnt != 0 or bcnt != 0 or gf_err:
            ei=2
        elif ccnt != 0:
            ei=1

        gt_error_code=int(gtcode[ei][0])



        with open(qhlog, 'a') as f:
            for line in pycr8ascii_log:
                # Make sure each line ends with a newline character
                if line[-1] != '\n':
                    f.write(line + '\n')
                else:
                    f.write(line)
            f.write('\n' + gtcode[ei] + '\n')


    return gt_error_code
