"""
GEOMS data version name checker tool

"""

import os
import re
import glob
from pathlib import Path
import numpy as np
import pycr8ascii



def determine_current_file(chFileGEOMSvnspec):
    """
    Determine the most up to date list file based on the version in the file name (e.g. 01R001)

    """

    file_list = glob.glob(chFileGEOMSvnspec)
    vn_cnt = len(file_list)

    if vn_cnt == 0:
        return -1
    elif vn_cnt == 1:
        return file_list[0]
    else:
        ver_list = []
        rev_list = []
        path_fix = chFileGEOMSvnspec.split('*')
        for f in file_list:
            # Assuming version/revision format 'ddrddd'
            #ver_rev = f.split('data_version_names_')[1].split('.')[0].split('r')
            ver_rev = f.split(path_fix[0])[1].split('.')[0].split('r')
            ver_list.append(ver_rev[0])
            rev_list.append(ver_rev[1])

        max_ver = np.max([int(i) for i in ver_list])
        max_rev = np.max([int(i) for i in rev_list])

        newest_file = path_fix[0] + "{:02d}".format(max_ver) + 'r' + "{:03d}".format(max_rev) + path_fix[1]

        return newest_file

def return_ga_check_values(ga_d):
    """
    Return all the global attribute values necessary to do the Version Name checks
    including VERSION_NAME, DATA_SOURCE_01, FILE_ACCESS, PI_NAME, AFFILIATION (from DATA_SOURCE)
    and DATA_LOCATION. Note: returns values in uppercase (1st list) and as written to the file
    (2nd list)

    """

    # r_vals_0 = [' ']*6
    # r_vals_1 = [' ']*6

    r_vals_0 = ['']*6
    r_vals_1 = ['']*6


    if 'DATA_SOURCE' in ga_d:
        dsv = ga_d['DATA_SOURCE']
        dsx = dsv.split('_')
        cdsx = len(dsx)
        if cdsx == 3:
            r_vals_0[0] = dsx[2].upper()
            r_vals_1[0] = dsx[2]
        if cdsx >= 1:
            r_vals_0[1] = dsx[0].upper()
            r_vals_1[1] = dsx[0]
        if cdsx >= 2:
            r_vals_0[4] = dsx[1][0:len(dsx[1])-3]
            r_vals_1[4] = dsx[1][0:len(dsx[1])-3]

    if 'FILE_ACCESS' in ga_d:
        dsv = ga_d['FILE_ACCESS']
        r_vals_0[2] = dsv.upper().strip()
        r_vals_1[2] = dsv.strip()

    if 'PI_NAME' in ga_d:
        dsv = ga_d['PI_NAME']
        r_vals_0[3] = ''.join([i.upper() for i in dsv.split()])
        r_vals_1[3] = ''.join(dsv.split())

    if 'DATA_LOCATION' in ga_d:
        dsv = ga_d['DATA_LOCATION']
        r_vals_0[5] = dsv.upper().strip()
        r_vals_1[5] = dsv.strip()

    #pdb.set_trace()

    return r_vals_0, r_vals_1


def incl_excl_checks(kw_name, chvn_chk, chvn_excl_incl, ga_chk_vals_0,
                     ga_chk_vals_1, log_file):
    """
    Check against excluded (~) or included DATA_SOURCE_01, FILE_ACCESS, PI_NAME, AFFILIATION
    (from DATA_SOURCE) and DATA_LOCATION values

    """

    chtxtvn = ' with this version name. Restriction criteria are: '
    ico = 0                       # default value for successful check - changes to -1 if an error is found
    n_chks = len(chvn_excl_incl)

    # Insert an empty element to avoid re-indexing the loops below
    chvn_chk.insert(0, '')


    if all([i == '' for i in chvn_chk]):
        chMessage = '  INFORMATION: No restrictions apply for keyword ' + kw_name
        print(chMessage)
        with open(log_file, 'a') as f:
            f.write(chMessage + '\n')

        return ico
    else:
        for i in range(1, n_chks):
            if chvn_chk[i] != '':
                # Separate out into component values (separated by '|' or '&')
                ds_vals = re.split('\||\&', chvn_chk[i])
                ds_vals = [i.strip() for i in ds_vals]

                # Separate inclusion and exclusion values
                evi = [True if i[0] == '~' else False for i in ds_vals]
                evcnt = evi.count(True)
                ivi = [not k for k in evi]
                ivcnt = ivi.count(True)

                # Lists of inclusion/exclusion values found
                if evcnt > 0:
                    ev_list = [ds_vals[k] for k in range(evcnt) if evi[k]]
                if ivcnt > 0:
                    iv_list = [ds_vals[k] for k in range(ivcnt) if ivi[k]]



                if chvn_excl_incl[i] == 'FILE_ACCESS':
                # Can have multiple values in the GEOMS file
                    ga_fa_vals = re.split(' |;', ga_chk_vals_0[i])
                    ga_fa_vals = [i.strip() for i in ga_fa_vals]
                    fa_cnt = len(ga_fa_vals)

                    #1. need to check that all ga_fa_vals are in the 'included' list
                    #2. need to check that none of the ga_fa_vals are in the 'excluded' list
                    fa_ok = [False]*fa_cnt
                    for j in range(fa_cnt):
                        if ga_fa_vals[0] in ds_vals:
                            fa_ok[j] = True
                        if '~' + ga_fa_vals[0] in ds_vals:
                            fa_ok[j] = False

                    if all(fa_ok):
                        chtxt = 'value' if fa_cnt == 1 else 'values'
                        chMessage = '  INFORMATION: '+ga_chk_vals_1[i]+' FILE_ACCESS '+chtxt+' permitted'
                    else:
                        failed_fa = []
                        for j in range(fa_cnt):
                            if not fa_ok[j]:
                                failed_fa.append(ga_fa_vals[j])

                        if len(failed_fa) == 1:
                            chtxt = 'value'
                        else:
                            chtxt = 'values'
                        failed_fa = ';'.join([k for k in failed_fa])
                        chMessage = '  ERROR: '+failed_fa+' FILE_ACCESS '+chtxt+' not permitted'
                        chMessage += chtxtvn+chvn_chk[i]
                        ico = -1

                    print(chMessage)
                    with open(log_file, 'a') as f:
                        f.write(chMessage + '\n')

                elif chvn_excl_incl[i] == 'DATA_SOURCE_01':
                    # Allow ds_vals to include multiple DATA_SOURCE_01 options e.g. UVVIS.DOAS
                    # Won't pass if it fails an exclusion check
                    passcheck = True     #set default to DATA_SOURCE_01 value being OK

                    if evcnt != 0:
                    # Check if file value is on the list of exclusions - if so then fails check
                        for j in range(evcnt):
                            if ev_list[j][1:] in ga_chk_vals_0[i]:
                                passcheck = False

                    # Note: only need to do inclusion checks if it has passed, or there are no,
                    # exclusion checks
                    if ivcnt != 0 and passcheck:
                    #check if file value is on the list of inclusions - if not then fails check
                        passcheck = False # set default to DATA_SOURCE_01 value not being OK
                        for j in range(ivcnt):
                            if iv_list[j] in ga_chk_vals_0[i]:
                                passcheck = True

                    if not passcheck:
                        chMessage = ('  ERROR: ' + ga_chk_vals_1[i] + ' ' + chvn_excl_incl[i]+
                                     ' value not permitted')
                        chMessage += chtxtvn + chvn_chk[i]
                        ico = -1
                    else:
                        chMessage = ('  INFORMATION: '+ga_chk_vals_1[i] + ' ' +
                                     chvn_excl_incl[i]+' value permitted')
                    print(chMessage)
                    with open(log_file, 'a') as f:
                        f.write(chMessage + '\n')



                else:
                # do checks on the rest of the criteria
                # Won't pass if it fails an exclusion check
                    passcheck = True    # set default to value being OK
                    if evcnt != 0:
                        # check if file value is on the list of exclusions - if so then fails check
                        for j in range(evcnt):
                            if ga_chk_vals_0[i] == ev_list[j][1:]:
                                passcheck = False
                    #Note: only need to do inclusion checks if it has passed, or there are no, exclusion checks
                    if ivcnt != 0 and passcheck:
                    #check if file value is on the list of inclusions - if not then fails check
                        passcheck = False # set default to DATA_SOURCE_01 value not being OK
                        for j in range(ivcnt):
                            if ga_chk_vals[i,0] == iv_list[j]:
                                passcheck = True

                    if not passcheck:
                        chMessage = ('  ERROR: ' + ga_chk_vals_1[i] + ' ' + chvn_excl_incl[i]+
                        ' value not permitted')
                        chMessage += chtxtvn + chvn_chk[i]
                        ico = -1
                    else:
                        chMessage = ('  INFORMATION: '+ga_chk_vals[i,1]+' '+
                                     chvn_excl_incl[i]+' value permitted')

                    print(chMessage)
                    with open(log_file, 'a') as f:
                        f.write(chMessage + '\n')

                    #  else:
                    # ;chMessage = '  INFORMATION: No '+chvn_excl_incl[i]+' restrictions apply'
                    # ;printf, lu, chMessage
                    # ;print, chMessage


    return ico


def analyze_GEOMS_VN(chvn_value, chFileGEOMSvn, ga_chk_vals_0, ga_chk_vals_1, log_file):
    """
    analyze_GEOMS_VN

    """

    tccode = ['0 (Passed version name checks)', '1 (Failed version name checks)']
    iCheckOut = 0

    chvn_excl_incl = ['VERSION_NAME', 'DATA_SOURCE_01', 'FILE_ACCESS', 'PI_NAME',
                      'ORGANIZATION', 'DATA_LOCATION'] # must match the order in the file


    #pdb.set_trace()
    if not os.path.isfile(chFileGEOMSvn):
        chMessage = '  ERROR: Version Name list file not found'
        print(chMessage)
        with open(log_file, 'a') as f:
            f.write(chMessage + '\n')
        iCheckOut = -1
    else:

        # read in contents of the list file
        #dum='' & n_vn = 0L
        with open(chFileGEOMSvn) as fu:
            dum = fu.readlines()


        # Determine number of version keywords - exclude comment lines or spaces
        chvernames_list = [i.split(',') for i in dum if i[0] != '!' and i[0] != '']
        chvernames_dict={}
        for key, value in zip([i[0] for i in chvernames_list], [i[1:] for i in chvernames_list]):
            chvernames_dict.update({key : value})

        # Check 1 - Does the version name consist of keyword part(s), optionally followed by
        # version number part(s)
        pres = chvn_value.split('.')
        chMessage = ''
        err_mess=False
        nni = [i.isdecimal() for i in pres]

        if '.' in (pres[0][0], pres[-1][-1]):
            chMessage += ('ERROR: Dot present at start or end of data version name: ' + chvn_value  + '\n')
            iCheckOut = 1
            err_mess=True
        if '' in pres:
            chMessage += ('ERROR: Data version name contains consecutive dots: ' + chvn_value  + '\n')
            iCheckOut = 1
            err_mess=True
        if ' ' in chvn_value:
            chMessage += ('ERROR: Data version name contains spaces: ' + chvn_value  + '\n')
            iCheckOut = 1
            err_mess=True


        if err_mess:
            print(chMessage)
            with open(log_file, 'a') as f:
                f.write(chMessage)


        if iCheckOut == 0:
            if not False in nni:
                # version name consists only of version numbers so no further checks required
                chMessage = ('  INFORMATION: Version name ' + chvn_value + ' consists of a processing '
                               'version number only. No further checks required.')
                print(chMessage)
                with open(log_file, 'a') as f:
                    f.write(chMessage + '\n')

                iCheckOut = 2
            elif False in nni and len(nni) != 0:
                # TODO:
                 #the keyword part(s) come(s) before the version number part(s) so write keywords to chvn_value
                if sorted(nni) != nni:
                    chMessage = ('  ERROR: Data version number must be after the data processing keyword: ')
                    chMessage += chvn_value
                    print(chMessage)
                    with open(log_file, 'a') as f:
                        f.write(chMessage + '\n')

                    iCheckOut = 1

        # Check 2 - Are the keyword part(s) present in the list
        if iCheckOut== 0:
            # Check each keyword or combination until all keywords are accounted for or an error is found
            keyword = chvn_value.upper()
            keyword_nonum = [i.upper() for i in keyword.split('.') if not i.isdecimal()]
            chvn_allwdup = [i[0].upper() for i in chvernames_list]

            kw_str = ''
            if len(keyword_nonum) > 1:
                kw_str = ' part'
            for kw_find in keyword_nonum:
                if kw_find in chvn_allwdup:
                    chMessage = '  INFORMATION: Version keyword match found for ' + kw_find + kw_str
                    # partial or full match found so do inclusion/exclusion checks
                    print(chMessage)
                    with open(log_file, 'a') as f:
                        f.write(chMessage + '\n')

                    ico = incl_excl_checks(kw_find, chvernames_dict[kw_find],
                                           chvn_excl_incl, ga_chk_vals_0, ga_chk_vals_1, log_file)
                    if ico == -1:
                        iCheckOut = ico

                else:
                    chMessage = '  ERROR: Version keyword match not found for ' + kw_find + kw_str
                    print(chMessage)
                    with open(log_file, 'a') as f:
                        f.write(chMessage + '\n')
                    iCheckOut = 1




    if iCheckOut == 2:
        iCheckOut = 0 # No checks required as version name consists of a version number only

    with open(log_file, 'a') as f:
        f.write('\n' + tccode[abs(iCheckOut)] + '\n')

    #pdb.set_trace()

    return(iCheckOut)



def geoms_vntools(dvndir, hdffile, QA_error_code, TC_error_code):
    """
    geoms_vntools

    """

    TC_error_code_2 = 0
    chFileGEOMSvnspec = os.path.join(dvndir, 'data_version_names_*.csv')

    # determine name of most up to date list file
    chFileGEOMSvn = determine_current_file(chFileGEOMSvnspec)

    #reterr=''
    #pycr8ascii(hdffile, ga, sds, reterr)

    # Find log file
    log_file = Path(os.path.abspath(hdffile)).with_suffix('.log')


    # Initialize object, read, and write metadata + data
    out_dir = None
    geoms_obj = pycr8ascii.Geoms(hdffile, out_dir) #
    # geoms_obj.set_global_log_file_name(log_file)
    geoms_obj.get_data()



    # Determine if DATA_SOURCE has a VERSION_NAME field
    # Get the global attribute values required for the checks
    ga_chk_vals_0, ga_chk_vals_1 = return_ga_check_values(geoms_obj.ga_d)
    # Version of the name as written in the GEOMS file instead of uppercase
    chvn_value = ga_chk_vals_1[0]

    if chvn_value != '':
        if chFileGEOMSvn == -1:
            chMessage = 'ERROR: GEOMS Version Name list file not found'
            print(chMessage)
            with open(log_file, 'a') as f:
                f.write(chMessage + '\n')
            TC_error_code_2 = -1
        else:
            chMessage = 'Analyzing compliance to GEOMS version name list: ' + \
                os.path.basename(chFileGEOMSvn)
            print(chMessage)
            with open(log_file, 'a') as f:
                f.write(chMessage + '\n')

            TC_error_code_2 = analyze_GEOMS_VN(chvn_value, chFileGEOMSvn, ga_chk_vals_0,
                                               ga_chk_vals_1, log_file)

        # if TC_error_code_2 eq -1 then chFileGEOMSvn is not present (i.e. problem with input file)
        if TC_error_code_2 == -1 or QA_error_code == 3:
            chMessage = '3 (Total QA/TC-VN error)'
        else:
            if TC_error_code_2 < TC_error_code:
                TC_error_code_2 = TC_error_code
            chMessage = str( QA_error_code + 3*TC_error_code_2 + 4 ) + ' (Total QA/TC-VN error)'
        with open(log_file, 'a') as f:
            f.write('\n' + chMessage + '\n')
        return TC_error_code_2
    else:
        if QA_error_code == 3:
            with open(log_file, 'a') as f:
                f.write('\n' + '3 (Total QA/TC error)' + '\n')

        return TC_error_code


if __name__ == "__main__":
    # For testing
    # chFileGEOMSvnspec = '/homevip/espen/repos/geoms-documentation/' + 'data_version_names_*.csv'
    # newest_file = determine_current_file(chFileGEOMSvnspec)
    # print(newest_file)
    pass
