#!/usr/bin/env python3

"""

pycr8ascii.py

DESCRIPTION
    Create ASCII Metadata and Data files from GEOMS files in hdf/netcdf format

REQUIREMENTS
    Tested with Python versions 3.7 & 3.10 under Ubuntu Linux. 
    Import libraries (h5py, pyhdf etc.) needs to be installed.

EXAMPLE USAGE
    Write file.meta and file.data files to current directory, using file.h5 as input:

      pycr8ascii.py -f file.h5 -o . 
    
CONTACT 
    eso@nilu.no (Espen Sollum)


"""

import os
import sys
import argparse
import pdb
from collections import OrderedDict
import numpy as np
from netCDF4 import Dataset
import h5py
from pyhdf import SD

# This does not quite work. Intended to read back data in creation order.
h5py.get_config().track_order=True

# log file is a global variable for now
global_log_file_name = None


class FileReader:
    """
    Class for reading and processing GEOMS data in various formats.

    """

    def __init__(self, file_name):
        """
        Initialize object with name of file to be read.

        """

        self.file_name = file_name
        self.file = None
        self.file_type = None
        self.all_datasets = None

    def open_file(self):
        """
        The specific command to open files differs between formats and is implemented in subclasses.

        """

        print("Reading {}".format(os.path.basename(self.file_name)))

    def get_global_attributes(self):
        """
        Get global attributes from file. Specific implementations in subclasses.

        """

    def get_all_datasets(self):
        """
        Get all datasets from file. Specific implementations in subclasses.

        """

    def get_variable_attributes(self):
        """
        Get variable attributes

        """

        self.var_attrs = OrderedDict()

    def get_var_data(self, key):
        """
        Get variable data from file

        """

    def read_data(self):
        """
        Operations common to all file format readers here
        Implementations of read_data return sdss_d (variables) and ga_d (global attributes)

        Method to read the contents of a GEOMS standard netCDF compatible file

        Possible Conditions for STOP_WITH_ERROR call
        1. No global attributes or datasets present in the root group of an HDF5 file.
        2. The number of global attributes or datasets is zero or not able to be determined
          in an HDF4 file.

        Information Conditions (when the program is able to make changes):
        1. DATA_VARIABLES global attribute not found or has no values.
        2. Number of datasets given under DATA_VARIABLES is not equal to the number saved to the
          file.
        3. The number of dimensions in a dataset exceeds 8.
        4. The data type of a dataset is not BYTE, SHORT, INTEGER, LONG, REAL, DOUBLE, or STRING.
        5. Dataset name is truncated or does not match VAR_NAME value.
        6. Metadata label VAR_NAME not found for the dataset.
        7. VAR_NAME does not match any DATA_VARIABLES values.
        8. No Variable Attributes returned after HDF call.
        9. Metadata Dataset listing order may not match the order of the Global Attribute
          DATA_VARIABLES Values.
        10. Variable is written to the HDF file as a coordinate variable.

        """

        # Note any calibrated (scaled) HDF4 dataset is corrected using the formula specified in
        # UG_print42r3.pdf pg. 3-107: orig = cal * (cal_val - offset)
        ncsa_cal=['scale_factor','scale_factor_err','add_offset','add_offset_err','calibrated_nt']

        # Initialize scalar to indicate whether VAR_DEPEND and VAR_SIZE values need to be reversed
        rev_vd_vs=2 # 1 = reverse attribute values; 0 = no reverse (and will generate message)
        # 2 = no action required; 3 = rev_vd_vs changed between 0 and 1 during checking, therefore error
        # Note if rev_vd_vs is 2 or 3 after all the checks then it will be changed to 1 so that dimensions
        # are swapped by default

        # List of Standard GEOMS Variable Attributes
        attr_arr_data=['VAR_NAME','VAR_DESCRIPTION','VAR_NOTES','VAR_SIZE','VAR_DEPEND',
                       'VAR_DATA_TYPE','VAR_UNITS','VAR_SI_CONVERSION','VAR_VALID_MIN',
                       'VAR_VALID_MAX','VAR_FILL_VALUE']

        # List of required attributes that will be checked for existence
        attr_data_required = ['VAR_DESCRIPTION', 'VAR_DEPEND',
                              'VAR_UNITS','VAR_SI_CONVERSION','VAR_VALID_MIN',
                              'VAR_VALID_MAX','VAR_FILL_VALUE']

        # for FILE_NAME checks
        in_file_chk=os.path.basename(self.file_name).lower()


        #Possible error messages for this procedure
        errtxt=[]
        errtxt.append(' found in the Root Group of the HDF5 File.')
        errtxt.append(' is zero or not able to be determined (check for corrupted file).')

        # Save values as is if dataset is of type STRING
        slabel=['VAR_UNITS','VAR_SI_CONVERSION','VAR_VALID_MIN','VAR_VALID_MAX','VAR_FILL_VALUE']

        # Dataset and attribute labels for data type checks
        dt_chk_labels=['Dataset','VAR_VALID_MIN','VAR_VALID_MAX','VAR_FILL_VALUE','VAR_DATA_TYPE']
        n_lab=len(dt_chk_labels)

        n_vn=0 # Initialize the number of Variable Names in the standard format h5 file

        # For storing the data variables
        #sdss_d = {}
        sdss_d = OrderedDict()

        self.get_all_datasets()
        n_sds = len(self.all_datasets)

        # Read in the Global Attribute labels and values
        ga_d = self.get_global_attributes()

        # Get all variable attributes
        self.get_variable_attributes()

        # Extract the list of the DATA_VARIABLES
        try:
            vn = ga_d['DATA_VARIABLES'].split(';')
            n_vn = len(vn)
        except KeyError:
            infotxt_output_a([0],[n_vn, n_sds , ':TODO: information message'])

            # If global attribute DATA_VARIABLES not present, construct it from
            # datasets found in file
            ga_d.update({'DATA_VARIABLES' : ';'.join(str(i) for i in self.all_datasets)})
            vn = ga_d['DATA_VARIABLES'].split(';')
            n_vn = len(vn)

        # Check that the actual file name matches the FILE_NAME entry
        try:
            if ga_d['FILE_NAME'].lower() != in_file_chk:
                infotxt_output_a([24], [ga_d['FILE_NAME']])
        except KeyError:
            print("ERROR: global attribute 'FILE_NAME' not found in file")
            ga_d.update({'FILE_NAME' : in_file_chk})


        if n_sds != 0:
            if n_vn != n_sds:
                # DATA_VARIABLES values not read successfully, or number of datasets given under DATA_VARIABLES
                # is not equal to the number saved to the file
                infotxt_output_a([0],[n_vn, n_sds,2, ':TODO: information message'])

            # Determine correct dataset order, and dimension order
            # oi=np.zeros((n_sds), dtype='int')

            # Iterate over all datasets found in file
            for i, key in enumerate(self.all_datasets):
                hdftype=['']*(n_lab-1)# to hold data types for dataset, min, max and fill values

                # Structure for variable data and attributes
                sds = SDS()
                
                # Get data, datatype and number of dimensions
                sds.data, dt_datatype, sds_dim = self.get_var_data(key)
                sds.dim = sds_dim

                sds_name = key

                # Need to swap non-alphanumeric characters to '_' to match names from h5_parse
                # for attribute values test
                # h5p_name="".join([ c if c.isalnum() else "_" for c in s ])
                # IF sds_name EQ '' THEN sds_name='N/A'

                # Determine dataset datatype
                # dt_datatype = self.file[key].dtype

                hdftype[0]=dt_datatype

                # Determine the number of attributes associated with the dataset
                sds_natts = len(self.var_attrs[key].keys())

                if sds_natts != 0:
                    ci=['']*4 # array containing catalog attributes

                    # Read in Dataset Attributes

                    # If attribute VAR_NAME is missing, set it to variable name in file
                    if 'VAR_NAME' not in self.var_attrs[key].keys():
                        sds.va_l.insert(0, 'VAR_NAME')
                        sds.va_v.insert(0, key)
                        infotxt_output_a([4,0], [key])


                    # If attribute VAR_DATA_TYPE is missing, set it to variable data type from file
                    if 'VAR_DATA_TYPE' not in self.var_attrs[key].keys():
                        sds.va_l.append('VAR_DATA_TYPE')
                        d_type = geoms_data_type(self.file[key].dtype.type.__name__)
                        sds.va_v.append(d_type)

                        infotxt_output_a([10,0], [key,1,2])


                    # If any required attributes are missing, print information
                    for a_attr in attr_data_required:
                        if a_attr not in self.var_attrs[key].keys():
                            infotxt_output_a([25], [a_attr, key])


                    for akey in self.var_attrs[key]:
                        da_name=akey

                        if 'bytes_' in type(self.var_attrs[key][akey]).__name__:
                            va_value = self.var_attrs[key][akey].decode()
                        else:
                            va_value = self.var_attrs[key][akey]

                        if isinstance(va_value, str):
                            at_datatype='STRING'
                            va_type='STRING'
                        else:
                            # VAR_FILL_VALUE, VAR_VALID_MAX, VAR_VALID_MEAN should be numeric types
                            #  Does not work with hdf4 - can not see the difference between 32/64 bits
                            if self.file_type == 'H4':
                                at_datatype=type(va_value)
                            else:
                                at_datatype=va_value.dtype

                            va_type='NON_STRING'

                        if da_name.upper() == 'VAR_NAME':
                        # Check that the dataset name matches the VAR_NAME
                            if va_value != key:
                                use_sds_name=1
                                va_value_tmp = va_value
                                sds_name_tmp = sds_name

                                if key in vn:
                                    # Dataset name exists in DATA_VARIABLES: change VAR_NAME
                                    va_value = key
                                else:
                                    # Replace variable name in DATA_VARIABLES
                                    vn = [va_value if i == key else i for i in vn]
                                    sds_name = va_value
                                    use_sds_name=0

                                infotxt_output_a([3,4, use_sds_name],[sds_name_tmp,va_value_tmp])

                            ci[0]=va_value

                        elif da_name.upper() == 'VAR_DATA_TYPE':
                            var_data_type = check_var_data_type(dt_datatype,
                                                                 va_value, sds_name, key)

                            # Note: the IDL code converts int8 to uint8 so we do the same
                            # here. Add some checks on valid min/max/fill
                            va_value = var_data_type
                            if var_data_type == "INT8":
                                va_value = "BYTE"

                            ci[1]=va_value

                        elif da_name.upper() == 'VAR_UNITS':
                            if va_type != 'STRING':
                                infotxt_output_a([20],['VAR_UNITS', sds_name ])

                        elif da_name.upper() == 'VAR_DEPEND:':
                            test_dim_order('VD',va_value,va_type,sds_dim,sds_name,rev_vd_vs)

                        elif da_name.upper() == 'VAR_SIZE':
                            test_dim_order('VS',va_value,va_type,sds_dim,sds_name,rev_vd_vs)

                        elif da_name.upper() == 'VAR_VALID_MIN':
                            hdftype[1] = at_datatype
                            # Hack for hdf4 files - check if it is necessary to rewrite
                            if self.file_type == 'H4':
                                va_value = np.array([va_value], dtype=dt_datatype)[0]
                            var_data_type = check_var_valid_fill(dt_datatype,
                                                                 va_value, sds_name, key,
                                                                 da_name.upper())

                        elif da_name.upper() == 'VAR_VALID_MAX':
                            if self.file_type == 'H4':
                                va_value = np.array([va_value], dtype=dt_datatype)[0]
                            var_data_type = check_var_valid_fill(dt_datatype,
                                                                 va_value, sds_name, key,
                                                                 da_name.upper())
                            #va_value = var_data_type
                            hdftype[2] = at_datatype

                        elif da_name.upper() == 'VAR_FILL_VALUE':
                            if self.file_type == 'H4':
                                va_value = np.array([va_value], dtype=dt_datatype)[0]

                            var_data_type = check_var_valid_fill(dt_datatype,
                                                                 va_value, sds_name, key,
                                                                 da_name.upper())
                            if var_data_type == "INT8":
                                va_value = np.uint(var_data_type)
                            hdftype[3] = at_datatype


                        # TODO: code for array dimension order - ignored for now


                        # Put data into structure
                        va_name = akey.upper()

                        sds.va_l.append(va_name)
                        sds.va_v.append(va_value)

                        # Done reading data and attributes

                        # Copy the variable attribute labels and values to a dictionary
                        sds.va_d = {}
                        for k, v in zip(sds.va_l, sds.va_v):
                            sds.va_d.update({k:v})

                sdss_d.update({sds_name : sds})

        return sdss_d, ga_d


    def close_file(self):
        """
        Close the file

        """

        if self.file is not None:
            self.file.close()
            self.file = None

class Hdf5Reader(FileReader):
    """
    Read hdf5 files

    """


    def open_file(self):
        """
        Open file

        """

        h5py.get_config().track_order=True
        self.file_type = 'H5'
        self.file = h5py.File(self.file_name, 'r', track_order=True)

    def read_data(self):
        """
        Perform common file operations

        """

        sdss_d, ga_d = super().read_data()

        return sdss_d, ga_d

    def get_global_attributes(self):
        """
        Get global attributes from file.

        """

        ga_d = OrderedDict()

        # Read in the Global Attribute labels and values
        for ka in self.file['/'].attrs.keys():
            kae = ka
            # Check for _GLOSDS suffix and strip (added for compatibility with H4toH5 utility)
            if kae.endswith("_GLOSDS"):
                kae = ka[:-7]

            # Remove trailing underscores
            if kae.endswith("_"):
                kae = kae[:-1]

            # Add attribute/data to dictionary
            ga_d[kae] = self.file['/'].attrs[kae]

            if 'bytes' in type(self.file['/'].attrs[kae]).__name__:
                ga_d[kae] = self.file['/'].attrs[kae].decode()
            else:
                ga_d[kae] = self.file['/'].attrs[kae]

        return ga_d


    def get_variable_attributes(self):
        """
        Get variable attributes

        """

        #self.var_attrs = {}
        self.var_attrs = OrderedDict()

        for v_key in self.all_datasets:
            va_values = OrderedDict()
            for a_key in self.file['/'][v_key].attrs.keys():
                va_values.update({a_key: self.file['/'][v_key].attrs[a_key]})
            self.var_attrs.update({v_key: va_values})


    def get_var_data(self, key):
        """
        Get variable data.

        Returns variable data, datatype and number of dimensions

        """

        # if key == "TEMPERATURE.EFFECTIVE.O3_SOURCE":
        #     pdb.set_trace()


        # Take into account that data may be scalar or in an array
        if type(self.file['/'][key][()]).__name__ == 'ndarray':
            if 'bytes_' in type(self.file['/'][key][()][0]).__name__:
                data_out = np.array([j.decode() for j in self.file['/'][key][()]])
            else:
                data_out = self.file['/'][key][()]
        else:
            if 'bytes_' in type(self.file['/'][key][()]).__name__:
                try:
                    data_out = np.array([j.decode() for j in self.file['/'][key][()]])
                except AttributeError:
                    data_out = self.file['/'][key].asstr(encoding='ascii')[()]
            else:
                data_out = self.file['/'][key][()]

        # return self.file['/'][key][()], self.file['/'][key].dtype, self.file['/'][key].ndim
        return data_out, self.file['/'][key].dtype, self.file['/'][key].ndim


    def get_all_datasets(self):
        """
        Get all datasets from file

        """

        # Get number of attributes within the root group (these should be the Global Attributes)
        n_ga=len(self.file.attrs.keys())

        # Get the number of objects within the root group (including datasets, groups etc)
        n_obj=len(self.file.keys())

        if n_obj != 0:

            # d_obj=np.zeros(n_obj, dtype='int')
            # g_obj=np.zeros(n_obj, dtype='int')

            # Check all objects in file and determine if it is group or dataset
            all_objs = []
            self.file['/'].visit(all_objs.append)
            h5py.get_config().track_order=True
            all_groups   = [ obj for obj in all_objs if isinstance(self.file['/'][obj],h5py.Group) ]
            all_datasets = [ obj for obj in all_objs if isinstance(self.file['/'][obj],h5py.Dataset) ]

            n_sds = len(all_datasets)
            n_grp = len(all_groups)
            # if (n_ga == 0) and (n_sds == 0) and (n_grp != 0):
            #     stdfm=0
        else:
            n_sds=0

        self.all_datasets = all_datasets


class Hdf4Reader(FileReader):
    """
    Read hdf4 files

    """
    
    def open_file(self):
        """
        Open file

        """

        self.file_type = 'H4'
        self.file = SD.SD(self.file_name)

    def read_data(self):
        """
        Perform common file operations

        """

        sdss_d, ga_d = super().read_data()

        return sdss_d, ga_d

    def get_global_attributes(self):
        """
        Get global attributes from file.

        """

        ga_d = {}

        # Read in the Global Attribute labels and values
        for i, ka in enumerate(self.file.attributes().keys()):
            kae = ka
            # Check for _GLOSDS suffix and strip (added for compatibility with H4toH5 utility)
            if kae.endswith("_GLOSDS"):
                kae = ka[:-7]

            # Remove trailing underscores
            if kae.endswith("_"):
                kae = kae[:-1]

            # Add attribute/data to dictionary
            ga_d[kae] = self.file.attributes()[kae]


        return ga_d


    def get_variable_attributes(self):
        """
        Get variable attributes

        """

        self.var_attrs = {}

        for v_key in self.all_datasets:
            va_values = {}
            #            for a_key in self.file[v_key].attrs.keys():

            #for a_key in self.file.select(v_key):

            data_var = self.file.select(v_key)
            for a_key in data_var.attributes():
#                va_values.update({a_key: self.file[v_key].attrs[a_key]})
                va_values.update({a_key: data_var.attributes()[a_key]})

            self.var_attrs.update({v_key: va_values})

    def get_all_datasets(self):
        """
        Get all datasets from file

        """

        # Get number of attributes within the root group (these should be the Global Attributes)
        n_ga=len(self.file.attributes().keys())

        # Get the number of objects within the root group (including datasets, groups etc)
        n_obj=self.file.info()[0]

        if n_obj != 0:

            # d_obj=np.zeros(n_obj, dtype='int')
            # g_obj=np.zeros(n_obj, dtype='int')


            # Check all objects in file and determine if it is group or dataset
            # all_objs = []
            # self.file.visit(all_objs.append)
            # all_groups   = [ obj for obj in all_objs if isinstance(self.file[obj], h5py.Group) ]

            # all_datasets = [ obj for obj in all_objs if isinstance(self.file[obj], h5py.Dataset) ]

            all_datasets = self.file.datasets().keys()

            n_sds = len(all_datasets)
            # n_grp = len(all_groups)
            # if (n_ga == 0) and (n_sds == 0) and (n_grp != 0):
            #     stdfm=0
        else:
            n_sds=0

        #        return all_datasets
        self.all_datasets = all_datasets


    def get_var_data(self, key):
        """
        Get variable data

        Return variable data, datatype and number of dimensions

        """

        var_data = self.file.select(key)

        # For hdf4, return datatype according to
        # http://fhs.github.io/pyhdf/modules/SD.html#data-types

        if var_data.info()[3] == 4:
            dt_datatype = np.dtype(str)
        elif var_data.info()[3] in (3, 21):
            dt_datatype = np.dtype(np.uint8)
        elif var_data.info()[3] == 20:
            dt_datatype = np.dtype(np.int8)
        elif var_data.info()[3] == 22:
            dt_datatype = np.dtype(np.int16)
        elif var_data.info()[3] == 24:
            dt_datatype = np.dtype(int)
        elif var_data.info()[3] == 5:
            dt_datatype = np.dtype(np.float32)
        elif var_data.info()[3] == 6:
            dt_datatype = np.dtype(np.float64)
        else:
            # Investigate
            pdb.set_trace()

        # Deal with strings as array of bytes
        if var_data.info()[3] == 4:
            char_data = var_data.get()
            char_dim = var_data.info()[2][0]
            ret_data = [] #np.ndarray((char_dim))

            for i in range(char_dim):
                ret_data.append(''.join([j.decode() for j in char_data[i,:]]))

            ret_data = np.array(ret_data)

        else:
            ret_data = var_data.get()


        return ret_data, dt_datatype, var_data.info()[1]

    def close_file(self):
        """
        Close the file

        """

        if self.file is not None:
            self.file.end()
            self.file = None


class NetCDFReader(FileReader):
    """
    NetCDF file reader

    """
    
    def open_file(self):
        """
        Open file.

        """
        self.file_type = 'N4'
        self.file = Dataset(self.file_name, 'r')

    def get_global_attributes(self):
        """
        Get global attributes from file.

        """

        ga_d = {}

        # Read in the Global Attribute labels and values
        for key in self.file.ncattrs():
            kae = key
            # Check for _GLOSDS suffix and strip (added for compatibility with H4toH5 utility)
            if kae.endswith("_GLOSDS"):
                kae = kae[:-7]

            # Remove trailing underscores
            if kae.endswith("_"):
                kae = kae[:-1]

            # Add attribute/data to dictionary
            ga_d[kae] = getattr(self.file, kae)

        return ga_d

    def get_all_datasets(self):
        """
        Get all datasets from file

        """

        self.all_datasets = [i for i in self.file.variables.keys()]

    def get_variable_attributes(self):
        """
        Get variable attributes

        """

        self.var_attrs = {}

        for v_key in self.all_datasets:
            va_values = {}
            for a_key in self.file[v_key].ncattrs():
                va_values.update({a_key : self.file[v_key].getncattr(a_key)})
            self.var_attrs.update({v_key: va_values})

    def get_var_data(self, key):
        """
        Get variable data

        Return variable data, datatype and number of dimensions

        """

        return self.file[key][()], self.file[key].dtype, self.file[key].ndim


    def read_data(self):
        """
        Read Data

        """

        sdss_d, ga_d = super().read_data()
        return sdss_d, ga_d

class Geoms:
    """
    Class for reading and writing GEOMS metadata and data
    
    """

    def __init__(self, file_name, out_dir):
        """
        Initialize

        """
        self.file_reader = None
        self.file_name = file_name
        self.out_dir = out_dir

        # Variable attribute labels/values
        self.va_l = []
        self.va_v = []

        # Dictionary of {va_l : va_v} pairs
        self.va_d = {}

        # Variable data
        self.data = None
        # self.sdss_d = {}
        self.sdss_d = OrderedDict()


    def get_file_format(self):
        """
        Identify format of the input file.

        For now, set from file name extension (netCDF4 of hdf5). Can eventually implement
        something similar to the 'file_format_a' function from idlcr8ascii.pro 

        """

        # Get file extension
        _, ext = os.path.splitext(self.file_name)

        # File is hdf5
        if ext == '.h5':
            self.file_type = 'H5'
            self.file_reader = Hdf5Reader(self.file_name)
        # File is netCDF4
        elif ext == '.nc':
            self.file_type = 'N4'
            self.file_reader = NetCDFReader(self.file_name)
        # File is hdf4
        elif ext == '.hdf':
            self.file_type = 'H4'
            self.file_reader = Hdf4Reader(self.file_name)

        else:
            raise ValueError("Unsupported file format")


        # h4id = np.array([14, 3, 19], dtype='uint8')
        # h5id = np.array([137, 72, 68, 70], dtype='uint8') # HDF
        # nc3id = np.array([67, 68, 70], dtype='uint8') # CDF

        # # First test using start of file byte markers
        # with open(infile, 'rb') as f:
        #     fid = np.fromfile(f, dtype='uint8', count=4)

        #     idfound = 'XX'
        #     if np.array_equal(fid[0:3], h4id):
        #         idfound = 'H4'
        #     elif np.array_equal(fid, h5id):
        #         idfound = 'H5'
        #     elif np.array_equal(fid[0:2], nc3id):
        #         idfound = 'N3'
        #     else:
        #         pass

        # # Second test, if required, using Python commands
        # if idfound == 'XX':
        #     if h5py.is_hdf5(infile):
        #         idfound = 'H5'  # netCDF4 or HDF5
        #     else:
        #         try:
        #             with Dataset(infile, 'r') as f:
        #                 idfound = 'N3'  # If get to here then file is netCDF
        #         except OSError:
        #             pass

        # if idfound == 'H5':
        #     # For H5, try to identify whether it is H5 or N4
        #     nc4ext = ['nc', 'nc4', 'netcdf', 'netcdf4']
        #     fileext = os.path.splitext(infile)[-1][1:].lower()
        #     if fileext in nc4ext:
        #         try:
        #             with Dataset(infile, 'r', format='NETCDF4') as f:
        #                 idfound = 'N4'  # If get to here then file can be opened using
        #                 # N4 libraries, so most likely a netcdf4 file
        #         except:
        #             pass


    def open_file(self):
        """
        Open the file

        """

        self.file_reader.open_file()


    def read_data(self):
        """
        Read data

        """

        if self.file_reader is None:
            raise RuntimeError("File not opened")

        # Perform the file reading. Get variables and global attributes .
        self.sdss_d, self.ga_d = self.file_reader.read_data()

    def close_file(self):
        """
        Close the file

        """

        if self.file_reader is not None:
            self.file_reader.close_file()
            self.file_reader = None

    def set_global_log_file_name(self, file_name):
        """

        """

        global global_log_file_name

        global_log_file_name = file_name


    def get_data(self):
        """
        Combine the file operations into one method, for simplicity

        """

        self.get_file_format()
        self.open_file()
        self.read_data()
        self.close_file()


    def output_data(self):
        """
        Output metadata and data to ASCII files.

        """

        # HDF4 and NETCDF pre-defined attributes, to be excluded
        ncsa_exclude=['long_name','units','format','coordsys','valid_range','_fillvalue','scale_factor',
              'scale_factor_err','add_offset','add_offset_err','calibrated_nt']


        # List of Standard GEOMS Variable Attributes. Use as sort order for output
        attr_arr_data=['VAR_NAME','VAR_DESCRIPTION','VAR_NOTES','VAR_SIZE','VAR_DEPEND',
                       'VAR_DATA_TYPE','VAR_UNITS','VAR_SI_CONVERSION','VAR_VALID_MIN',
                       'VAR_VALID_MAX','VAR_FILL_VALUE']

        # Standard global attributes sort order.
        attr_arr_glob=['PI_NAME','PI_AFFILIATION','PI_ADDRESS','PI_EMAIL',
                       'DO_NAME','DO_AFFILIATION','DO_ADDRESS','DO_EMAIL',
                       'DS_NAME','DS_AFFILIATION','DS_ADDRESS','DS_EMAIL',
                       'DATA_DESCRIPTION','DATA_DISCIPLINE','DATA_GROUP','DATA_LOCATION',
                       'DATA_SOURCE','DATA_VARIABLES','DATA_START_DATE','DATA_STOP_DATE',
                       'DATA_FILE_VERSION','DATA_MODIFICATIONS','DATA_CAVEATS','DATA_RULES_OF_USE',
                       'DATA_ACKNOWLEDGEMENT','DATA_QUALITY','DATA_TEMPLATE','DATA_PROCESSOR',
                       'FILE_NAME','FILE_GENERATION_DATE','FILE_ACCESS','FILE_PROJECT_ID','FILE_DOI',
                       'FILE_ASSOCIATION','FILE_META_VERSION']

        sorted_ga = []
        unsorted_ga = []
        for i in attr_arr_glob:
            if i in self.ga_d.keys():
                sorted_ga.append(i)
            else:
                unsorted_ga.append(i)

        # Check for global attributes not in attr_arr_glob
        for i in self.ga_d.keys():
            if i not in sorted_ga:
                sorted_ga.append(i)

        sorted_ga += unsorted_ga


        # Create output filenames
        in_dir = os.path.split(os.path.abspath(self.file_name))[0]
        out_dir = self.out_dir
        in_file = os.path.basename(self.file_name)

        fn_base, _ = os.path.splitext(in_file)
        out_f1 = os.path.join(out_dir, fn_base + '.meta')
        out_f2 = os.path.join(out_dir, fn_base + '.data')

        # Open files
        with open(out_f1, 'w') as f_meta:

            # Write meta file
            f_meta.write('! Output from pycr8ascii.py v1.0' + os.linesep)
            f_meta.write('! '+ os.path.join(in_dir, in_file) + os.linesep + '!' + os.linesep)
            f_meta.write('! Global Attributes' + os.linesep)
            #for key in self.ga_d:
            for key in sorted_ga:
                if key in self.ga_d:
                    # Note: removing leading/trailing whitespace
                    print(key, '=', self.ga_d[key].strip(), sep='', file=f_meta)

            f_meta.write('!' + os.linesep)

            # Try to maintain order from DATA_VARIABLES global attribute
            for data_var in self.ga_d['DATA_VARIABLES'].split(';'):

                d_attr = self.sdss_d[data_var]

                # Write variable attributes
                f_meta.write('! Variable Attributes' + os.linesep)
                sorted_va = []
                unsorted_va = []

                for i in attr_arr_data:
                    if i in d_attr.va_d.keys():
                        sorted_va.append(i)
                    else:
                        unsorted_va.append(i)

                # Check for variable attributes not in attr_arr_data
                for i in d_attr.va_d.keys():
                    if i not in sorted_va:
                        sorted_va.append(i)

                sorted_va += unsorted_va

                for k in sorted_va:
                    if k not in d_attr.va_d.keys():
                        sorted_va.remove(k)

                # for i_txt in [str(i) + '=' + str(d_attr.va_d[i]) + os.linesep for i in
                #               d_attr.va_d.keys() if i.lower() not in ncsa_exclude]:

                # Note: removing leading/trailing whitespace
                for i_txt in [str(i) + '=' + str(d_attr.va_d[i]).strip() + os.linesep for i in
                              sorted_va if i.lower() not in ncsa_exclude]:
                    f_meta.write(i_txt)
                f_meta.write('!' + os.linesep)

            f_meta.write('! End of output file created by pycr8ascii.py' + os.linesep)
        #f_meta.close()

        print(os.path.basename(out_f1), "created!")

        with open(out_f2, 'w') as f_data:
            # Write data file
            # Loop over data variables
            # Maintain order from DATA_VARIABLES global attribute
            for data_var in self.ga_d['DATA_VARIABLES'].split(';'):

                var = self.sdss_d[data_var]


                var_name = var.va_d['VAR_NAME']
                var_type = var.va_d['VAR_DATA_TYPE']

                if "bytes" in type(var_name).__name__ :
                    f_data.write(var_name.decode('utf-8)' + os.linesep))
                else:
                    f_data.write(var_name + os.linesep)

                # Set number formatting similar to IDL code
                if var_name in ('DATETIME', 'DATETIME.START', 'DATETIME.STOP'):
                    if var_type == 'DOUBLE':
                        fmt = "{:14.9f}"
                    elif var_type == 'REAL':
                        fmt = "{:11.6f}"
                else:
                    if var_type in ('BYTE', 'SHORT', 'INTEGER'):
                        fmt = "{:d}"
                    else:
                        fmt = "{:12.6E}"


                # if type is bytes, interpret as 1 element of string
                if "bytes" in type(var.data).__name__:
                    f_data.write(var.data.decode('utf-8') + os.linesep)
                    # If reading h5 files with netCDF4 library ('str' datatype)
                elif type(var.data).__name__ == 'str':
                    # Make sure lineshift is included
                    if var.data[-1] != '\n':
                        f_data.write(var.data + '\n')
                    else:
                        f_data.write(var.data)
                # Deal with strings that were read as array of bytes.
                elif type(var.data).__name__ == 'ndarray' and var_type == 'STRING':
                    fmt = "{:s}"
                    for i in var.data.flatten():
                        f_data.write(fmt.format(i) + os.linesep)

                else:
                    try:
                        # Quick fix multidimensional data :TODO:
                        for i in var.data.flatten(): # Write in single column
                            if "bytes" in type(i).__name__:
                                f_data.write(i.decode('utf-8') + os.linesep)
                                # This is for writing non-standard types like 'int8'
                            elif 'int' in var.data.dtype.type.__name__:
                                fmt = "{:d}"
                                f_data.write(fmt.format(i) + os.linesep)
                            elif 'str_' in var.data.dtype.type.__name__:
                                fmt = "{:s}"
                                f_data.write(fmt.format(i) + os.linesep)
                            else:
                                f_data.write(fmt.format(i) + os.linesep)

                    except TypeError: # var.data is a scalar
                        if type(var.data) is bytes:
                            f_data.write(var.data.decode('utf-8'))
                        else:
                            f_data.write(fmt.format(var.data) + os.linesep)


        print(os.path.basename(out_f2), "created!")


class SDS:
    """
    Structure to hold variable attribute labels/values and data.

    """

    def __init__(self):
        """
        Define initial variables

        """

        self.va_l = []
        self.va_v = []
        self.data = None


def where_eq(inlist, inel):
    """
    Similar to the IDL 'where' function
    Return the list indices of inlist that matches inel, and the number of indices found

    """

    si = np.where(np.array(inlist) == inel)[0][0]
    #si = np.where(np.array(inlist) == inel)[0]
    if "arr" in type(si).__name__:
        scnt = len(si)
    else:
        scnt = 0

    return si, scnt


def stop_with_error_a(txt1, txt2):
    """
     Procedure called when an error in the program is detected. An error message is displayed
     and the program stopped and reset. If necessary, open files are closed. The error message
     is displayed in one or more of the following: a Pop-up dialog window; the Pop-up logging
     window; the Output Logging window in the IDLDE.
      ----------
     Written by Ian Boyd for the EVDC/AVDC - iboyd@bryanscientific.org

       History:
         20050729: Original IDLCR8ASCII Routine - Version 1.0
         20061004: Set-up so that the error output is displayed in the output window dependent on the
                   method that IDLCR8ASCII is called. If txt1 is preceeded by 'D_' or is null, the
                   error output is to a Pop-up Dialog window. Otherwise output will be to a logging
                   window. Renamed from STOP_WITH_ERROR to STOP_WITH_ERROR_A to avoid a name conflict
                   with the equivalent procedure in IDLCR8HDF. Common variable definition WIDGET_WIN_A
                   added - Version 2.0
         20080302: Added code which sends the error message to the IDLDE output window and/or an external
                   file (as determined by the dux array values) - Version 3.0
         20101122: Allow routine to return to the calling program, rather than stop the application, if
                   a 'reterr' argument is included in the call to idlcr8ascii - Version 4.0b1


       Inputs: txt1 - the initial text line of the error message. Generally contains the name of the routine
                      where the error was detected.
               txt2 - the second text line which generally describes the error state.

       Output: Nil

       Called by: The routine in which the error was detected. The following routines call STOP_WITH_ERROR:
                  READ_HDF_SDS; IDLCR8ASCII

       Subroutines Called: IDLCR8ASCII_EVENT (via XMANAGER)

    """

    print("stop_with_error_a called")
    print(txt1)
    print(txt2)


def geoms_data_type(data_type):
    """
    From input data_type, return the corresponding GEOMS datatype.
    Returns "UNDEFINED" in case of invalid GEOMS type

    """

    if data_type in ("byte", "uint8"):
        ret_val = "BYTE"
    elif data_type == "int8":
        ret_val = "INT8" # Not a valid GEOMS type, will be converted if possible
    elif data_type == "int16":
        ret_val = "SHORT"
    elif data_type == "int32":
        ret_val = "INTEGER"
    elif data_type in ("float32", "float"):
        ret_val = "REAL"
    elif data_type == "float64":
        ret_val = "DOUBLE"
    # This might not work for all h5py library versions
    elif data_type == h5py.h5t.string_dtype():
        ret_val = "STRING"
    # For H5 variable-length strings
    elif data_type == "bytes_":
        ret_val = "STRING"
    else:
        ret_val = "UNDEFINED"


    return ret_val


def check_var_data_type(f_type, va_type, sds_name, var_name):
    """
    Check that VAR_DATA_TYPE corresponds to the dataset data type

    """

    av_dt=['BYTE','SHORT','INTEGER','REAL','DOUBLE', 'STRING']

    # First check if the f_type-string is of type byte
    if type(va_type).__name__ == 'bytes_':
        va_type = va_type.decode()

    # Convert variable type to geoms type
    if f_type == str:
        var_data_type= "STRING"
    else:
        var_data_type = geoms_data_type(f_type.type.__name__)


    if var_data_type in ("INT8", "UNDEFINED"):
#        dt_err = var_name + ' ' + 'VAR_DATA_TYPE' + '=' + f_type + ' is not a valid GEOMS data type'
        dt_err = var_name + ' ' + 'VAR_DATA_TYPE' + '=' + f_type.type.__name__ + \
            ' is not a valid GEOMS data type'
        infotxt_output_a([2],[dt_err])

    elif va_type not in av_dt:
#        dt_err = var_name + ' ' + 'VAR_DATA_TYPE' + '=' + va_type.type.__name__ + ' is not a valid GEOMS data type'
        dt_err = var_name + ' ' + 'VAR_DATA_TYPE' + '=' + va_type + ' is not a valid GEOMS data type'
        infotxt_output_a([2],[dt_err])

    elif var_data_type != va_type:
        # import pdb; pdb.set_trace()
        dt_err = var_name + ' Dataset value(s) are ' + var_data_type + ' but VAR_DATA_TYPE ' + \
            'is ' + va_type
        infotxt_output_a([2],[dt_err])

    return var_data_type


def check_var_valid_fill(f_type, va_value, sds_name, var_name, var_attr):
    """
    Check that VAR_VALID_MIN, VAR_VALID_MAX and VAR_FILL_VALUE match the dataset data type.

    """

    av_dt=['BYTE','SHORT','INTEGER','REAL','DOUBLE', 'STRING']

    # Convert variable type to geoms type
    #    if f_type == str:
    # Include hdf4 syntax
    if f_type == str or f_type.type.__name__ == 'str_':
        var_data_type= "STRING"
    else:
        var_data_type = geoms_data_type(f_type.type.__name__)

    # Convert variable value type to geoms type
    if type(va_value) == str or type(va_value).__name__ == 'str_':
        value_data_type = "STRING"
    else:
        value_data_type = geoms_data_type(type(va_value).__name__)

    if var_data_type == "UNDEFINED":
#        dt_err = var_name + ' ' + 'VAR_DATA_TYPE' + '=' + f_type + ' is not a valid GEOMS data type'
        dt_err = var_name + ' ' + 'VAR_DATA_TYPE' + '=' + f_type.type.__name__ + \
            ' is not a valid GEOMS data type'
        infotxt_output_a([2],[dt_err])
    elif value_data_type not in av_dt:
        dt_err = var_name + ' ' + var_attr + '=' + value_data_type + ' is not a valid GEOMS data type'
        infotxt_output_a([2],[dt_err])

    elif var_data_type != value_data_type:
        dt_err = var_name + ' Dataset value(s) are ' + var_data_type + ' but VAR_DATA_TYPE ' + \
            'is ' + value_data_type


        infotxt_output_a([2],[dt_err])


def infotxt_output_a(in0, in1):
    """
    Procedure called to report information relevant to the reading of the HDF file.


     Inputs: in0 - Integer array containing input used to make correct text output
             in1 - String array containing input used to make correct text output

    """

    global global_log_file_name

    # Convert input to str
    in1 = [str(i) for i in in1]


    if in0[0] ==  0:
        infotxt=['']*3

        if in1[0] == 0:
            infotxt[0]='  INFORMATION: Global Attribute DATA_VARIABLES not found or has no values.'
            infotxt[1]='    Dataset listing order matches the order that the dataset is read in the GEOMS file.'
        else:
            infotxt[0]='  INFORMATION: Number of Global Attribute DATA_VARIABLES values ('+in1[0]+')'
            infotxt[1]='    does not match the number of datasets saved to the GEOMS file ('+in1[1]+')'

        infotxt[2]='    Number of datasets determined from '+in1[2]+' call'

    elif in0[0] == 1:
        infotxt='  INFORMATION: '+in1[0]+' has '+in1[1]+' data dimensions'
    elif in0[0] == 2:
        infotxt='  INFORMATION: '+in1[0]
    elif in0[0] == 3:
        infotxt=['']*4
        if (str(in1[1]).upper().find(str(in1[0]).upper()) != -1) and (in0[1] == 4) and (in0[2] == 0):
            infotxt[0]='  INFORMATION: Dataset Name is truncated (File created with HDF4.2r1 library or earlier):'
        else:
            infotxt[0]='  INFORMATION: Dataset Name does not match VAR_NAME value:'
        infotxt[1]='    SDS_NAME: '+in1[0]
        infotxt[2]='    VAR_NAME: '+in1[1]
        if in0[2] == 0:
            infotxt[3]='    Output uses VAR_NAME from Metadata'
        else:
            infotxt[3]='    Output uses Dataset Name from the HDF file'
    elif in0[0] == 4:
        infotxt=['']*2
        if in0[1] == 0:
            infotxt[0]='  INFORMATION: Metadata label VAR_NAME not found for dataset:'
            infotxt[1]='    SDS_NAME: '+in1[0]
        elif (in0[2] == 0) and (in0[3] != 0):
            infotxt[0]='  INFORMATION: VAR_NAME value does not match any DATA_VARIABLES values:'
            infotxt[1]='    VAR_NAME: '+in1[1]
    elif in0[0] == 5:
        infotxt=['']*2
        infotxt[0]='  INFORMATION: No Variable Attributes returned after call to '+in1[1]
        infotxt[1]='    SDS_NAME: '+in1[0]
    elif in0[0] == 6:
        infotxt='  INFORMATION: Metadata Dataset listing order may not match the order of'
        infotxt=infotxt+' the Global Attribute DATA_VARIABLES Values'
    elif in0[0] == 7:
        infotxt='  INFORMATION: Can read one HDF or netCDF file into session memory.'
        infotxt=infotxt+' Only first file in the array will be read'
    elif in0[0] == 8:
        infotxt='  INFORMATION: '+in1[0]+' is a coordinate variable'
    elif in0[0] == 9:
        infotxt=['']*3
        infotxt[0]='  INFORMATION: Calibrated data in dataset '+in1[0]
        infotxt[1]='    has been converted back to its original form using the formula:'
        infotxt[2]='    Orig_Data = Scale_Factor * (Cal_Data - Offset)'
    elif in0[0] == 10:
        infotxt=['']*2
        infotxt[0]='  INFORMATION: VAR_DATA_TYPE='+in1[1]+' for Calibrated data in dataset '+in1[0]
        infotxt[1]='    has been changed to the original datatype of '+in1[2]
    elif in0[0] == 11:
        infotxt=['']*2
        if in0[1] == 20:
            infotxt[0]='  INFORMATION: Order of multi-dimensional datasets changed to match the programming'
            infotxt[0]=infotxt[0]+' language convention:'
            infotxt[1]='    IDL/Fortran - fastest changing dimension first (datasets transposed when'
            infotxt[1]=infotxt[1]+' reading/writing to the HDF file)'
        elif np.mod(in0[1],10) == 0:
            infotxt[0]='  INFORMATION: Listing order of multi-dimensional VAR_DEPEND and VAR_SIZE values'
            infotxt[1]='    in the HDF file changed to match corresponding dataset dimension ordering'
        else:
            infotxt[0]='  INFORMATION: Unable to determine the correct listing order of multi-dimensional'
            infotxt[1]='    VAR_SIZE and VAR_DEPEND values in the HDF file'
    elif in0[0] == 12:
        infotxt=['']*2
        infotxt[0]='  INFORMATION: Table Attribute Values file must be in the same directory as the input file'
        infotxt[1]='    if wanting to convert from one Data Format to another'
    elif in0[0] == 13:
        infotxt=['']*2
        infotxt[0]='  INFORMATION: IDLcr8HDF called but it is not in the IDL Search Path or it encountered an'
        infotxt[1]='    unexpected error. If the latter please contact Ian Boyd at iboyd@bryanscientific.org'
    elif in0[0] == 14:
        infotxt=['']*2
        infotxt[0]='  INFORMATION: /POPUP keyword cannot be used together with the "reterr" argument.'
        infotxt[1]='    The request for a POPUP window has been ignored'
    elif in0[0] == 15:
        infotxt=['']*2
        infotxt[0]='  INFORMATION: Argument \'reterr\' must be of type string.' # a returnable variable of type string.
        infotxt[1]='    idlcr8ascii will stop normally if an error is encountered'
    elif in0[0] == 16:
        infotxt=['']*2
        infotxt[0]='  INFORMATION: /LOG, /FORMAT, and /DUMP keywords cannot be used in IDL DEMO mode and will be ignored.'
        infotxt[1]='    To generate ASCII files please use the free IDL Virtual Machine or a licenced version of IDL'
    elif in0[0] == 17:
        infotxt=['']*2
        infotxt[0]='  INFORMATION: NetCDF file create feature is disabled in IDL DEMO mode.'
        infotxt[1]='    To generate files please use the free IDL Virtual Machine or a licenced version of IDL'
    elif in0[0] == 18:
        infotxt='  INFORMATION: Value for Attribute Label '+in1[0]+' is an invalid Data Type'
    elif in0[0] == 19:
        infotxt='  INFORMATION: '+in1[0]+' is not a valid GEOMS Metadata Attribute'
    elif in0[0] == 20:
        infotxt='  INFORMATION: '+in1[0]+' entry must be written to the file as a STRING for '+in1[1]
    elif in0[0] == 21:
        infotxt='  INFORMATION: '+in1[0]+' sub-values must be separated by '';'' for '+in1[1]
    elif in0[0] == 22:
        infotxt='  INFORMATION: Spaces not permitted in the '+in1[0]+' entry for '+in1[1]
    elif in0[0] == 23:
        infotxt=['']*2
        itxt1='  WARNING: This program can\'t check that BYTE datasets and attributes contain only'
        itxt2=' 8-bit unsigned values (0-255) in netCDF3 files.'
        infotxt[0]=itxt1+itxt2
        itxt1='    Please check that the '+in1[0]+' dataset and attributes'
        itxt2=' with BYTE data type do not contain negative 8-bit signed values'
        infotxt[1]=itxt1+itxt2
    elif in0[0] == 24:
        infotxt=['']*2
        infotxt[0]='  INFORMATION: Input file name does not match the Global Attribute entry'
        infotxt[1]='    FILE_NAME=' + str(in1[0])
    elif in0[0] == 25:
        infotxt = '  INFORMATION: Variable attribute ' + in1[0] + ' is missing for variable ' + \
            in1[1]

    # infotxt is list or string for now
    if type(infotxt).__name__ == 'str':
        print(infotxt)
    else:
        print(*infotxt, sep='\n')

    if global_log_file_name is not None:
        with open(global_log_file_name, 'a') as f:
            if type(infotxt).__name__ == 'str':
                f.write(infotxt + '\n')
            else:
                i_txt = '\n'.join(infotxt)
                f.write(i_txt + '\n')


def nc_dimension_chk(ftype, sd_id, di):
    """

    This function checks whether each dataset in a HDF or netCDF file is a netCDF dimension name or not.
    It returns the index values of non-dimension dataset names and the number of dimension datasets.
    """

    dimen_names = ['constant', 'independent_', '_strlen']
    n_dn = len(dimen_names)
    n_sds = sd_id.nmembers if ftype == 'H5' else sd_id.nvars

    ds_names = [None] * n_sds
    gdi = [None] * n_sds
    for i in range(n_sds):
        if ftype == 'H5':
            sds_name = sd_id.get_member_name('/', di[i])
        else:  # netCDF file
            varstruct = sd_id.inqVar(i)
            sds_name = varstruct['name']
        ds_names[i] = sds_name.strip().lower()

    for i in range(n_sds):
        dnicnt = 0
        j = 0
        while dnicnt == 0 and j <= n_dn - 1:
            if dimen_names[j] in ds_names[i]:
                dnicnt += 1
            else:
                j += 1
        if dnicnt != 0:
            # Check that names are actually netCDF dimension names, otherwise treat as a dataset name
            if j == 0 and ds_names[i] != dimen_names[j]:  # i.e. not a netCDF dimension name
                dnicnt = 0
            elif j == 1:
                # Check that first part of the name is INDEPENDENT_ and the second part is a number
                if not ds_names[i].startswith(dimen_names[j]):
                    dnicnt = 0
                else:
                    res = ds_names[i].split('_')
                    res.append('')  # to ensure at least 2 extracted values in ds_names[i]
                    if not res[1].isdigit():
                        dnicnt = 0  # i.e. not a netCDF dimension name
            elif j == 2:
                # Check that first part of the name is a named dataset
                res = ds_names[i].split('_')
                ni = [k for k, name in enumerate(ds_names) if res[0] == name]
                if not ni:
                    dnicnt = 0  # i.e. not a netCDF dimension name

        if dnicnt == 0:
            gdi[i] = i  # i.e. this is a valid dataset name and not a netCDF dimension name

    dgi = [i for i, index in enumerate(gdi) if index is not None]
    n_sdsg = len(dgi)

    return dgi, n_sdsg


def test_dim_order(*args):
    """
    Not implemented yet

    """


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        prog="pycr8ascii.py",
        description="Read a GEOMS file in hdf5 or netCDF4 format and write .meta and .data files",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-f', '--file',
                        dest='file_name',
                        help="GEOMS file in netCDF or HDF format",
                        default=None)

    parser.add_argument('-o', '--outdir',
                        dest='out_dir',
                        help="Directory to write output. Same as input directory if not given.",
                        default=None)


    args = parser.parse_args()
    file_name = args.file_name
    out_dir = args.out_dir


    # Check input
    if len(sys.argv) == 1:
        parser.print_help()
        parser.exit()

    if file_name is None:
        print("File_name is required input (-f file_name)")
        parser.exit()

    if out_dir is None:
        # If out_dir is not given, set to input directory
        out_dir = os.path.dirname(os.path.abspath(file_name))

    # Initialize object, read, and write metadata + data
    geoms_obj = Geoms(file_name, out_dir)
    geoms_obj.get_data()
    geoms_obj.output_data()
