#!/usr/bin/env python3

"""

pycr8hdf.py

DESCRIPTION
    Create a GEOMS file in hdf5 format from .meta and .data files

REQUIREMENTS
    Tested with Python versions 3.7 & 3.10 under Ubuntu Linux. 

EXAMPLE USAGE
    pycr8hdf.py -m metadatafile.meta -d datafile.data -t tavfile.tav
    
CONTACT 
    eso@nilu.no (Espen Sollum)

"""

import os
import sys
import re
import argparse
import copy
import pdb
from pathlib import Path
from collections import OrderedDict
import datetime
import numpy as np


# log file is a global variable for now
global_log_file_name = None

# qa_yes is used as a global variable
global_qa_yes = False

# In case of errors detected in infotxt_output, this variable will tell the main
# program not to generate hdf file
global_o34_idl = '0'


class Tabledata():
    """
    Common block variables in the original IDL code

    """
    def __init__(self):
        self.tab_arr = None
        self.tab_ver = None
        self.tab_type = None


class Metadata():
    """
    Common block variables in the original IDL code

    """

    def __init__(self):
        self.meta_arr = None
        self.attr_arr_glob = None
        self.attr_arr_glob_prov =['DATA_TEMPLATE', 'DATA_QUALITY']
        self.attr_arr_data = None
        self.attr_free = None
        self.mv_lng = None
        self.mv_dbl = None
        self.mv_str = None


class Data():
    """
    Common block variables in the original IDL code

    """

    def __init__(self):
        """

        """

        self.ds = None
        self.ndm = None
        self.nvn = None
        self.vn = None
        self.vu = None
        self.vnchange = None
        self.vserror = None
        self.qa_yes = None
        self.rerr = ['NA', 'NA'] # initialize return error string

        # variable attributes and data
        self.va_l = []
        self.va_v = []
        self.data = OrderedDict()
        self.inf = -1

        self.va_d = {}
        self.sdss_d = {}

        self.data_read = False


    def extract_data(self, vc, dtest, infowrite):
        """

        Note: dtest is not being used for output of the data values, as in the IDL code

        """

        # Possible error messages for this procedure
        errtxt=[]
        errtxt.append('EOF encountered but not expected')
        errtxt.append(' when trying to find VAR_NAME in ')
        errtxt.append(' when trying to determine VAR_SIZE for VAR_NAME=')
        errtxt.append(' - Number of data lines should be ')
        errtxt.append('Invalid data point found in the dataset at line ')
        errtxt.append('Two instances of this VAR_NAME present in the Data File')
        errtxt.append('Too many data values present in the dataset or invalid data line entry following the dataset')
        errtxt.append('Number of data values does not match product of VAR_SIZE: ')
        errtxt.append('ISO8601 format not valid: ')
        errtxt.append('Unable to determine correct number of data values due to VAR_SIZE calculation error')
        proname='extract_data procedure: '


        dtest_out = -1

        # initialize data line count
        ndl=1

        vsi, vcnt = where_eq(self.vserror, self.vn[vc]) # Check whether there is a VAR_SIZE calculation error for this dataset
        sdata=self.ndm[vc, 8] == 6         # Check for String data type


        if self.inf == 1:
        # Read data file
            if not self.data_read:
                try:
                    with open(self.datafile, 'r') as f:
                        self.raw_data = [s.rstrip("\n\r") for s in f.readlines()]
                        # Strip for comments
                        self.raw_data = [i for i in self.raw_data if i[0:1] not in ('!', '#')]

                except Exception as e:
                    print(e)
                    # Exit if problem with reading the file
                    sys.exit()


                self.data_read = True         # Only read file once

                # Find indices to all variable names
                # Put in a dictionary for look up, in case the metadata order differs from the
                # order in the data file
                var_name_indices = []
                var_name_index_dict = {}

                for v in self.vn:
                    # Check if same VAR_NAME found twice in the data file,
                    # or not found
                    num_var_found = len([i for i, x in enumerate(self.raw_data) if x == v])
                    if num_var_found > 1:
                        stop_with_error(proname+data.vn[vc] + ': ', errtxt[5])
                    elif num_var_found ==0:
                        stop_with_error(proname + v, errtxt[0]+errtxt[1]+ self.datafile)

                    var_name_indices.append(self.raw_data.index(v))
                    var_name_index_dict.update({v : self.raw_data.index(v)})

                var_name_indices.sort()

                # For each variable, store start/end indices to the data
                var_name_index_range = {}


                # Set start indices
                for k, v in var_name_index_dict.items():
                    # variable start index
                    s_idx = var_name_indices.index(v)
                    # get index value from the sorted list
                    l_idx = var_name_indices[s_idx]

                    # Set end indices
                    if s_idx + 1 == len(var_name_indices): # last variable in file
                        e_idx = len(self.raw_data)
                    else:
                        e_idx = var_name_indices[s_idx + 1]

                    var_name_index_range.update({k: (l_idx, e_idx)})


                # Populate data dict
                for k, v in var_name_index_range.items():
                    l_idx, e_idx = var_name_index_range[k]
                    self.data.update({k: self.raw_data[l_idx + 1: e_idx]})


            dum=''
            vntest=self.vn[vc]
            # Add check in case VAR_NAME has been changed from original input
            if self.vnchange[vc] != '':
                vntest=self.vnchange[vc]

            # Read through data file until the correct Variable Name is found
            # vn_index = None
            try:
                vn_index = [i.upper() for i in self.raw_data].index(vntest.upper())
            except ValueError as ve: # vntest not found among variable names
                try:
                    vn_index = [i.upper() for i in self.raw_data].index(self.vn[vc].upper())
                except ValueError as ve: # vn[vc] not found among variable names
                    stop_with_error(proname + self.vn[vc], errtxt[0]+errtxt[1])



            self.ndl = len(self.data[self.vn[vc]])

            if dtest[0] == '-1': # need to determine the total number of data lines only (ndl)
                # self.ndl = len(self.data[self.vn[vc]])
                pass
            else:
                dtest_out = self.data[self.vn[vc]]


        elif self.inf == 0: # data is a structure
            if dtest[0] == '-1': # Determine the total number of data lines only
                self.ndl = len(self.geoms.sdss_d[self.vn[vc]].data)
                dtest_out = ['-1']
            else:
                self.ndl = len(self.geoms.sdss_d[self.vn[vc]].data)
                dtest_out = self.geoms.sdss_d[self.vn[vc]].data



        return dtest_out


    def set_global_log_file_name(self, file_name):
        """
        Set log file name (for running QA)

        """

        global global_log_file_name

        global_log_file_name = file_name


    def meta_data_to_geoms(self, md):
        """
        Store variable metadata and corresponding data in a structure.

        """

        natts = len(md.attr_arr_data)

        ALLOWABLE_DATA_TYPES=['BYTE','SHORT','INTEGER','LONG64','REAL','DOUBLE','STRING']

        for i in range(self.nvn):
            sds = SDS()
            sds.va_d = OrderedDict()
            mi, _ = where_eq(md.meta_arr, md.attr_arr_data[0] + '=' + self.vn[i])
            ninc=0
            j=0
            while j <= natts-1:
                ma=mi[0]+ninc
                # separate out the Meta entry into two components
                # res[0]: attribute name; res[1]: attribute value
                res=md.meta_arr[ma].split('=', maxsplit=1)
                while res[0] != md.attr_arr_data[j]: # in case optional variable attributes are not included
                    j=j+1

                sds.va_l.append(res[0])
                sds.va_v.append(' ' if len(res[1]) == 0 else res[1])

                ninc=ninc+1
                j=j+1

            # Copy the variable attribute labels and values to a dictionary
            for k, v in zip(sds.va_l, sds.va_v):
                sds.va_d.update({k:v})

            # Get the data
            sds.data = self.data[self.vn[i]]

            self.sdss_d.update({self.vn[i] : sds})


class SDS():
    """
    Structure to hold variable attribute labels/values and data

    """

    def __init__(self):
        """
        Define initial variables

        """

        self.va_l = []
        self.va_v = []
        self.data = None


def is_number(s):
    """
    Check that string contains a valid number

    """

    try:
        float(s)
        return True
    except ValueError:
        return False


def stop_with_error(txt1, txt2):
    """

    Procedure called when an error in the program is detected.

    Inputs: txt1 - the initial text line of the error message. Generally contains the name of the routine
                    where the error was detected.
            txt2 - the second text line which generally describes the error state.

    """

    err_txt = ('  ERROR in ' + txt1 + '\n    ' + txt2 + '\npycr8hdf stopped - Program Ended on {}'
               .format(datetime.datetime.utcnow()))

    if global_qa_yes:
        err_txt += '\n  INFORMATION: QA stopped before all checks could be completed'

    print(err_txt)

    if global_log_file_name is not None:
        with open(global_log_file_name, 'a') as f:
            f.write(err_txt)

    sys.exit()



def where_eq(inlist, inel):
    """
    Similar to IDL 'where'.
    Return the list indices of inlist that matches inel, and the number of indices found.

    """

    # If empty input list, return [-1], 0
    if len(inlist) == 0:
        return [-1], 0

    si = np.where(np.array(inlist) == inel)[0]
    scnt = len(si)

    return si, scnt



def infotxt_output(infotxt):
    """
    Report information to user.

    First character of infotxt will be changed as follows:

    0 = INFORMATION
    1 = NON-STANDARD COMPLIANCE NOTIFICATION (QA) o/w INFORMATION
    2 = ERROR (QA) o/w INFORMATION
    3 = ERROR
    4 = DEBUG

    """

    global global_o34_idl

    # If infotxt is a list, convert to string
    infotxt = ''.join([i for i in infotxt])

    if global_qa_yes:
        if infotxt[0] == '0':
            infotxt = '  INFORMATION:' + infotxt[1:]
        elif infotxt[0] == '1':
            infotxt = '  NON-STANDARD COMPLIANCE NOTIFICATION:' + infotxt[1:]
        elif infotxt[0] == '4':
            infotxt = '  DEBUG:' + infotxt[1:]
        else:
            infotxt ='  ERROR:' + infotxt[1:]

        # Remove everything after '|'
        infotxt = infotxt.split('|')[0]

    else: # program called in HDF file create mode
        if infotxt[0] == '3':
            infotxt='  ERROR:' + infotxt[1:]
            global_o34_idl = 'NOHDF'
        elif infotxt[0] == '4':
            infotxt='  DEBUG:' + infotxt[1:]
        else:
            infotxt='  INFORMATION:' + infotxt[1:]

        # Remove '|' from the message
        infotxt = infotxt.replace('|', '')


    # Assume infotxt is either list or string.
    if type(infotxt).__name__ == 'str':
        print(infotxt)
    else:
        print(*infotxt, sep='\n')

    if global_log_file_name is not None:
        with open(global_log_file_name, 'a') as f:
            if type(infotxt).__name__ == 'str':
                f.write(infotxt + '\n')
            else:
                i_txt = '\n'.join(infotxt)
                f.write(i_txt + '\n')


def var_units_test(data, vuvalue, rd, tab_type): #, var_si_conv, errstate):
    """
    Procedure to perform checks on the VAR_UNITS value in the metadata and, based on the VAR_UNITS input,
    calculate and return the VAR_SI_CONVERSION value.




    Inputs: vuvalue - a string containing the Metadata VAR_UNITS value to be checked (everything
                      after the '=')
                 rd - integer flag to indicate if VAR_DATA_TYPE is real/double (-1) or not (1). Used to
                      make VAR_SI_CONVERSION values of the same type
           tab_type - integer flag differentiating between AVDC (0) and original Envisat (1) styles

    Output: var_si_conv - a string containing the VAR_SI_CONVERSION value determined by the program
                       (returns '' if an error is encountered)
            errstate - string describing an error state encountered during testing (o/w set to '')


    """

    var_unit_arr=['%;0;0.01;1','A;0;1;A','C;0;1;s A','cd;0;1;cd','d;0;86400;s','deg;0;1.74533E-2;rad',
                  'degC;273.15;1;K','1;0;1;1','h;0;3600;s','Hz;0;1;s-1','J;0;1;m2 kg s-2','K;0;1;K',
                  'l;0;1E-3;m3','lm;0;1;cd sr','lx;0;1;cd sr m-2','m;0;1;m','min;0;60;s',
                  'MJD2K;0;86400;s','mol;0;1;mol','molec;0;1.66054E-24;mol','Np;0;1;1',
                  'N;0;1E3;m kg s-2','Pa;0;1;m-1 kg s-2','pH;0;1E-12;m2 kg s-2 A-2','photons;0;1;photons',
                  'ppbv;0;1E-9;1','ppmv;0;1E-6;1','pptv;0;1E-12;1','ppv;0;1;1','psu;0;1;psu',
                  'rad;0;1;rad','s;0;1;s','sr;0;1;sr','V;0;1;m2 kg s-3 A-1','W;0;1;m2 kg s-3','kg;0;1;kg',
                  'DU;0;4.4614E-4;mol m-2','Gal;0;1E-2;m s-2','dB;0;1;1']

    unit_pre_arr=['Y;yotta;1E24','Z;zetta;1E21','E;exa;1E18','P;peta;1E15','T;tera;1E12','G;giga;1E9',
                  'M;mega;1E6','k;kilo;1E3','h;hecto;1E2','da;deka;1E1','d;deci;1E-1','c;centi;1E-2',
                  'm;milli;1E-3','u;micro;1E-6','n;nano;1E-9','p;pico;1E-12','f;femto;1E-15',
                  'a;atto;1E-18','z;zepto;1E-21','y;yocto;1E-24']

    #Set up text in case an error is found in the input VAR_UNITS value
    errtxt=['']*2
    if tab_type == 1:
        errtxt[0]='No match with Table.Dat BASE_UNIT/UNIT_PREFIX values'
    else:
        errtxt[0]='No match with Table Attribute Values file VAR_UNITS/UNIT_PREFIX values'
    errtxt[1]='Not valid'
    errtxt.insert(0, 'VAR_UNITS='+vuvalue+': ')
    var_si_conv=''
    errstate=''  # initialize outputs

    nta=len(var_unit_arr)
    #extract var_unit_arr/unit_pre_arr sub-values into vu/up arrays
    vuhold=var_unit_arr[0].split(';') # test for number of sub-values
    nvu=len(vuhold)
    data.vu=np.zeros((nvu,nta), dtype=object)
    data.vu[:,:]='' # Array for strings

    for j in range(nta):
        vuhold=var_unit_arr[j].split(';')
        data.vu[:,j]=vuhold

    nta=len(unit_pre_arr)
    vuhold=unit_pre_arr[0].split(';')  # test for number of sub-values
    nup=len(vuhold)
    up=np.zeros((nup,nta), dtype=object) ; up[:,:]='' # Array for strings
    for j in range(nta):
        vuhold=unit_pre_arr[j].split(';')
        up[:,j]=vuhold

    # Separate out metadata sub-values into component parts, and set-up holding arrays
    vp=[v.strip() for v in vuvalue.split(' ')]
    vpn=len(vp)
    vpx=np.zeros((2,vpn), dtype=object) ; vpx[:,:]='' # 0 holds VAR_UNIT value, 1 holds POWER component
    bpx= ['']*vpn               # holding string array for base unit
    ex=[1]*vpn              # holding integer array for power value (defaults to 1)
    tm=np.zeros((vpn),dtype=np.float64)                # holding array for scale factor
    j=0

    
    while j <= vpn-1 and errstate == '':
        # test to see if the sub-value is a base unit in the TAV file
        ti, tcnt = where_eq(data.vu[0,:], vp[j])

        if tcnt != 0:
            vpx[0,j]=vp[j] # it is a base unit
        else:               # separate out into unit and power values as required
            vpx[0,j]=vp[j][0:1] # save first character of vp(j)
            stopchk=0
            if len(vp[j]) >= 2:
                for k in range(1, len(vp[j])):
                    ah=vp[j][k:k+1]
                    test1=ah.isalpha() # A-Z a-z

                    if test1:
                        if stopchk == 0:
                            vpx[0,j]=vpx[0,j]+ah
                        else:
                            stopchk=2
                            # if stopchk EQ 2 THEN this means that VAR_UNITS is not legal
                    elif stopchk == 0:
                        stopchk=1      # first non-alpha character so check for numeric or '-' character
                        test1=ah.isalnum() or ah == '-' and ah != '0' # - or 1-9
                        if not test1:
                            stopchk=2
                        else:
                            vpx[1,j]=ah
                    elif stopchk == 1:
                        # need to check for a numeric character only
                        test1=ah.isdecimal()
                        if not test1:
                            stopchk=2
                        else:
                            vpx[1,j]=vpx[1,j]+ah
                if vpx[1,j] != '':
                    ex[j]=int(vpx[1,j])

            if stopchk == 2:
                vpx[0,j]=vp[j] # in the event that the value is not valid, so will create error
                                # Do TAV check on the VAR_UNIT value
            ti, tcnt = where_eq(data.vu[0,:], vpx[0,j])


        if tcnt != 0:    # VAR_UNIT is a BASE_VALUE
            bemult=np.float64(data.vu[nvu-2,ti[0]])**ex[j]
            tm[j]=bemult
            bpx[j]=data.vu[nvu-3,ti[0]]+';'+ '{:.1E}'.format(bemult)  +';'+data.vu[nvu-1,ti[0]]
        elif vpx[0,j] == 'g':
             # check for VAR_UNIT EQ g (gram) for AVDC style TAV file
            bemult=np.float64(0.001)**ex[j]
            tm[j]=bemult
            bpx[j]='0;' + '{:.1E}'.format(bemult) + ';kg'
        else:                    # separate out vpx value into prefix and base-value and test
                                # check for valid prefix - first check for 'deka' (da)
            pref = vpx[0,j][0:2] ; bas=vpx[0,j][2:]
            pi, pcnt = where_eq(up[0,:], pref)
            if pcnt == 0: # test for the remaining prefixes
                pref = vpx[0,j][0:1] ; bas=vpx[0,j][1:]
                pi, pcnt = where_eq(up[0,:], pref )

            if pcnt != 0:
                pmult=np.float64(up[nup-1,pi[0]])
                # check for valid base
                ti, tcnt = where_eq(data.vu[0,:], bas)
                if tcnt != 0:
                    bpmult=(np.float64(data.vu[nvu-2,ti[0]])*pmult)**ex[j] ; tm[j]=bpmult
                    bpx[j]=data.vu[nvu-3,ti[0]]+';' + '{:.1E}'.format(bpmult) +';'+data.vu[nvu-1,ti[0]]
                elif bas == 'g': # check for VAR_UNIT EQ g (gram) for AVDC style TAV file
                    bpmult=pmult*np.float64(0.001)**ex[j] ; tm[j]=bpmult
                    bpx[j]='0;' + '{:.1E}'.format(bpmult) +';kg'
                else:
                    errstate=errtxt[0]
            else:
                errstate=errtxt[0]
        j=j+1


    if errstate == '':  # No errors detected so continue

        # Create VAR_SI_CONVERSION value
        tmult=1.0
        for j in range(vpn):
            tmult=tmult*tm[j]

        # reformat the multiplier e.g. 1.000E+003 becomes 1E3
        # convert to Exponential form if necessary
        if tmult == 273.15 or np.mod(tmult, 60) == 0 or ((tmult >= 0.01) and (tmult < 1.e2)
                                                         and (np.mod(tmult*1.e4, 1.e2) == 0.)):
            tmults = str(tmult).upper().strip()  # keep the same format
        else:
            tmults = '{:14.6E}'.format(tmult).strip()

        epos=-1
        if 'E' in tmults:
            epos=re.search('E', tmults).start()
        ppos=-1
        if '.' in tmults:
            ppos=re.search('\.', tmults).start()

        if epos != -1:  # remove unnecessary characters after 'E'
            ep=int(tmults[epos+1:])
            epx=str(ep).strip()
            tmults=tmults[:epos+1]+epx
        if ppos != -1:   # remove any trailing zeroes after the decimal place
            if epos != -1:
                ep=tmults[ppos+1:epos]
            else:
                ep=tmults[ppos+1:]

            while ep[len(ep)-1:len(ep)] == '0':
                ep=ep[0:len(ep)-1]

            if ep != '':
                tmults=tmults[:ppos+1]+ep
            else:
                tmults=tmults[:ppos]
            if epos != -1:
                tmults=tmults+'E'+epx

        # Scale the units by the power e.g. W2 = (m2 kg s-3)^2
        for j in range(vpn):
            vsc=bpx[j].split(';')
            vspl=vsc[2].split(' ') ; vscnt=len(vspl)
            sichk=['']*vscnt ; pwchk=['']*vscnt
            if (vpx[1,j] != '') and (vpx[1,j] != '1') and (vsc[2] != '1'):
                bpx[j]=vsc[0]+';'+vsc[1]+';'
                for k in range(vscnt):
                    # separate out SI units and power values
                    sires=re.split(r'-|\d+', vspl[k])
                    sires = [sire for sire in sires if sire]
                    sichk[k]=sires[0] # SI Unit

                    if len(sichk[k]) != len(vspl[k]):
                        pwchk[k]=vspl[k][len(sichk[k]):]
                    else:
                        pwchk[k]='1' # Power value

                    pwm=int(pwchk[k])*int(vpx[1,j])
                    if k == 0:
                        sp=''
                    else:
                        sp=' '
                    bpx[j]=bpx[j]+sp+sichk[k]+str(pwm)

        # Put together VAR_SI_CONVERSION
        vsc=bpx[0].split(';')
        var_si_conv=vsc[0]+';'+tmults+';'+vsc[2]

        # ESO dbg
        # import pdb; pdb.set_trace()
        #if 'mol' in vsc[2]:
        #    import pdb; pdb.set_trace()



        # Add remaining base units to VAR_SI_CONVERSION
        if vpn > 1:
            for j in range(1,vpn):
                vsc=bpx[j].split(';')
                var_si_conv=var_si_conv+' '+vsc[2]

        # Check VAR_SI_CONVERSION for repeated SI units e.g. m m-3 becomes m-2
        schk = re.split(' |;', var_si_conv)
        scnt = len(schk)
        if scnt > 3: # more than one SI unit in VAR_SI_CONVERSION
            sichk=['']*(scnt-2) ;  pwchk=['']*(scnt-2)
            for j in range(scnt-2): # separate out SI units and power values
                sires=re.split(r'-|\d+', schk[j+2])
                sires = [sire for sire in sires if sire]
                sichk[j]=sires[0] # SI Unit

                if len(sichk[j]) != len(schk[j+2]):
                    pwchk[j]=schk[j+2][len(sichk[j]):]
                else:
                    pwchk[j]='1'    # Power value

            j=0 ; pwval=0

            while j < scnt-3:
                si, sicnt = where_eq(sichk[j+1:scnt-2], sichk[j])
                si = [l for l in si if l != '']
                sicnt = len(si)

                if sicnt != 0:
                    for k in range(sicnt):
                        si[k]=si[k]+j+1
                        pwval=int(pwchk[j]) + int(pwchk[si[k]])
                        if int(str(pwval)[0]) == 0 and k == sicnt-1:
                            sichk[j]=''
                            pwchk[j]=''
                        elif int(str(pwval[0])) == 1 and k == sicnt-1:
                            pwchk[j]=''
                        else:
                            pwchk[j]=pwval[0]
                            # make null all the other SI values
                        sichk[si[k]]=''
                        pwchk[si[k]]=''
                elif pwchk[j] == '1':
                    pwchk[j]=''
                j=j+1

            # Also do check on the power value of the last SI Unit
            if pwchk[scnt-3] == '1':
                pwchk[scnt-3]=''

            # Put units in power value order
            pwhold=copy.deepcopy(pwchk)

            oi, ocnt=where_eq(pwhold, '')
            if ocnt != 0:
                for x in oi:
                    pwhold[x]='1'


            # Convert to np to allow argsort, astype
            pwhold = np.array(pwhold)
            sichk = np.array(sichk)
            pwchk = np.array(pwchk)
            pws = np.argsort(pwhold.astype(int))
            
            sichk = sichk[pws][::-1]
            pwchk = pwchk[pws][::-1]

            var_si_conv=schk[0]+';'+schk[1]+';'+sichk[0]+pwchk[0]
            si = [l for l in schk if l != '']
            sicnt = len(si)
            if sicnt == 0:
                var_si_conv=var_si_conv+'1'  # i.e. values cancelled out
            else:
                for j in range(1,scnt-2):
                    #si = [l for l in sichk[0:j-1] if l != '']
                    si = [l for l in sichk[0:j] if l != '']
                    sicnt = len(si)
                    if (sichk[j] == '') or (sicnt == 0):
                        var_si_conv=var_si_conv+sichk[j]+pwchk[j]
                    else:
                        var_si_conv=var_si_conv+' '+sichk[j]+pwchk[j]

        if rd < 0:
        # VAR_DATA_TYPE is Real or Double so make VAR_SI_CONVERSION values floating point
            tchkh=var_si_conv.split(';') ; tup=[st.upper() for st in tchkh]
            for j in range(2):
                if '.' not in tup[j]:
                    if 'E' not in tup[j]:
                        tchkh[j]=tchkh[j]+'.0'
                    else:
                        tchkh[j]=tup[j][0: re.search('E', tup[j]).start()] +'.0' + \
                        tup[j][re.search('E', tup[j]).start():]
#                            tup[j][re.search('E', tup[j]).end():]
            var_si_conv=tchkh[0]+';'+tchkh[1]+';'+tchkh[2]

        # Check for invalid VAR_UNITS - third VAR_SI_CONVERSION is 1 plus extra values
        vsc=var_si_conv.split(';')
        vsc[2]=vsc[2].strip()
        if (errstate == '') and (vsc[2][0] == '1') and (len(vsc[2]) > 1):
            errstate=errtxt[1]
            var_si_conv=''


    return var_si_conv, errstate


def geoms_rule_changes(code, *args):
    """
    Called for functions not yet implemented

    """

    print("geoms_rule_changes called with code", code, " (not implemented)")
    # print("geoms_rule_changes called with parameters ", code)


def geoms_rule_changes_0(meta_arr, mh_good):
    """
    Check for redundant attributes

    """

    redundant=['DATA_TYPE','DATA_LEVEL','VAR_MONOTONE','VAR_DIMENSION',
               'VAR_AVG_TYPE','VIS_LABEL','VIS_FORMAT','VIS_PLOT_TYPE',
               'VIS_SCALE_TYPE','VIS_SCALE_MIN','VIS_SCALE_MAX']

    for i_att in redundant:
        rdl = len(i_att)
        fa_i, fa_cnt = where_eq([i[0:rdl] for i in meta_arr], i_att)

        if fa_cnt != 0:
            for i in fa_i:
                mh_good[i] = False
            if i_att[0:3] in ('VAR', 'VIS'):
                att_type = 'Variable'
            else:
                att_type='Global'
            infotxt = '2 '+i_att+' not a GEOMS '+att_type+' attribute|.'
            infotxt += ' Removed from the metadata saved to the output file'
            infotxt_output(infotxt)


def geoms_rule_changes_1(vncnt, dv, data):
#def geoms_rule_changes_1(meta_data, data):
    """
    Check for redundant _START/STOP sub-values and replace with
    DATETIME.START/STOP/INTEGRATION.TIME (for single occurences only)

    """

    time_attr=['START','STOP','INTEGRATION','RESOLUTION']
    n_tim=len(time_attr)
    iscnt=0                 # Used for RESOLUTION/INTEGRATION.TIME checks
    for i in range(n_tim):
        si = [j for j, s in enumerate(data.vn) if '_'+time_attr[i]+'.TIME' in s]
        if len(si) > 0:
            pdb.set_trace()
            if iscnt == 0:  # change name and also relevant VAR_NAME if present
                if i == 2:
                    iscnt=1 # used in the event both INTEGRATION and RESOLUTION.TIME are present
            data.vnchange[si[0]]=data.vn[si[0]]
            if i > 2:
                data.vn[si[0]]=time_attr[2]+'.TIME'
            else:
                data.vn[si[0]]='DATETIME.'+time_attr[i]
            if data.qa_yes:
                itxt=' expected to be '
            else:
                itxt=' renamed '
            infotxt='2 Dataset name '+vnchange[si[0]]+itxt+in2[si[0]]
            infotxt_output(infotxt)
    # Check for double underscore and replace with single underscore
    for i, k in enumerate(data.vn):
        if '__' in k:
            data.vnchange[i]=data.vn[i]
            infotxt=[' ']*2
            infotxt[0]='2 Double underscore present in Variable Name '+data.vn[i]+'|'
            infotxt[1]='    replaced with a single underscore'
            infotxt_output(infotxt)
            data.vn[i].replace('__', '_')




def geoms_rule_changes_2(vdtest, vnv, vdv, holdvd):
    """
    Check for and Apply GEOMS rule changes for INDEPENDENT and axis variables
    RULE: Any variable that is mentioned in VAR_DEPENDS is by definition an axis

    """

    avi, avcnt = where_eq(vdtest, vnv)                     # i.e. is the VAR_NAME an axis variable


    if avcnt != 0:                     # possible axis variable so do axis variable checks
        test1=(vdv == 'INDEPENDENT') or (vdv == vnv) #value is independent or self-referencing
        test2=(vdv == 'INDEPENDENT') and (vdv != vnv) and (vdv != 'CONSTANT') # neither independent, self-referencing, nor constant
        test6=(vdv == 'CONSTANT') and (vnv == 'DATETIME') # i.e. VAR_NAME=DATETIME and VAR_DEPEND=CONSTANT
        if (test1) or (test6):
            if vdv != vnv:
                infotxt='2 ' + holdvd + ' should be self-referencing for axis variable VAR_NAME='+vnv+'|.'
                infotxt=infotxt+' VAR_DEPEND value changed in metadata'
                infotxt_output(infotxt)
                vdv=vnv        # self-referencing VAR_DEPEND value
            elif test2:
                holdvd[0]='E1'       # i.e. axis variable cannot be dependent on another variable so return error
    else:         # not an axis variable so ensure VAR_DEPEND value is not self-referencing
        if vdv == vnv:
            holdvd[0]='E3' # i.e. VAR_DEPEND value cannot be self-referencing

    return vdv, holdvd


def geoms_rule_changes_3(in1, in2, in3):
    """
    Change MJD2000 to MJD2K

    """
    in1='VAR_UNITS=MJD2K'
    in2='MJD2K'
    if in3 == 0:  # First instance of MJD2000 so include comment
        infotxt='2 VAR_UNITS=MJD2000 not GEOMS compliant|. Changed to VAR_UNITS=MJD2K'
        infotxt_output(infotxt)


def geoms_rule_changes_5(vncnt, dv, vuv, data_source, data):
    """
    Look for obsolete DATA_VARIABLES

    """
    
    # List of obsolete DATA_VARIABLE_01 values
    obs_list_1=['AIR.TEMPERATURE','TEMPERATURE.AIR','COLUMN.VERTICAL','H2O.RELATIVE.HUMIDITY',
                'PRESSURE.SURFACE','TEMPERATURE.SURFACE','TEMPERATURE.INTERNAL.BOX',
                'TEMPERATURE.INTERNAL.INSTRUMENT','TEMPERATURE.SEA.SURFACE','AIR.NUMBER.DENSITY',
                'SEA.SURFACE.TEMPERATURE','OPACITY.ATMOSPHERIC.VERTICAL']
    #Corresponding list of replacement DATA_VARIABLE_1 values (direct replacement)
    rep_list_1=['TEMPERATURE','TEMPERATURE','COLUMN','HUMIDITY.RELATIVE',
                'SURFACE.PRESSURE','SURFACE.TEMPERATURE','INTERNAL.BOX.TEMPERATURE',
                'INTERNAL.INSTRUMENT.TEMPERATURE','SEA.SURFACE.TEMPERATURE','NUMBER.DENSITY',
                'SEA.SURFACE.TEMPERATURE.SKIN','OPACITY.ATMOSPHERIC']
    #Obsolete DATA_VARIABLE_01 which need special handling
    obs_list_spec_1=['.AMF','.AVK','.CONCENTRATION','.MIXING.RATIO']
    #Obsolete DATA_VARIABLE_02 values
    obs_list_2=['SOLAR','VERTICAL.SOLAR',
                'EMISSION.VERTICAL','VERTICAL.EMISSION',
                'VERTICAL','SLANT','SOLAR.OCCULTATION','VERTICAL.ZENITH','VERTICAL.NADIR',
                'VERTICAL.LUNAR','VERTICAL.SOLAR.FOCUS']
    #[0,1] for FTIR/UVVIS (except Brewer and Dobson) only, [2,3] for MWR only
    #Corresponding list of replacement DATA_VARIABLE_2 values (direct replacement)
    rep_list_2=['ABSORPTION.SOLAR|SCATTER.SOLAR','ABSORPTION.SOLAR|SCATTER.SOLAR',
                'EMISSION','EMISSION',
                '','','OCCULTATION.SOLAR','ZENITH','NADIR',
                'ABSORPTION.LUNAR','ABSORPTION.SOLAR.FOCUS']
    #Replacements for .CONCENTRATION - dependent on VAR_UNITS
    rep_list_conc=['.MIXING.RATIO.VOLUME','.MIXING.RATIO.MASS','.NUMBER.DENSITY','.PARTIAL.PRESSURE']
    #Conditional Miscellaneous changes
    misc_list_1=['TEMPERATURE_SKIN','AIR.MASS.FACTOR']
    #Corresponding change to miscellaneous variables
    misc_list_2=['SEA.SURFACE.TEMPERATURE.SKIN','O3.COLUMN_AMF']
    #Obsolete DATA_VARIABLE_03 descriptors (may or may not be able to fix these)
    obs_list_3=['DU','UNCERTAINTY.STDEV','UNCERTAINTY.RMS','UNCERTAINTY.RELATIVE','UNCERTAINTY.TOTAL', 
                'UNCERTAINTY.RANDOM','UNCERTAINTY.SYSTEMATIC']
    #Corresponding list of replacement DATA_VARIABLE_03 descriptors
    rep_list_3=['','UNCERTAINTY.COMBINED.STANDARD','UNCERTAINTY.COMBINED.STANDARD', 
                'UNCERTAINTY.COMBINED.STANDARD.RELATIVE','UNCERTAINTY.COMBINED.STANDARD', 
                'UNCERTAINTY.RANDOM.STANDARD','UNCERTAINTY.SYSTEMATIC.STANDARD']
    rep_list_4=['','','','','UNCERTAINTY.COMBINED.COVARIANCE','UNCERTAINTY.RANDOM.COVARIANCE', 
                'UNCERTAINTY.SYSTEMATIC.COVARIANCE']

    in1 = len(dv)
    vn_v=np.ndarray((in1,3), dtype='object')
    vn_new=['']*in1
    # Separate out Variable Name/[Mode or Descriptor]/[Descriptor]

    for i, val in enumerate(dv):
        res=val.split('_')
        nres = min(3, len(res))
        if nres > 3:
            nres=3
        for j in range(nres):
            vn_v[i, j]=res[j]


    # Test for values in obs_list_1
    for i in range(len(obs_list_1)):
        # Change to exact match from v4.0b55 (20200709) - due to introduction of DRY.AIR.... variable names
        ri, rcnt = where_eq([k.upper() for k in vn_v[:,0]], obs_list_1[i])
        if rcnt != 0:# obsolete values found so replace
            vlen=len(obs_list_1[i])
            for j in range(rcnt):
                change_val=1
                # Special case for SEA.SURFACE.TEMPERATURE (only change if mode is _SKIN)
                if obs_list_1[i] == 'SEA.SURFACE.TEMPERATURE':
                    if vn_v[ri[j],1].upper() == 'SKIN':
                        vn_v[ri[j],1]=''
                    else:
                        change_val=0
                if change_val == 1:
                    vpos=vn_v[ri[j],0].upper().find(obs_list_1[i])
                    vn_v[ri[j],0] = vn_v[ri[j],0][0 : vpos] + rep_list_1[i] + vn_v[ri[j],0][vpos+vlen:]


    # TODO: implement the rest


def geoms_rule_changes_6(vncnt, dv, data_source):
    """

    Look for obsolete DATA_SOURCE

    """

    obs_list=['RO.CHAMP','RO.SAC.C','RO.F3C.FM','RO.CNOFS','RO.GRACE.A','FTIR_',
              'UVVIS.DOAS_','MICROWAVE.RADIOMETER','LIDAR.BACKSCATTER',
              'LIDAR.DIAL','LIDAR.OLEX','LIDAR.RIEGL','LIDAR.RMR']

    # Corresponding replacement list
    rep_list=['CHAMP','SAC.C','F3C.FM','CNOFS','GRACE.A','FTIR.','UVVIS.DOAS.','MWR.','LIDAR.']

    # Possible list of species for the above instruments (note: gases will take precedence
    # over Aerosol and Temperature, o/w the first species found will be considered the
    # primary species measured)
    spc_list=['BrO.','C2H2.','C2H6.','CCl2F2.','CCl3F.','CH4.','CHF2Cl.','CHOCHO.','ClO.','ClONO2.','CO.',
              'CO2.','COF2.','H2CO.','H2O.','HCFC22.','HCl.','HCN.','HCOOH.','HF.','HNO3.','HONO.','IO.',
              'N2O.','NO.','NO2.','O3.','OClO.','OCS.','SF6.','SO2.','AEROSOL','TEMPERATURE']


    oli=-1                   # will change to obs_list index value if obsolete DATA_SOURCE_01 found
    n_obsl=len(obs_list)
    lid_i=n_obsl-5    # needs to be start index of the lidar measurements in obs_list
    spc_i=n_obsl-8    # needs to be start index of the instruments requiring species information in obs_list
    uv_i=n_obsl-7     # needs to be UVVIS.DOAS index in obs_list

    for i, obs_l in enumerate(obs_list):
        if obs_l in data_source.upper():
            print("obsolete DATA_SOURCE value found")
            print("implement geoms_rule_changes_6")
            sys.exit()


def geoms_rule_changes_7(facnt, fav, data):
    """
    Look for obsolete CALVAL and NDSC FILE_ACCESS values

    """

    obs_list=['CALVAL','NDSC']
    rep_list=['EVDC','NDACC']
    
    for i in range(facnt-1):
        oi, ocnt = where_eq(obs_list, fav[i])
        if ocnt != 0:
            if data.qa_yes:
                itxt=' expected to be '
            else:
                itxt=' renamed '
            infotxt='2 FILE_ACCESS value ' + fav[i] + itxt + rep_list[oi[0]]
            infotxt_output(infotxt)
            fav[i]=rep_list[oi[0]]


def geoms_rule_changes_8(in1, in2, in3, qa_yes=False):

    # vucnt, vuv, vdv
    """

    Look for VAR_UNITS=DIMENSIONLESS or NONE and replace with '' or '1'

    """

    vuv=[i.upper() for i in in2]
    for i in range(in1): # len(in2)
        if (vuv[i] == 'DIMENSIONLESS') or (vuv[i] == 'NONE'):
            if in3[i] == 'STRING':
                newv=''
            else:
                newv='1'
            if qa_yes:
                itxt=' expected to be '
            else:
                itxt=' renamed '
            infotxt='2 VAR_UNITS='+vuv[i]+itxt+'VAR_UNITS='+newv+' based on VAR_DATA_TYPE='+in3[i]
            infotxt_output(infotxt)
            in2[i]=newv

    return in2


def geoms_rule_changes_9(m_v0, m_v1, ga_chk, md, td):
    """

    DATA_QUALITY/DATA_TEMPLATE checks

    Inputs:
    md, td - metadata, tabledata

    """

    attr_arr_glob_prov=['DATA_TEMPLATE', 'DATA_QUALITY']
    geoms_te=''
    meta_te=''
    std_txt=' Required Metadata Template: '

    # Check if DATA_TEMPLATE field is present in the TAV file
    try:
        dti = td.tab_arr[attr_arr_glob_prov[0]]
        dtcnt = len(dti)
    except KeyError as k_e:
        dtcnt = 0

    # Check if DATA_TEMPLATE label is present in the Metadata
    gi, gcnt = where_eq(m_v0, attr_arr_glob_prov[0])

    if gcnt == 1:
        if m_v1[gi[0]] != '':
            meta_te = m_v1[gi[0]].strip().upper()

    if dtcnt != 0:
        tab_arr_sub = dti # Valid Template values from TAV file


    if dtcnt == 0:
        infotxt='3'+std_txt+'DATA_TEMPLATE field is not present in the TAV file. Checks cannot be performed'
        infotxt_output(infotxt)
        std_txt=''
    elif meta_te != '':
        # check if value is present in the TAV file
        ci, ccnt = where_eq (tab_arr_sub, meta_te)
        if ccnt != 0:
            geoms_te=tab_arr_sub[ci[0]]
            m_v1[gi[0]] = geoms_te
            infotxt='0'+std_txt+geoms_te
            infotxt_output(infotxt)
            std_txt=''



def geoms_rule_changes_11(in1):
    """
    Check DATA_VERSION_NAME part of the DATA_SOURCE

    """

    test2=False
    test3=False     # will change to 1 if test conditions are true

    test1 = all([i.isalnum() for i in in1.split('.')])
    test2 = in1[0] == '.' or in1[-1] == '.'
    test3 = False
    if re.search(('\.\.'), in1) is not None:
        test3 = True

    if not test1 or test2 or test3:
        # Write error message
        infotxt='3 VERSION_NAME sub-value of the DATA_SOURCE must consist of dot-separated alpha-numeric words: '
        infotxt=infotxt+in1
        infotxt_output(infotxt)


def read_tablefile(table_file, meta_data, table_data):
    """
    Read TAV file

    """

    proname="read_tablefile"

    errtxt = ['']*4
    errtxt[0]='Table Attribute Values file version not identified'
    errtxt[1]='Envisat table.dat ASC2HDF program version not found'
    errtxt[2]='First Envisat table.dat field value not found'
    errtxt[3]='Table Attribute Values file version is invalid (should be ddRddd or ddRdddd): '
    for i in range(3):
        errtxt[i]=errtxt[i]+' with the search criteria used by this program'

    # Minimum TAV version
    MIN_FMV = 4

    # Envisat Field names to be changed to equivalent AVDC Field names
    enviname=['_NAME','_AFFILIATION','DATA_VARIABLES_00_00','BASE_UNIT']
    avdcname=['ORIGINATOR','AFFILIATION','DATA_VAR_ALL_00_00','VAR_UNITS']

    # Read file, remove newline characters
    print('Input TAV File: ' + os.path.abspath(table_file))
    try:
        # NB: without this encoding hack (ISO-8859-1) we sometimes get errors like
        # utf-8' codec can't decode byte 0xe5 in position 79354: invalid continuation byte.
        # Investigate furter.
        with open(table_file, 'r', encoding="ISO-8859-1") as f:
            tab_data = f.read().splitlines()

    except Exception as e:
        print(e)
        print("ERROR: TAV file not found (pycr8hdf.py):\n", table_file, "\nExiting.")
        sys.exit(0)

    # Write to log file
    if global_log_file_name is not None:
        try:
            with open(global_log_file_name, 'a') as f:
                f.write('Input TAV File: ' + os.path.abspath(table_file) + '\n')
        except Exception as e:
            print("ERROR: Can not write to log file:\n{}\nExiting.".format(global_log_file_name))
            print(e)
            sys.exit(0)


    # Remove spaces
    tab_data = [i.replace(' ','') for i in tab_data]

    # Uppercase
    dum_up = [i.upper() for i in tab_data]

    # Determine TAV file version
    tab_type = None
    for val in dum_up:
        if '!TABLE.DATVERSION' in val:
            tab_type=1
            n_title=4
            infotxt='0 EVDC original style table.dat'
            break
        if '!VERSION' in val:
            tab_type=0
            n_title=5
            infotxt='0 GEOMS compliant Table Attribute Values'
            break

    if tab_type is None:
        stop_with_error(proname, errtxt[0])

    table_data.tab_type = tab_type

    infotxt=infotxt + ' file input. ' + infotxt[2:2+n_title] + ' Reporting Guidelines apply\n'
    infotxt_output(infotxt)

    # Ensure Meta Version has a valid format
    # i. Check third character is an 'R'
    # ii. Check number of characters is 6 or 7
    # iii. Check dd and ddd(d) are numeric
    # iv. Check that version number is 03 or greater for AVDC file type
    # v. Issue warning and convert to ddRddd if format is ddRdddd

    tmp_meta = [v.split(' ') for v in dum_up]
    f_mv = None
    for v_md in tmp_meta:
        if 'VERSION' in v_md[0]:
            f_mv = v_md[0].split('VERSION')[-1]
            break

    if f_mv is None:         # VERSION not found in dum_up
        stop_with_error(proname, errtxt[0] + ': ' + dum_up)

    if f_mv[2:3] != 'R' or len(f_mv) < 6 or len(f_mv) > 7:
        stop_with_error(proname, errtxt[3] + f_mv)

    if int(f_mv[0:2]) < MIN_FMV and tab_type == 0:
        infotxt = ['3 Old version of the Table Attribute Values file used as input: ']
        infotxt.append(' TAV Version ' + f_mv)
        infotxt.append('    Please update from http://avdc.gsfc.nasa.gov/Tools')
        infotxt_output(infotxt)


    if tab_type == 1:
        # identify asc2hdf version in table.dat and read past the CHECK_ATTRIBUTE line
        print("envitest, not implemented, exiting")
        sys.exit()

    table_data.tab_ver = f_mv + ';pycr8hdf'

    # Remove comments
    tab_data = [i for i in tab_data if i[0:1] not in ('!', '#', '')]

    # Find labels
    tab_labels = [i for i in tab_data if i[0:1] not in ('!', '#', '=', '')]

    # Change name of Envisat field to AVDC equivalent
    for i_pos, env_name in enumerate(enviname):
        #enviname=['_NAME','_AFFILIATION','DATA_VARIABLES_00_00','BASE_UNIT']
        pos_env, res_ln = where_eq(tab_labels, env_name)
        if res_ln == 1:
            tab_labels[pos_env[0]] = avdcname[i_pos]

        if res_ln > 1:
            print("Error: duplicate label in TAV file?")
            sys.exit()

    # Place the table data in a dictionary with field/labels as keys

    # Get number of entries for each label
    n_fields = []
    for lab_name in tab_labels:
        field_pos, num_field = where_eq(tab_data, lab_name)
        n_fields.append(field_pos[0])

    n_fields.append(len(tab_data))

    tab_arr = {}
    for i, label in enumerate(tab_labels):
        lab_dat = tab_data[n_fields[i]+1 : n_fields[i+1]]
        # Strip '='
        lab_dat = [i.replace('=', '') for i in lab_dat if i[0:1] == '=']
        tab_arr.update({label: lab_dat})

    #tab_data = [i.replace('=', '') for i in tab_data if i[0:1] == '=']


    table_data.tab_arr = tab_arr


#def read_metadata(metafile, tabledata, md, data, inf, qa_yes=False):
def read_metadata(metafile, tabledata, md, data, inf):

    """
    1. Ensures uniform input (making ATTRIBUTE_NAME uppercase, deleting unwanted spaces etc).
    2. Checks that attribute length falls within allowable limits
    3. Checks for repeated, missing, new, obsolete attributes
    4. Update FILE_META_VERSION with TAV and software version (tab_ver)

    Please also note the following:
    1. Any extra valid Global Attributes will be included (and added to the attr_arr_glob array).
    2. If redundant attributes are detected they will be omitted from the Metadata written
       to the HDF file.

    Inputs: metafile - the filename of the input file containing the Metadata, or
                       the metadata stored as a list
            tabledata - tabledata
            md      - metadata object
            data    - data object
            inf - flag, where 0=Metafile is a list; 1=Metafile is a file.
            qa_yes   - True if running qa


        Possible Conditions for STOP_WITH_ERROR call (plus [line number] where called):
          1. Metadata value contains too many characters
          2. No or invalid DATA_VARIABLES and VAR_NAME attributes present in the metadata
          3. DATA_VARIABLES or VAR_NAME value appears more than once
          4. No valid Metadata in command line parameter file or arrays
          5. VAR_NAME does not match corresponding DATA_VARIABLES value
          6. Number of DATA_VARIABLES values does not match the number of VAR_NAME occurrences
          7. A required Global Attribute Label is not present in the Metadata
          8. A required Global Attribute Label is repeated in the Metadata
          9. Number of Attribute occurrences does not match the number of datasets
         10. DATA_QUALITY is not present but is a mandatory attribute when DATA_TEMPLATE is reported

        Information Conditions (when the program is able to make changes):
          1. Invalid Attribute Label ignored
          2. Missing Global Attribute added
          3. DATA_VARIABLES attribute added to metadata based on VAR_NAME values
          4. Values for DATA_VARIABLES appended to the metadata based on VAR_NAME values
          5. VAR_NAME value appended to label based on DATA_VARIABLES value
          6. New Global Attribute added

    """


    attr_arr_glob=['PI_NAME','PI_AFFILIATION','PI_ADDRESS','PI_EMAIL',
                   'DO_NAME','DO_AFFILIATION','DO_ADDRESS','DO_EMAIL',
                   'DS_NAME','DS_AFFILIATION','DS_ADDRESS','DS_EMAIL',
                   'DATA_DESCRIPTION','DATA_DISCIPLINE','DATA_GROUP','DATA_LOCATION',
                   'DATA_SOURCE','DATA_VARIABLES','DATA_START_DATE','DATA_STOP_DATE',
                   'DATA_FILE_VERSION','DATA_MODIFICATIONS','DATA_CAVEATS','DATA_RULES_OF_USE',
                   'DATA_ACKNOWLEDGEMENT','DATA_QUALITY','DATA_TEMPLATE','DATA_PROCESSOR',
                   'FILE_NAME','FILE_GENERATION_DATE','FILE_ACCESS','FILE_PROJECT_ID','FILE_DOI',
                   'FILE_ASSOCIATION','FILE_META_VERSION']

    attr_arr_glob_opt=['DATA_DESCRIPTION','DATA_MODIFICATIONS','DATA_CAVEATS','DATA_RULES_OF_USE',
                       'DATA_ACKNOWLEDGEMENT','DATA_QUALITY','DATA_TEMPLATE','DATA_PROCESSOR',
                       'FILE_PROJECT_ID','FILE_ASSOCIATION']

    attr_arr_glob_ins=['DATA_START_DATE','DATA_STOP_DATE','FILE_GENERATION_DATE',
                       'FILE_DOI','FILE_META_VERSION']

    # attr_arr_glob_prov=['DATA_TEMPLATE','DATA_QUALITY']

    attr_arr_data=['VAR_NAME','VAR_DESCRIPTION','VAR_NOTES','VAR_SIZE','VAR_DEPEND',
                   'VAR_DATA_TYPE','VAR_UNITS','VAR_SI_CONVERSION','VAR_VALID_MIN',
                   'VAR_VALID_MAX','VAR_FILL_VALUE']

    attr_arr_data_opt=['VAR_NOTES']

    #Attributes that have user defined values
    attr_free=['DATA_DESCRIPTION','DATA_MODIFICATIONS','DATA_CAVEATS',
               'DATA_RULES_OF_USE','DATA_ACKNOWLEDGEMENT','DATA_QUALITY',
               'DATA_PROCESSOR','FILE_PROJECT_ID','FILE_ASSOCIATION',
               'VAR_DESCRIPTION','VAR_NOTES']

    #Atrributes that have numeric values (except for string datatype)
    num_atts=['VAR_VALID_MIN','VAR_VALID_MAX','VAR_FILL_VALUE','_fillvalue','valid_range']

    #To check for pre-defined attributes
    ncsa=['long_name','units','format','coordsys','valid_range','_fillvalue','scale_factor',
          'scale_factor_err','add_offset','add_offset_err','calibrated_nt']

    #Possible sub-value separators
    sv_sep=['.','_',';']

    nchar=65535 # limit on number of characters

    # Possible error messages for this procedure
    errtxt=['']*9
    proname='Read_Metadata procedure: '
    errtxt[0]=' value contains too many characters (maximum permitted: '+ str(nchar)+')'
    errtxt[1]='No or invalid DATA_VARIABLES and VAR_NAME attributes present in the metadata'
    errtxt[2]=' value appears more than once: '
    errtxt[3]='No valid Metadata in input'
    errtxt[4]='VAR_NAME does not match corresponding DATA_VARIABLES value: '
    errtxt[5]='Number of DATA_VARIABLES values does not match the number of VAR_NAME occurrences '
    errtxt[6]='Global Attribute Label not present in the Metadata: '
    errtxt[7]='Global Attribute Label repeated in the Metadata: '
    errtxt[8]=' occurrences does not match the number of datasets (should be '

    dvfound=-1
    nd=len(attr_arr_data)
    ng=len(attr_arr_glob)
    nna=len(num_atts)
    dsei=np.where(np.array(attr_arr_glob) == 'DS_EMAIL')[0] # any new global attributes should be after this attribute
    # NOTE: not used?
    vni=np.where(np.array(attr_arr_data) == 'VAR_NAME')[0]  # used in VAR_NAME checks
    # ndv=0 # will reset to the number of datasets

    if inf ==1:
        # metafile is a file name
        with open(metafile, 'r') as lu:
            mh = lu.readlines()
            nrline = len(mh)
    else:
        # metadata is already in a list
        mh = metafile

    # Remove leading whitespace
    mhhold = [i.lstrip() for i in mh]

    # Indicate wanted/unwanted lines with True/False
    mhgood=[True]*len(mh)

    # Check for Redundant Attribute Labels
    geoms_rule_changes_0(mhhold, mhgood)

    # test1 - check first character is A-Z or a-z
    test1=[i[0].isalpha() for i in mhhold]
    test1b=[i[0:10]=='_FILLVALUE' for i in mhhold]
    test1=[ta or tb for ta, tb in zip(test1, test1b)]

    # test2 - check for '=' and not a redundant label
    test2=[a.count('=') > 0 for a in mhhold]
    test2=[ta and tb for ta, tb in zip(test2, mhgood)]

    # Indices to valid data
    nri=np.where(test1 and test2)[0]
    nrline=len(nri)

    if nrline==0:
        stop_with_error(proname, errtxt[3])

    # Generate an error for invalid metadata attributes
    bnri, bnrline = where_eq(test1, False)
    if bnrline != 0:
        for i in range(bnrline):
            infotxt='2 Invalid Metadata Attribute: '+mhhold[bnri[i]].rstrip()+'|'
            infotxt=infotxt+' removed from the metadata saved to the output file'
            infotxt_output(infotxt)

    # Set holding arrays for metadata
    mh = [mh[i] for i in nri]
    md.mv_lng = [0]*len(nri)
    md.mv_dbl = [0]*len(nri)

    # m_v2=INTARR(nrline) ;global/variable attribute switch

    # for metadata labels and values
    m_v0 = ['']*nrline
    m_v1 = ['']*nrline
    m_v2 = [0]*nrline

    # Separate out the Meta entry into two components
    for i in range(nrline):
        m_v0[i], m_v1[i] = mh[i].split('=', maxsplit=1)

    # Strip newline endings from the values
    m_v1 = [i.rstrip(os.linesep) for i in m_v1]

    if len(m_v1) != nrline:
        print("Unexpected length of metadata array {len(m_v1)} != {nrline} ")
        sys.exit()

    # Strip leading and trailing whitespace from labels
    m_v0 = [i.strip() for i in m_v0]

    # Force uppercase attribute labels, except for HDF predefined labels
    for inum, val in enumerate(m_v0):
        if val not in ncsa:
            m_v0[inum] = val.upper()

    # Check for and remove leading and trailing spaces in Metadata (sub-)values (except for free-text
    # attributes
    # and VAR_FILL_VALUEs). Also for non-free-text attributes check if the first and last characters are
    # sub-value separators, or if there are consecutive sub-value separators in the attribute value.
    # Also check for values with excessive length.

    if data.qa_yes:
        wstxt='are not permitted'
    else:
        wstxt='removed'  # used when reporting white space in sub-values

    for i, val in enumerate(m_v0):
        if val not in attr_free: # not a free text attribute / IDL scnt==0
            nmcnt = len(np.where(val.upper == np.array([j.upper() for j in num_atts])[0]))
            res = m_v1[i].split(';')
            svcnt = len(res)
            if not(all([j.strip() == j for j in res])): # Leading/trailing Whitespace detected
                if svcnt == 1 and data.qa_yes:
                    if nmcnt != 0 and res[0].strip() != '':
                        infotxt='' # value should be a number not a string so do not check for spaces
                    else:
                        if res[0] == ' ':
                            infotxt='' # there could be a single space in the HDF file if there is no entry
                        else:
                            infotxt='2 Leading or trailing spaces in '+m_v0[i]+'='+m_v1[i]+' value '+wstxt
                elif svcnt > 1:
                    infotxt='2 Leading or trailing spaces between '+m_v0[i]+' sub-values '+wstxt
                else:
                    infotxt='' # Quietly fix if only single value and not doing QA

                if infotxt != '':
                    infotxt_output(infotxt)
                res=[j.strip() for j in res]
                m_v1[i]=res[0]
                if svcnt > 1:
                    for j in range(1,svcnt):
                        m_v1[i]=m_v1[i]+';'+res[j]

            if nmcnt == 0:
                # Check that first and last characters of the metadata value aren't sub-value separators
                # and that there are no consecutive sub-value separators
                mv1len=len(m_v1[i])        # string length of attribute value
                mv1start=m_v1[i][0]  # first character
                mv1end=m_v1[i][-1] # last character
                for k in (0,1):
                    writeonce=True      # so that error message is only written once for each attribute
                    if k == 0:
                        infotxt='3 Sub-value separators at the start or end of the attribute value are not permitted'
                    else:
                        infotxt='3 Consecutive sub-value separators are not permitted'
                    for j in range(len(sv_sep)):
                        spchk=sv_sep[j]+sv_sep[j]
                                    # Note: in some cases a '.' may be present at the end of a sub-value e.g. DO_NAME=Boyd;Ian S.
                        if k == 0 and sv_sep[j] != '.':
                            test=mv1start == sv_sep[j] or mv1end == sv_sep[j]
                        elif k == 1:
                            test=spchk in m_v1[i]
                        else:
                            test=False
                        if test and writeonce:
                            if mv1len > 50:
                                infotxt=infotxt+' in '+m_v0[i]+' attribute'
                            else:
                                infotxt=infotxt+': '+m_v0[i]+'='+m_v1[i]
                            infotxt_output(infotxt)
                            writeonce=False


        if len(m_v1[i]) > nchar:
            infotxt('3 ' + m_v0[i] + errtxt[0])
            infotxt_output(infotxt)

    for i, val in enumerate(m_v0):
        # Do check for DATA_VARIABLES and separate out the values
        if val == 'DATA_VARIABLES':
            dv=m_v1[i].split(';') ; ndv=len(dv) ;  dvfound=i

        # Check for FILE_META_VERSION and add value derived from READ_TABLEFILE if not doing QA
        if val == 'FILE_META_VERSION':
            if not data.qa_yes and m_v1[i] != tabledata.tab_ver:
                m_v1[i]=tabledata.tab_ver
                infotxt='0 FILE_META_VERSION updated with corrected TAV file version and tool name'
                infotxt_output(infotxt)
            elif data.qa_yes:
                # Check Revision number is less than or equal to the TAV revision number, must be '04''R''xxx'
                revno = m_v1[i].split(';')
                fmvcnt = len(revno)
                tavno=tabledata.tab_ver.split(';')
                if fmvcnt != 2:
                    infotxt='2 FILE_META_VERSION values must include the TAV file version and the HDF file '
                    infotxt=infotxt+'creation tool name e.g. FILE_META_VERSION='+tavno[0]+';CUSTOM'
                    infotxt_output(infotxt)

                revok=0              # default is to report a TAV file version error unless revok eq 3
                if len(revno[0]) == len(tavno[0]):
                    vrev=revno[0][:2]
                    vtav=tavno[0][:2]
                    rrev=revno[0][2]
                    rtav=tavno[0][2]
                    srev=revno[0][3:]
                    stav=tavno[0][3:]

                    if vrev.isnumeric():
                        if int(vrev) >= 4 and int(vrev) <= int(vtav):
                            revok += 1 # TAV version OK

                    if rrev == rtav:
                        revok += 1 # 'R' is present as the third character
                    if srev.isnumeric() and revok == 2:
                        if int(srev) <= int(stav) or (int(srev) > int(stav) and int(vrev) < int(vtav)):
                            revok += 1 # TAV sub-version OK
                if revok != 3:
                    infotxt='2 FILE_META_VERSION sub-value '+revno[0]+' is not valid. It must be the TAV file version e.g. '+tavno[0]
                    infotxt_output(infotxt)


        # Check whether a Global Attribute (1), GEOMS Variable Attribute (2), or pre-defined Variable Attribute (3)
        gi=np.where(val == np.array(attr_arr_glob))[0] ; gcnt=len(gi)
        vi=np.where(val == np.array(attr_arr_data))[0] ; vcnt=len(vi)
        pi=np.where(val.lower() == np.array(ncsa))[0] ; pcnt=len(pi)

        if gcnt != 0:
            m_v2[i]=1
        elif vcnt != 0:
            m_v2[i]=2
        elif pcnt != 0:
            m_v2[i]=3
        else:  # check for new Global Attribute
            if i > 0:
                test1=m_v2[i-1] == 1
            else:
                test1=1  # i.e. the previous attribute was also a global attribute

            test2=(val[0:3] != 'VAR') and (val[0:3] != 'VIS') # Does not start with 'VAR' or 'VIS'
            test3=len(val.split()) == 1                       # No spaces in the label
            test4=True 
            if (test1) and (test2) and (test3) and (test4):
                m_v2[i]=1  # it is considered a new global attribute


    # Check for any invalid attribute labels and remove
    bi=np.where(np.array(m_v2) == 0)[0]
    bcnt = len(bi)
    if bcnt != 0:
        for i in range(bcnt):
            infotxt='2 Unknown Attribute Label '+m_v0[bi[i]]+'| ignored'
            infotxt_output(infotxt)

    # Resize arrays, or stop the program if no valid data
    gi=np.where(np.array(m_v2) != 0)[0] ; nrline = len(gi)
    if nrline != 0:
        # Use np.array for the indexing and convert back to list
        m_v0 = list(np.array(m_v0)[gi])
        m_v1 = list(i for i in np.array(m_v1)[gi])
        m_v2 = list(i for i in np.array(m_v2)[gi])

        md.mv_lng = list(np.array(md.mv_lng)[gi])
        md.mv_dbl = list(np.array(md.mv_dbl)[gi])
    else:
        stop_with_error(proname, errtxt[3])

    # Check to see that Global and Variable Attributes are present
    gi = np.where(np.array(m_v2) == 1)[0] ; gcnt=len(gi)
    vi = np.where(np.array(m_v2) == 2)[0] ; vcnt=len(vi)
    pi = np.where(np.array(m_v2) == 3)[0] ; pcnt=len(pi) # to do pre-defined attribute checks if necessary
    if (gcnt == 0) or (vcnt == 0):
        stop_with_error(proname, errtxt[3])

    # Do QA on pre-defined attributes
    if pcnt != 0:
        pre_defined_att_checks(-1, [m_v0[i] for i in pi])


    # Do DATA_VARIABLES checks
    if ndv == 0:
        test1 = np.where(np.array(m_v0) == 'VAR_NAME')[0]
        test2 = np.where(np.array(m_v1) == '')[0]
        dvi = [x for x in test1 if x in test2] ; ndv = len(dvi)

        if ndv != 0:
            dv = list(np.array(m_v1)[dvi])
            # Create or append to DATA_VARIABLES attribute
            if dvfound == -1:
                nrline += 1
                m_v0.append('DATA_VARIABLES')
                m_v1.append('')
                m_v2.append(1)
                md.mv_lng.append(0)
                md.mv_dbl.append(0.0)
                infotxt='2 Missing DATA_VARIABLES Global Attribute| added to metadata based on VAR_NAME values'
            else:
                infotxt='2 Missing DATA_VARIABLES attribute values| appended to the metadata based on VAR_NAME values'
            infotxt_output(infotxt)
        else:
            stop_with_error(proname, errtxt[1])
            # No or invalid DATA_VARIABLES and VAR_NAME attributes present in the metadata so cannot proceed

    # Check for repeated DATA_VARIABLES
    for i in dv:
        if (dv.count(i) > 1) and (i != ''):
            stop_with_error(proname, 'DATA_VARIABLES' + errtxt[2] + dv[i])


    # Do VAR_NAME checks
    vi=np.where(np.array(m_v0) == attr_arr_data[vni[0]])[0] # identify VAR_NAME indices
    vcnt = len(vi)
    if vcnt != 0:
        # Check if VAR_NAME values are present, if not add value from DATA_VARIABLES (if
        # number of DATA_VARIABLES values match vcnt)
        if ndv == vcnt:
            for i in range(vcnt):
                if (m_v1[vi[i]] != '') and (m_v1[vi[i]] != dv[i]):
                    stop_with_error(proname, errtxt[4]+'('+m_v1[vi[i]]+'/'+dv[i]+')')
                elif m_v1[vi[i]] == '':
                    m_v1[vi[i]]=dv[i]
                    infotxt='2 Missing VAR_NAME value for '+m_v0[vi[i]]+'|. '+dv[i]+'| appended to label based on DATA_VARIABLES value'
                    infotxt_output(infotxt)
        else:
            # number of DATA_VARIABLES values does not match the number of VAR_NAME occurrences
            stop_with_error(proname, errtxt[5] + '(' + str(ndv) + '/' + str(vcnt) + ')')

        # Check for repeated VAR_NAMES
        for i in range(vcnt):
            ri=np.where(np.array(m_v1)[vi[i]] == np.array(m_v1)[vi])[0]
            if len(ri) > 1 and np.array(m_v1)[vi[i]] != '':
                stop_with_error(proname,'VAR_NAME'+errtxt[2]+m_v1[vi[i]])

    else:
        # No VAR_NAME labels so create from DATA_VARIABLES
        for i in range(ndv):
            pdb.set_trace()
            m_v0.append('VAR_NAME')
            m_v1.append(dv[i])
            m_v2.append(2)
            md.mv_lng.append(0)
            md.mv_dbl.append(0.0)
            if i == 0:
                vi=[nrline]
            else:
                vi.append(nrline+1)

        infotxt='1 Missing VAR_NAME attribute labels and values| added to metadata based on DATA_VARIABLES values'
        infotxt_output(infotxt)


    # Check for obsolete FTIR, UVVIS.DOAS, LIDAR, and MWR DATA_SOURCE_01 values and modify of possible
    dsi, dsicnt=where_eq(m_v0, 'DATA_SOURCE')
    vni, vncnt=where_eq(m_v0, 'VAR_NAME')
    dv=[m_v1[i] for i in vni] 
    data_source=''

    if dsicnt == 1:
        if m_v1[dsi[0]] != '':
            data_source=m_v1[dsi[0]]
            dschange=data_source
            # geoms_rule_changes(6, vncnt, dv, data_source)
            geoms_rule_changes_6(vncnt, dv, data_source)
            if data_source != dschange:
                m_v1[dsi[0]]=data_source


    # Check for missing Global Attributes which have values which can be calculated by the program
    for i, val in enumerate(attr_arr_glob_ins):
        gaii, gaicnt = where_eq(m_v0, attr_arr_glob_ins[i])
        if gaicnt ==  0: # missing mandatory global attribute so add to the array
            if attr_arr_glob_ins[i] == 'FILE_META_VERSION':
                itxt=tabledata.tab_ver
            else:
                itxt=''
            nrline += 1
            m_v0.append(attr_arr_glob_ins[i])
            m_v1.append(itxt)
            m_v2.append(1)
            md.mv_lng.append(0)
            md.mv_dbl.append(0.0)
            infotxt='2 Missing mandatory Global Attribute '+attr_arr_glob_ins[i]+'| added'
            infotxt_output(infotxt)

    # Do DATA_TEMPLATE/DATA_QUALITY checks
    ngp=len(md.attr_arr_glob_prov)
    ga_chk=['']*ngp # initialize DATA_TEMPLATE/QUALITY values
    geoms_rule_changes_9(m_v0, m_v1, ga_chk, md, tabledata)
    # Check to see if Global Attributes have to be added
    for i in range(ngp):
        if ga_chk[i] != '':
            m_v0.append(md.attr_arr_glob_prov[i])
            if i == 0:
                itxt='='+ga_chk[i]
                m_v1.append(ga_chk[i])
            else:
                itxt=''
                m_v1.append('')
            m_v2.append(1)
            nrline += 1
            md.mv_lng.append(0)
            md.mv_dbl.append(0.0)
            infotxt='2 Missing Global Attribute ' + md.attr_arr_glob_prov[i]+itxt+'| added'
            infotxt_output(infotxt)


    # Calculate expected size of meta_arr array
    gcnt=m_v2.count(1)
    if gcnt > ng:
        ngv=gcnt
    else:
        ngv=ng
    n_arr=ngv+(nd*ndv)
    md.meta_arr=['']*n_arr
    mlh=[0]*n_arr
    mdh=[0.0]*n_arr
    # mlh and mdh are holding arrays to hold numeric values in the metadata (will be renamed mv_lng and mv_dbl)

    # Check for missing or repeated standard Global Attributes
    i=0
    while i <= ng-1:
        gai, gacnt =where_eq(list(np.array(m_v0)[gi]), attr_arr_glob[i])
        if gacnt == 0: # Check to see if Global Attribute Optional or Mandatory
            gaoi, gaocnt=where_eq(attr_arr_glob[i], attr_arr_glob_opt)
            if gaocnt == 0: # it is mandatory, therefore missing
                infotxt='3 Global Attribute Label not present in the Metadata: '+attr_arr_glob[i]
                infotxt_output(infotxt)
        elif gacnt > 1: # Global Attribute Repeated ;
            pdb.set_trace()
            stop_with_error(proname, errtxt[7]+attr_arr_glob[i])

        if gacnt == 0: # remove attribute from the attr_arr_glob array
            print("remove global attribute", attr_arr_glob[i])
            attr_arr_glob.pop(i)
            ng=ng-1
        else:
            i += 1

    # Check for all Standard Global Attributes, add any new ones, and put in expected order
    xg=0

    for i in range(gcnt):
        gai, gacnt=where_eq(attr_arr_glob, m_v0[gi[i]] )
        if gacnt == 0: # New Global Attribute
            attr_arr_glob.insert(i, m_v0[gi[i]])

            ng=ng+1
            infotxt='1 Non-standard Global Attribute '+m_v0[gi[i]]+'| added'
            infotxt_output(infotxt)

    for i in range(gcnt):
        gai, gacnt = where_eq(attr_arr_glob, m_v0[gi[i]])
        md.meta_arr[gai[0]] = m_v0[gi[i]] + '=' + m_v1[gi[i]]

    # Check for repeated new Global Attributes
    for i in range(ng):
        gai, gacnt=where_eq(np.array(m_v0)[gi], attr_arr_glob[i])
        if gacnt > 1:  # Global Attribute Repeated
            stop_with_error(proname, errtxt[7]+attr_arr_glob[i])

    # Do checks for all the variable attributes and put in expected order
    vni=-1
    for i in range(nd):
        si=ng+i # start index of meta_arr for variable attributes
        di, dcnt=where_eq(m_v0, attr_arr_data[i])
        if attr_arr_data[i] == 'VAR_NAME':
            vni=di # to enable checks on optional variable attributes
        if dcnt == ndv: # i.e. correct number of attribute labels found so add to meta_arr in correct order
            for j in range(ndv):
                md.meta_arr[si+(nd*j)]=m_v0[di[j]]+'='+m_v1[di[j]]
                mlh[si+(nd*j)]=md.mv_lng[di[j]]
                mdh[si+(nd*j)]=md.mv_dbl[di[j]]

        elif dcnt == 0: # attribute not present - so add with no label
            oi, ocnt = where_eq(attr_arr_data_opt, attr_arr_data[i])  # check whether missing attribute is optional
            if ocnt == 0:
                for j in range(ndv):
                    md.meta_arr[si+(nd*j)]=attr_arr_data[i]+'='
                if data.qa_yes:
                    itxt=' must be present in the metadata'
                else:
                    itxt=' added to the metadata with no value'
                infotxt='2 Mandatory variable attribute '+attr_arr_data[i]+itxt
                infotxt_output(infotxt)

        else: # incorrect number of attributes - report with error unless attribute is optional
            oi, ocnt=where_eq(attr_arr_data_opt, attr_arr_data[i])
            if (ocnt != 0) and (vni[0] >= 0) and (di[0]-vni[0] > 0): # match VAR_NOTES attributes to the corresponding VAR_NAME index value ;
                for j in range(dcnt):
                    vnih=di[j]-vni[0] ; jh=0
                    for k in range(len(vni)):
                        dvn=di[j]-vni[k]
                        if (dvn < vnih) and (dvn > 0):
                            vnih=di[j]-vni[k] ; jh=k
                    md.meta_arr[si+(nd*jh)]=m_v0[di[j]]+'='+m_v1[di[j]]
                    mlh[si+(nd*jh)]=md.mv_lng[di[j]]
                    mdh[si+(nd*jh)]=md.mv_dbl[di[j]]
            else: # Not optional or too much out of order, therefore report as an error
                stop_with_error(proname,'Number of ' + attr_arr_data[i] + errtxt[8] + str(ndv))

    # Remove any empty meta_arr cells (can happen due to optional attributes)
    # and save numeric metadata values in correct order
    gi, _ = where_eq(md.meta_arr, '')
    mv_lng=[]
    mv_dbl=[]
    for i in range(len(md.meta_arr)):
        if i not in gi:
            mv_lng.append(mlh[i])
            mv_dbl.append(mdh[i])
    md.meta_arr = [i for i in md.meta_arr if i != '']
    md.mv_lng = mv_lng
    md.mv_dbl = mv_dbl


    # Call GEOMS_RULE_CHANGES to check for obsolete Dataset names and update if possible
    # Identify VAR_NAME and VAR_UNITS indices
    var_arr = [i[0:8] for i in md.meta_arr]
    unit_arr = [i[0:9] for i in md.meta_arr]
    vni, vncnt = where_eq(var_arr, 'VAR_NAME')
    vui, vucnt = where_eq(unit_arr, 'VAR_UNITS')

    data.vnchange=['']*vncnt # array to hold original VAR_NAME values if they have been changed
    # extract VAR_NAME values
    dv=['']*vncnt

    for i in range(vncnt):
        res=re.split('=', md.meta_arr[vni[i]])
        if len(res) == 2:
            dv[i]=res[1]

    # Check for obsolete _START/STOP/INTEGRATION sub-values and replace with DATETIME.START/STOP
    # or INTEGRATION.TIME (for single occurences only). Also check for double underscores and
    # replace with single underscores in the Variable Names

    # Set list of data variables
    data.vn = dv
    geoms_rule_changes_1(vncnt, dv, data)

    # TODO: TEST
    # Create a geoms structure to simplify checks later
    # md.attr_arr_glob = attr_arr_glob
    # md.attr_arr_data = attr_arr_data
    # md.attr_free = attr_free
    # data.nvn = len(data.vn)
    # data.inf = 1
    # dtest=None ; vc=0 ; infowrite=''
    # data.vserror=['']*data.nvn
    # dtest = data.extract_data(vc, dtest, infowrite) # return the dataset
    # data.meta_data_to_geoms(md)

    # TEST END

    # Check for other obsolete variable names
    if vncnt == vucnt:
        vuv=['']*vucnt
        for i in range(vucnt):
            res=md.meta_arr[vui[i]].split('=')
            if len(res) == 2:
                vuv[i]=res[1].strip()
        # geoms_rule_changes(5,vncnt,dv,vuv,data_source)
        geoms_rule_changes_5(vncnt, dv, vuv, data_source, data)

    # Update meta_arr in the case of modified Dataset name values
    vci = np.where(np.array(data.vnchange) != '')[0]
    if len(vci) != 0: # need to update VAR_NAME and DATA_VARIABLES attribute values
        var_arr = [i[0:14] for i in md.meta_arr]
        dvi, _ = where_eq(var_arr, 'DATA_VARIABLES')
        md.meta_arr[dvi[0]]='DATA_VARIABLES='
        for i in range (vncnt):
            # Check for new VAR_NAME value and update
            if data.vnchange[i] != '':
                md.meta_arr[vni[i]]='VAR_NAME='+dv[i]
                # Append to new DATA_VARIABLES list
            if i == 0:
                itxt=''
            else:
                itxt=';'
            md.meta_arr[dvi[0]]=md.meta_arr[dvi[0]]+itxt+dv[i]


    # Check for obsolete CALVAL and EVDC FILE_ACCESS values and replace
    var_arr = [i[0:11] for i in md.meta_arr]
    fai, _ = where_eq(var_arr, 'FILE_ACCESS')
    res = re.split(' |=|;', md.meta_arr[fai[0]])
    facnt = len(res)
    if facnt >= 2:
        fav= [i.upper() for i in res[1:facnt]]
        fachange=fav
        #geoms_rule_changes(7,facnt-1,fav)
        geoms_rule_changes_7(facnt, fav, data)
        if fav != fachange: # FILE_ACCESS values changed so update meta_arr
            md.meta_arr[fai[0]]='FILE_ACCESS='
            for i in range(facnt-1):
                if i == 0:
                    itxt=''
                else:
                    itxt=';'
                md.meta_arr[fai[0]]=md.meta_arr[fai[0]]+itxt+fav[i]


    # Check for VAR_UNITS=DIMENSIONLESS or NONE and fix if possible.
    # Need to check VAR_DATA_TYPE. If string then VAR_UNITS value should be blank, o/w replace with '1'
    var_arr = [i[0:13] for i in md.meta_arr]
    vdi, vdcnt = where_eq(var_arr, 'VAR_DATA_TYPE')

    if vdcnt != 0 and vdcnt == vucnt:
        vdv=['']*vdcnt
        for i in range(vdcnt):
            res=re.split('=', md.meta_arr[vdi[i]])
            nres = len(res)
            if nres == 2:
                vdv[i]=res[1].upper()
            vuchange=vuv
            vuv = geoms_rule_changes_8(vucnt, vuv, vdv, data.qa_yes)
            for i in range (vucnt): # Check for changes and update meta_arr
                if vuchange[i].upper() != vuv[i]:
                    md.meta_arr[vui[i]]='VAR_UNITS='+vuv[i]


    # Check for non-permitted characters (must be ISO646-US character set)
    geoms_rule_changes(10, md.meta_arr)

    md.attr_arr_glob = attr_arr_glob
    md.attr_arr_data = attr_arr_data
    md.attr_free = attr_free

    #pdb.set_trace()
    return md


def set_up_structure(md, data):
    """

    do additional checks on VAR_NAME, VAR_DEPEND, VAR_DATA_TYPE, VAR_UNITS, and
    VAR_FILL_VALUE metadata attributes and set up a structure which will be used to hold the
    data values.

    """

    ALLOWABLE_DATA_TYPES=['BYTE','SHORT','INTEGER','LONG64','REAL','DOUBLE','STRING']
    num_atts=['VAR_VALID_MIN','VAR_VALID_MAX','VAR_FILL_VALUE']
    nna=len(num_atts)

    proname="set_up_structure"

    errtxt = ['']*17

    errtxt[0]='Does not match the equivalent VAR_NAME in DATA_VARIABLES: '
    errtxt[1]='Number of dimensions too high to make HDF file (max. allowed 8)'
    errtxt[2]='VAR_DEPEND attribute value missing or does not match the number of VAR_SIZE values'
    errtxt[3]='VAR_DEPEND value not CONSTANT,INDEPENDENT nor self-referencing, and doesn\'t match a previous VAR_NAME: '
    errtxt[4]='VAR_DEPEND variable name(s) must have VAR_DEPEND value of CONSTANT or be self-referencing: '
    errtxt[5]='VAR_DEPEND variable name must have matching VAR_SIZE value (= '
    errtxt[6]='Data type not BYTE, SHORT, INTEGER, REAL, DOUBLE, or STRING: '
    errtxt[7]='Unexpected number of ATTRIBUTE values extracted: '
    errtxt[8]='VAR_DATA_TYPE must be DOUBLE for VAR_UNITS=MJD2K: '
    errtxt[9]='VAR_UNITS must be MJD2K for ' # no longer used (infotxt message instead)
    errtxt[10]='VAR_FILL_VALUE not a valid value: '
    errtxt[11]=' not valid'
    errtxt[12]='Second VAR_SIZE value should be ''2'' for this VAR_NAME'
    errtxt[13]='Axis variable cannot be dependent on another variable: '
    errtxt[14]=' is already the axis variable in the vertical dimension'
    errtxt[15]='is not an axis variable so VAR_DEPEND value cannot be self-referencing'
    errtxt[16]='Type conversion error.  Entry is not a valid number: '

    # Do a search for a DATA_TEMPLATE value. If not present will do all .BOUNDARIES checks, otherwise
    # these checks will be done by the Template QA program - interim until .BOUNDARIES dimension ordering
    # is consistent for all templates (FTIR retains original convention as of 20140521)
    dtfound=0                     # default to template not present

    dt_tmp = [i[0:13] for i in md.meta_arr]
    dti, dtcnt = where_eq(dt_tmp, 'DATA_TEMPLATE')
    if dtcnt != 0:       # check there is a value
        res = re.split(' |=|;', md.meta_arr[dti[0]])
        rescnt = len(res)
        if rescnt == 2:
            dtfound=1  # valid DATA_TEMPLATE value found

    # Determine a set of possible Axis Variables to be tested (i.e. all VAR_DEPEND values
    # that aren't CONSTANT or INDEPENDENT)
    vdi, vdcnt=where_eq([i[0:10] for i in md.meta_arr], 'VAR_DEPEND')
    vardeptest=['']
    if vdcnt != 0:
        for i in range(vdcnt):
            res=re.split(' |=|;', md.meta_arr[vdi[i]])
            rescnt=len(res)
            if rescnt > 1: # VAR_DEPEND values found
                for j in range(1, rescnt):
                    resuc = res[j].upper()
                    #                    ni, ncnt = where_eq(resuc, vardeptest.upper())
                    ni, ncnt = where_eq([i.upper() for i in vardeptest], resuc)
                    if ncnt == 0 and resuc != 'INDEPENDENT' and resuc != 'CONSTANT':
                        # Add VAR_DEPEND value to list of axis-variables to be checked
                        #vardeptest.append=[vardeptest,res[j]]
                        vardeptest.append(res[j])

    # Remove '' from vardeptest
    vardeptest=[i for i in vardeptest if i != '']

    lc=0
    vc=0
    writeonce=0
    notlong=False                    # Boolean to check whether VAR_SIZE values are of integer data type

    while lc < len(md.meta_arr):
        valid=False                   # set to test for successful string conversion to a number
        if (md.meta_arr[lc][0:14]) == 'DATA_VARIABLES': # determine number of variables reported
            lcdv=lc
            vn=re.split(' |=|;', md.meta_arr[lc])
            data.nvn=len(vn)-1                     # total number of variables reported
            vnold=vn # DBG
            vn=vn[1:]                         # remove the DATA_VARIABLES label
            data.vn = vn                 # copy to 'data' structure
            data.vu=['']*data.nvn
            vfvd=np.zeros((nna,data.nvn))    # holding array for floating point num_atts array values
            vfvl=np.zeros((nna,data.nvn), dtype=np.int64)  # holding array for integer num_atts array values
            data.ndm=np.zeros((data.nvn,9), dtype=np.int64)      # to contain the dimensions and data type for each dataset
            data.vserror=['']*data.nvn # to hold variable names for datasets where the VAR_SIZE values cannot be determined


        if md.meta_arr[lc][0:8] == 'VAR_NAME':
            vname=md.meta_arr[lc]       # used in case of error
            lcz=lc                   # VAR_NAME index value
            res=re.split(' |=', md.meta_arr[lc])
            if len(res) != 2:
                res=[res[0],'-1']
            if res[1] != vn[vc]:
                stop_with_error(proname + vname + ': ', errtxt[0]+vn[vc])
                # Does not match the equivalent VAR_NAME in DATA_VARIABLES

            while md.meta_arr[lc][0:8] != 'VAR_SIZE':
                lc += 1
            holdvs = md.meta_arr[lc]  ; lcx=lc  # do checks after determining VAR_DATA_TYPE
            while md.meta_arr[lc][0:10] != 'VAR_DEPEND':
                lc += 1
            holdvd = md.meta_arr[lc] ; lcy=lc  # do checks after determining VAR_DATA_TYPE
            while md.meta_arr[lc][0:13] != 'VAR_DATA_TYPE':
                lc +=1

            res=re.split(' |=|;', md.meta_arr[lc].upper())
            if len(res) == 2:
                gi, gcnt = where_eq(ALLOWABLE_DATA_TYPES, res[1]) # Check VAR_DATA_TYPE is valid
            else:
                gcnt=0
            if gcnt != 1:
                stop_with_error(proname+vname+': ', errtxt[6] + md.meta_arr[lc])
                # Data type not BYTE, SHORT, INTEGER, LONG64, REAL, DOUBLE, or STRING
            if gi[0] == 3:
                if writeonce == 0:
                    writeonce=1
                    infotxt='2 64-bit LONG Data Type is not currently supported|. '
                    infotxt=infotxt+'Will attempt to write data in 32-bit INTEGER Data Type'
                    infotxt_output(infotxt)
                gi[0]=2
                md.meta_arr[lc]='VAR_DATA_TYPE=INTEGER'

            data.ndm[vc,8]=gi[0]          # index number representing the data type

            # Do VAR_SIZE and VAR_DEPEND checks
            resvs=re.split(' |=|;', holdvs)  # VAR_SIZE entries
            resvd=re.split(' |=|;',holdvd)
            nvs = len(resvs)
            nd = len(resvd)
            nvs=nvs-1
            nd=nd-1  # Number of Dimensions based on number of VAR_SIZE/VAR_DEPEND values

            ndmok = False if nvs == 0 else True
            # If there are no VAR_SIZE value entries, need to determine from Data file or structure or
            # from the VAR_SIZE values of the VAR_DEPEND dependencies
            if (nvs != nd) and ndmok or (nd == 0):
                stop_with_error(proname+vname+': ', errtxt[2])
                # VAR_DEPEND attribute value missing or does not match the number of VAR_SIZE values
            elif nd > 8:
                # Number of dimensions too high to make HDF file (max. allowed 8)
                stop_with_error(proname + vname + ': ', errtxt[1])

            elif ndmok:
                data.ndm[vc,0:nd]=[int(i) for i in resvs[1:nvs+1]]  # save VAR_SIZE values in ndm array

                # Check that original input values are of integer data type
                for k in range(1, nvs+1):
                    if not resvs[k].strip().isdigit():
                        resvs[k]=np.int64(resvs[k].strip())
                        notlong=True

                if notlong:
                    for k in range(1, nvs+1):
                        if k == 1:
                            vshold=resvs[k]
                        else:
                            vshold=vshold+';'+resvs[k]
                    md.meta_arr[lcx]='VAR_SIZE='+vshold
                    holdvs=md.meta_arr[lcx]

            vshold=[0]*nd       # to hold VAR_SIZE of the dependencies

            # Do checks on VAR_DEPEND values
            ci1, _ = list(where_eq([i.upper() for i in resvd], 'INDEPENDENT'))
            ci2, _ = list(where_eq([i.upper() for i in resvd], 'CONSTANT'))
            ci = ci1 + ci2
            ccnt = len(ci)

            if ccnt != 0:
                resvd[ci]=resvd[ci].upper()
            # Rules for .BOUNDARY VAR_NAME
            if '.BOUNDARIES' in vn[vc]:
                test1=False
                test2=False

                if nd in (2, 3) and dtfound == 0:
                    # First VAR_DEPEND value must be INDEPENDENT
                    test1=resvd[1] != 'INDEPENDENT'
                    # Second VAR_DEPEND value must be present in the Variable Name e.g.
                    # ALTITUDE for ALTITUDE.BOUNDARIES
                    secvdsplt=re.split('.', resvd[2])
                    test2=vn[vc] not in secvdsplt[0]

                elif (nd <= 1) or (nd >= 4):
                    test2=True  # incorrect number of VAR_DEPEND values
                if nd == 3 and resvd[3] != 'DATETIME':
                    test1=True  # Third VAR_DEPEND value must be DATETIME
                if (test1) or (test2):
                    if test1 == 1:
                        itxt=' in IDL/Fortran Dimension Ordering'
                    else:
                        itxt=''
                    infotxt='3 '+holdvd+' not valid for '+vname+itxt
                    infotxt_output(infotxt)
                    bounderror=1

            for i in range(1, nd+1):
                if i == 1:
                # Check for and Apply GEOMS rule changes for INDEPENDENT and axis variables
                    vdval=resvd[i]
                    vnval=vn[vc]
                    vdval, holdvd = geoms_rule_changes_2(vardeptest, vnval, vdval, holdvd)
                    if vdval != resvd[i]: # First VAR_DEPEND value has been changed
                        pdb.set_trace()
                        # Rebuild VAR_DEPEND entry
                        holdvd='VAR_DEPEND='+vdval
                        if nd > 1:
                            for j in range(2,nd+1):
                                holdvd=holdvd+';'+resvd[j]
                    if holdvd[0:1] == 'E': # error found while attempting to identify axis variables
                        if holdvd == 'E1':
                            stop_with_error(proname+vname+': ', errtxt[13]+vdval)
                        elif holdvd == 'E3':
                            stop_with_error(proname+vname+' ', errtxt[15])
                    resvd[i]=vdval
                    md.meta_arr[lcy]=holdvd

                test1=(resvd[i] == 'CONSTANT') and ((i > 1) or (nd >= 2))  # CONSTANT can be the only VAR_DEPEND
#                test2=(resvd[i] == vn[vc]) and (i > 1 or nd > 2)  # e.g. ALTITUDE;DATETIME are possible dependencies for ALTITUDE
                test2=(resvd[i] == vn[vc]) and (nd > 2)  # e.g. DATETIME;ALTITUDE are possible dependencies for ALTITUDE
                test3=(resvd[i] == 'INDEPENDENT') and ((i > 1) or (nd >= 2)) and (not '.BOUNDARIES' in vn[vc])
                test4=(resvd[i] != 'CONSTANT') and (resvd[i] != 'INDEPENDENT') and (resvd[i] != vn[vc])
                # Check for CONSTANT,INDEPENDENT or self-referencing and more than one VAR_DEPEND value

                # DEBUG: Disabling test2 for now (VAR_DEPEND = DATETIME;ALTITUDE not allowed for ALTITUDE)...
                #if (test1) or (test2) or (test3):
                if (test1) or (test3):
                    stop_with_error(proname + ' ' + vname + ': ',holdvd+errtxt[11])
                    # VAR_DEPEND value(s) not valid

                # Check for valid dependencies, and hold VAR_SIZE values of the dependencies if required
                if vc >= 1:
                    gi, gcnt = where_eq(vn[0:vc], resvd[i])
                else:
                    gcnt=0
                if gcnt == 0 and test4:
                    stop_with_error(proname+vname+': ',errtxt[3]+resvd[i])
                    # VAR_DEPEND value not CONSTANT,INDEPENDENT nor Self-referencing, and doesn't match a previous VAR_NAME

                if gcnt != 0: # do remaining checks on VAR_DEPEND values
                    errfound=False
                    gi, _ = where_eq(md.meta_arr, 'VAR_NAME='+resvd[i])
                    # check VAR_DEPEND value of the VAR_DEPEND variable name
                    vi = np.flatnonzero(np.core.defchararray.find(np.array(md.meta_arr[gi[0]:]), 'VAR_DEPEND') != -1)

                    vex=re.split(' |=|;', md.meta_arr[gi[0]+vi[0]])
                    nvex=len(vex)
                    if nvex != 2:
#                        vex=[vex, '']  # missing VAR_DEPEND value so add dummy value
                        vex.append('')  # missing VAR_DEPEND value so add dummy value
                    #if (vex[1] != 'CONSTANT') and (vex[1] != 'INDEPENDENT') and (vex[1] != resvd[i]):
                    # DEBUG: vex[1]
                    if (vex[1] != 'CONSTANT') and (vex[1] != 'INDEPENDENT') and (vex[1] != resvd[i]):
                        vert_var=['ALTITUDE','ALTITUDE.GPH','PRESSURE','DEPTH']
                        # does VAR_DEPEND variable name reference a vertical axis
                        res, vcnt = where_eq(vert_var, resvd[i])
                        # does VAR_DEPEND of the VAR_DEPEND variable name reference a vertical axis
                        #res, xcnt=where_eq(vert_var, vex[1])
                        # DEBUG: vex[1] is DATETIME when dimensions are reversed
                        res, xcnt=where_eq(vert_var, vex[i])
                        if (vcnt != 0) and (xcnt == 0):
                            pdb.set_trace()
                            # Vertical axis so should be self-referencing
                            infotxt='2 ' + md.meta_arr[gi[0]+vi[0]]+' should be self-referencing for axis variable'
                            infotxt=infotxt+' VAR_NAME='+resvd[i]+'|. VAR_DEPEND value changed in metadata'
                            infotxt_output(infotxt)
                            md.meta_arr[gi[0]+vi[0]]='VAR_DEPEND='+resvd[i]
                            if nvex > 2:
                                for j in range(2,nvex):
                                    md.meta_arr[gi[0]+vi[0]]=md.meta_arr[gi[0]+vi[0]]+';'+vex[j]
                        else:
                            pass

                    if errfound: # redundant from v4.0
                        pdb.set_trace()
                        stop_with_error(proname+vname+': '+md.meta_arr[lcy]+': ',errtxt[4]+md.meta_arr[gi[0]+vi[0]])
                        # VAR_DEPEND variable name must have VAR_DEPEND value of CONSTANT, INDEPENDENT or be self_referencing


                    # check VAR_SIZE value of the VAR_DEPEND variable name
                    vi = np.flatnonzero(np.core.defchararray.find(np.array(md.meta_arr[gi[0]:]), 'VAR_SIZE') != -1)

                    vex=re.split(' |=|;', md.meta_arr[gi[0]+vi[0]])
                    nvex=len(vex)
                    if nvex < 2 or nvex > 3:
                        errfound=True
                        # DEBUG: check this.
                        # Should implement the rest of this function...
                    # DEBUG: disabling this for now - data looks ok?
                    #elif ndmok and int(vex[i]) != data.ndm[vc,i-1]:
                    #elif ndmok and (int(vex[1]) != data.ndm[vc,i-1]):

                        #errfound=True
                    if errfound:
                        stop_with_error(proname + vname+ ': '+'VAR_DEPEND='+resvd[i]+': ', errtxt[5] + str(data.ndm[vc,i-1]) + '): ' + md.meta_arr[gi[0]+vi[0]])
                        # VAR_DEPEND variable name must have matching VAR_SIZE value
                    else:
                        vshold[i-1]=int(vex[1])

                # If i eq 2 and value is INDEPENDENT then check that the corresponding VAR_SIZE value is 2
                #(if present), o/w generate errors
                if i == 2 and resvd[i] == 'INDEPENDENT':
                    if ndmok and data.ndm[vc,i-1] != 2:
                        stop_with_error(proname+vname+': '+holdvs+': ',errtxt[12])
                        # Second VAR_SIZE value should be '2' for this VAR_NAME
                vshold[i-1]=2     # set VAR_SIZE value in the event that it is not part of the Metadata

            if not ndmok: # no VAR_SIZE value so need to determine from the data or the vshold array
                pdb.set_trace()
                if vshold[0] == 0:
                    # VAR_DEPEND is Self-Referencing or CONSTANT so determine VAR_SIZE
                    # from the Data file or structure
                    dtest=['-1']  # -1 identifies to the EXTRACT_DATA routine that ndl is all that is required
                    dtest = data.extract_data(vc)
  #             IF STRLEN(rerr[0]) GT 2 THEN RETURN
  #                               ;Check if any other VAR_SIZE values have already been determined
  #                               ;This could happen in the case VAR_DEPEND=ALTITUDE;DATETIME for VAR_NAME=ALTITUDE for e.g.
  #             gi=WHERE(vshold NE 0L,gcnt)
  #             ndltot=' ('+STRTRIM(ndl,2)+')'
  #             IF gcnt NE 0 THEN BEGIN
  #                FOR i=0,gcnt-1 DO ndl=ndl/FLOAT(vshold[gi[i]])
  #             ENDIF
  #                               ;Make up VAR_SIZE string based on available values
  #             IF vshold[0] EQ 0L THEN vsstr=' (VAR_SIZE=x' ELSE vsstr='VAR_SIZE='+STRTRIM(vshold[0],2)
  #             IF nd GT 1 THEN FOR i=1,nd-1 DO $
  #                IF vshold[i] EQ 0L THEN vsstr=vsstr+';x' ELSE vsstr=vsstr+';'+STRTRIM(vshold[i],2)
  #             vsstr=vsstr+') '
  #             vshold[0]=LONG(ndl)
  #             bi=WHERE(vshold EQ 0,bcnt) ;Expected VAR_SIZE values are missing if bcnt not equal to 0
  #             IF (ndl NE vshold[0]) OR (bcnt NE 0) THEN BEGIN
  #                infotxt=STRARR(2)
  #                infotxt[0]='3 Unable to determine missing VAR_SIZE value(s) for '+vname+' based on'
  #                infotxt[1]='    '+holdvd+vsstr+'and the total number of dataset values'+ndltot
  #                INFOTXT_OUTPUT,infotxt
  #                vserror[vc]=vn[vc]
  #             ENDIF
  #             meta_arr[lcx]='VAR_SIZE='+STRTRIM(vshold[0],2)
  #             ndm[vc,0:nd-1]=vshold
  #             IF nd GT 1 THEN FOR i=1,nd-1 DO meta_arr[lcx]=meta_arr[lcx]+';'+STRTRIM(vshold[i],2)
  #          ENDIF ELSE BEGIN     ;can determine VAR_SIZE from the vshold array
  #             meta_arr[lcx]='VAR_SIZE='
  #             FOR i=0,nd-1 DO BEGIN
  #                IF i EQ 0 THEN meta_arr[lcx]=meta_arr[lcx]+STRTRIM(vshold[i],2) $
  #                ELSE meta_arr[lcx]=meta_arr[lcx]+';'+STRTRIM(vshold[i],2)
  #                ndm[vc,i]=vshold[i] ;EQ array dimensions from VAR_SIZE of the dependencies
  #             ENDFOR
  #          ENDELSE
  #                               ;Check to see if VAR_SIZE calculation is valid or not based on contents of vserror
  #          vsok=1
  #          FOR i=1,nd DO BEGIN
  #             ei=WHERE(resvd[i] EQ vserror,ecnt)
  #             IF ecnt NE 0 THEN vsok=0
  #          ENDFOR
  #          IF vsok EQ 1 THEN BEGIN
  #             IF data.qa_yes THEN qtxt=' not recorded. Calculated as '+meta_arr[lcx] ELSE qtxt=' added: '+meta_arr[lcx]
  #             infotxt='2 VAR_SIZE value(s) for '+vname+qtxt
  #             INFOTXT_OUTPUT,infotxt
  #          ENDIF
            #       ENDIF

            if resvd[1] == 'CONSTANT' and data.ndm[vc,0:7].sum() > 1.:
                infotxt='3 Dataset '+vn[vc]+' must only have a single value for VAR_DEPEND=CONSTANT: ' + md.meta_arr[lcx]
                infotxt_output(infotxt)


            #   Do DATETIME/MJD2000 checks
            while md.meta_arr[lc][0:9] != 'VAR_UNITS':
                lc += 1
            resv=re.split(' |=', md.meta_arr[lc])
            if len(resv) == 1:
                resv=[resv,'[MISSING]'] # add filler for checks
            test1=(vn[vc] == 'DATETIME') or (vn[vc] == 'DATETIME.START') or (vn[vc] == 'DATETIME.STOP')
            if test1 and resv[1].upper() != 'MJD2K':
                lch=lc                # hold setting for VAR_UNITS
                # Find VAR_SI_CONVERSION values and test for correct values for MJD2K
                while md.meta_arr[lch][0:17] != 'VAR_SI_CONVERSION':
                    lch += 1
                res=re.split(' |=|;', md.meta_arr[lch])
                test1=resv[1].lower() == 's' and int(res[1]) ==  0 and int(res[2]) == 86400 and \
                    res[3].lower() == 's'
                if (resv[1] == '[MISSING]') or (resv[1] == '1') or (resv[1].lower() == 'd') or (test1):
                    errid='2 ' ; itxt='|. Value changed in Metadata'
                else:
                    errid='3 ' ; itxt=''

                infotxt=errid + md.meta_arr[lc]+' must be VAR_UNITS=MJD2K for '+vname+itxt
                infotxt_output(infotxt)
                md.meta_arr[lc]='VAR_UNITS=MJD2K'
                resv=['VAR_UNITS','MJD2K']
                # Change VAR_SI_CONVERSION value also

                if data.ndm[vc,8] <= 3:
                    md.meta_arr[lch]='VAR_SI_CONVERSION=0;86400;s'
                elif data.ndm[vc,8] <= 5:
                    md.meta_arr[lch]='VAR_SI_CONVERSION=0.0;86400.0;s'
                else:
                    md.meta_arr[lch]='VAR_SI_CONVERSION='

            data.vu[vc]=resv[1]   # used to check for MJD2K units when reading in data as well as NCSA units value
            if len(resv) > 2:
                for i in range(2, len(resv)):
                    data.vu[vc]=data.vu[vc]+' '+resv[i]

            # Do checks on the numeric attributes
            for i in range(0,nna):
                while md.meta_arr[lc][0:len(num_atts[i])] != num_atts[i]:
                    lc += 1
                resv = re.split(' |=|;', md.meta_arr[lc])
                nresv = len(resv)
                test1=(data.vu[vc].upper() != 'MJD2K') or (num_atts[i] == 'VAR_FILL_VALUE')
                test2=(nresv != 2) and (data.ndm[vc,8] != 6)
                if test1 and test2:
                    infotxt=[]
                    infotxt.append('3 Unexpected number of ATTRIBUTE values extracted for '+vname+':')
                    infotxt.append('    ' + md.meta_arr[lc])
                    infotxt_output(infotxt)
                    resv=[resv,'']     # ensure at least 2 values
                    md.meta_arr[lc]=resv[0]+'='+resv[1]
                    nresv=2

                if (data.ndm[vc,8] != 6) and (nresv == 2):
                # If VAR_UNITS=MJD2000 then check for ISO8601 format and change to MJD2000 if necessary
                    if (data.vu[vc].upper() == 'MJD2K') and ('Z' in resv[1]):
                        mjd2000=julian_date(resv[1])  # return time in MJD2000 format
                        if mjd2000 == -99999.:
                            stop_with_error(proname + vn[vc]+': ', errtxt[10] + md.meta_arr[lc])
                        vfvd[i,vc]=mjd2000
                        infotxt=[]
                        md.meta_arr[lc]=resv[0]+'='+'{:18.9f}'.format(mjd2000).strip()
                        if data.qa_yes:
                            qtxt='    must be in MJD2K format|'
                        else:
                            qtxt='    replaced with MJD2K formatted value '
                            infotxt.append('2 '+resv[0]+'='+resv[1]+' for VAR_NAME='+vn[vc])
                            infotxt.append(qtxt + md.meta_arr[lc])
                            infotxt_output(infotxt)
                    else:
                    #  Check that the VAR_VALID_MIN/MAX or VAR_FILL_VALUE is numeric and save to vfvl or vfvd arrays
                        lcx=lc
                        if data.ndm[vc,8] <= 3: # value is integer type
                            if md.mv_lng[lc] != 0:
                                vfvl[i,vc]=md.mv_lng[lc]
                            else:
                                vfvl[i,vc]=np.int64(np.float64(resv[1]))
                        else: # value is floating point type
                            if md.mv_dbl[lc] != 0.:
                                vfvd[i,vc]=md.mv_dbl[lc]
                            else:
                                vfvd[i,vc]=np.float64(resv[1])
            vc += 1
        lc += 1
        valid=True

    if notlong: # write out INFORMATION or ERROR to log
        if data.qa_yes:
            itxt=' must be '
        else:
            itxt=' changed to '
        infotxt='2 VAR_SIZE values'+itxt+'integer format'
        infotxt_output(infotxt)


    # Save numeric metadata values to common arrays
    md.mv_lng=vfvl
    md.mv_dbl=vfvd

    if data.qa_yes:
        #pass
        #print("set_up_structure called with qa_yes=True")
        # pdb.set_trace()
        # TODO
        # Need to ensure that ndm is stored in correct order.
        # Re-sort the geoms sdss_d dict according to data.vn sort order
        new_dict = OrderedDict((key, data.geoms.sdss_d[key]) for key in data.vn)
        # Replace in place
        data.geoms.sdss_d.clear()
        data.geoms.sdss_d.update(new_dict)



    # Check VAR_VALID_MIN, VAR_VALID_MAX, VAR_FILL_VALUE and Dataset data types match VAR_DATA_TYPE
#      s_dims=SIZE(sds,/DIMENSIONS)
#      pchkd=PTR_VALID(sds.data)
#      pchkv=PTR_VALID(sds.va_v)
#      pchkl=PTR_VALID(sds.va_l)

#      IF s_dims[0] GE 1 THEN BEGIN
#         attnamechk=['VAR_NAME','VAR_DATA_TYPE','VAR_VALID_MIN','VAR_VALID_MAX','VAR_FILL_VALUE']
#         n_anc=N_ELEMENTS(attnamechk)
#         FOR i=0,s_dims[0]-1 DO BEGIN
#            vdtx=BYTARR(n_anc)+255B ;VAR_DATA_TYPE for the attnamechk values
#            vnh=''
#            FOR j=0,s_dims[1]-1 DO BEGIN
#               IF (pchkl[i,j]) AND (pchkv[i,j]) THEN BEGIN
#                  res=*sds[i,j].va_l & res=STRTRIM(STRUPCASE(res),2)
#                  ai=WHERE(res EQ attnamechk,acnt)
#                  IF acnt NE 0 THEN BEGIN
#                     attval=*sds[i,j].va_v
#                     IF ai[0] EQ 0 THEN vnh=STRTRIM(STRUPCASE(attval),2) $
#                     ELSE IF ai[0] EQ 1 THEN BEGIN
#                        vdth=STRTRIM(STRUPCASE(attval),2)
#                        vi=WHERE(vdth EQ allowable_data_types,vcnt)
#                        IF vcnt NE 0 THEN vdtx[1]=idl_data_type[vi[0]]
#                     ENDIF ELSE vdtx[ai[0]]=SIZE(*sds[i,j].va_v,/TYPE)
#                  ENDIF
#               ENDIF
#            ENDFOR
#            IF pchkd[i,0] THEN dvdtx=SIZE(*sds[i,0].data,/TYPE) ELSE dvdtx=255B ;Data type of the dataset
#                                 ;PRINT,vnh,dvdtx,vdtx[1:4]
#            IF vdtx[1] NE 255B THEN BEGIN
#               IF (dvdtx NE vdtx[1]) AND (dvdtx NE 255B) THEN BEGIN
#                  vi=WHERE(dvdtx EQ idl_data_type,vcnt)
#                  IF vcnt NE 0 THEN svdt=allowable_data_types[vi[0]] ELSE svdt=''
#                  IF svdt EQ '' THEN $
#                     infotxt='3 '+vnh+' dataset has data type that does not match VAR_DATA_TYPE='+vdth $
#                  ELSE $
#                     infotxt='3 '+vnh+' dataset has data type '+svdt+', which does not match VAR_DATA_TYPE='+vdth
#                  INFOTXT_OUTPUT,infotxt
#               ENDIF
#               FOR j=2,n_anc-1 DO BEGIN
#                  IF vdtx[j] NE vdtx[1] THEN BEGIN
#                     vi=WHERE(vdtx[j] EQ idl_data_type,vcnt)
#                     IF vcnt NE 0 THEN svdt=allowable_data_types[vi[0]] ELSE svdt=''
#                     IF svdt EQ '' THEN $
#                        infotxt='3 '+vnh+' '+attnamechk[j]+ $
#                                ' value has data type that does not match VAR_DATA_TYPE='+vdth $
#                     ELSE $
#                        infotxt='3 '+vnh+' '+attnamechk[j]+' value has data type '+svdt+ $
#                                ', which does not match VAR_DATA_TYPE='+vdth
#                     INFOTXT_OUTPUT,infotxt
#                  ENDIF
#               ENDFOR
#            ENDIF
#         ENDFOR
#      ENDIF
#   ENDIF

# ;Define the HDF storage structure
#   ds_set={data: PTR_NEW()}      ;SDS data array
#   ds=REPLICATE(ds_set,data.nvn)      ;Dimension the structure to the size of the SDS datasets

# TypeConversionError:
#   IF valid EQ 0 THEN BEGIN
#      STOP_WITH_ERROR,o3[3]+proname+vname+': ',errtxt[16]+meta_arr[lcx],lu & RETURN
#   ENDIF

    #pdb.set_trace()


def julian_date(date_ut, iso=True, mjd2000=True):
    """

    Computes the Julian date (jd) or date in MJD2000 format in double precision.
     ----------
       Bojan R. Bojkov
       bojan.r.bojkov@nasa.gov
       06/03/2001
       09/24/2001: Code cleanup                        - Function Version 1.0
       10/16/2001: Mode corrections                    - Function Version 1.01
       03/07/2002: Comment cleanup and variable change - Function Version 2.0
       10/08/2002: Added jdf option                    - Function Version 2.1
       07/21/2003: Converted to a function             - Function Version 3.0
       05/25/2005: Added ISO8601 and MJD2000 keywords.
                   Input can either be a string or a numeric array
                   (see below for input details). Returns -99999.d
                   if the Input is invalid - Ian Boyd

     References ----------
       Hughes, D.W., Yallop, B.D. and Hohenkerk, C.Y., "The Equation of Time",
         Mon. Not. R. astr. Soc., 238, pp 1529-1535. 1989.

       Results verified using:
         Meeus, J, "Astronomical Algorithms", William-Bell, Inc., Richmond VA,
         p62, 1991.

     Caveats ----------
       NONE

     Input ----------
       date_ut: List of (UT: YYYY, MM, DD, [hh, mm, ss]), or if ISO8601
                keyword set, a string containing datetime as YYYYMMDDThhmmssZ

     Output ---------
       jd:      julian day, or if MJD2000 keyword set,
                a double precision day in MJD2000 format

     External subroutines ---------
       NONE

    """
    from datetime import datetime

    valid=0                       # Type conversion check
    input_error=False              # Check that input is as expected
    #Check input format
    if iso: # input is YYYYMMDDThhmmssZ
        if len(date_ut) == 16:
            # test for valid date (will return type conversion error value if not a number)
            try:
                date_check = datetime(date_ut, '%Y%m%dT%H%M%SZ')
            except ValueError as ve:
                input_error=True

            dt=[]
            dt.append(date_ut[:4])
            dt.append(date_ut[4:6])
            dt.append(date_ut[6:8])
            dt.append(date_ut[9:11])
            dt.append(date_ut[11:2])
            dt.append(date_ut[13:2])
            date_ut=dt ; achk=[0,6]  # to add hhmmss component
        else:
            input_error=True

    else:
        # check that input contains at least YMD (and at most YMDhms) information and is either
        # an integer, long, float or double array
        achk=len(date_ut)
        if (achk < 3) or (achk > 6):
            input_error=True


    if not input_error:  # can perform JD calculation
        if date_ut[1] > 2:
            y=np.float64(date_ut[0])
            m=np.float64(date_ut[1]-3)
            d=np.float64(date_ut[2])
        else:
            y=np.float64(date_ut[0]-1)
            m=np.float64(date_ut[1]+9)
            d=np.float64(date_ut[2])


    # Compute Julian date:
    j=np.int64(365.25*(y+4712.0)) + np.int64(30.6*m+0.5)+59.+d-0.5

    # Check for Julian or Gregorian calendar:
    if j < 2299159.5:
        jd=j # If Julian calendar.
    else:            # If Gregorian calendar.
        gn=38. - np.int64(3.*np.int64(49.0 + y/100.)/4.)
        jd=j+gn

    # add hhmmss values if present
    if achk == 4:
        jd=jd + np.float64(date_ut[3])/24.                                    # hh only
    elif achk == 5:
        jd=jd+(np.float64(date_ut[3])*3600. + np.float64(date_ut[4])*60.)/86400. # hhmm only

    elif achk == 6:
        jd=jd+(np.float64(date_ut[3])*3600. + np.float64(date_ut[4])*60. + \
               np.float64(date_ut[5]))/86400. # hhmmss

    valid=1                     # Type conversion OK

                                # Set to MJD2000 if required
    if mjd2000:
        jd=jd-2451544.5


    if valid == 0:
        return -99999.
    else:
        return jd



def read_data(md, data):
    """

    Perform checks on the data file

    Inputs:
      md - Metadata object
      data - Data object

    """

    mint=np.float64(90000.)
    maxt=np.float64(-90000.)  # initialize min and max datetimes

    valfound=0  # count variable that increments when DATETIME and related datasets are found
    errorfound=1  #variable to indicate error in DATETIME checks
    infowrite=0 # do not write ISO8601 to MJD2K information message in EXTRACT_DATA routine (will write on last EXTRACT_DATA call)

    vchk=['VAR_VALID_MIN', 'VAR_VALID_MAX']
    nvchk=len(vchk)
    mmi=['']*nvchk

    for i in range(nvchk):
        mi, _ = where_eq(md.attr_arr_data, vchk[i])
        mmi[i] = mi[0]

    vavk=[]*data.nvn # holding array for AVK attribute (used below)

    # Determine number of DATETIME values, and set arrays for DATETIME checks
    dtfv=[mint]*3
    dti, dtcnt = where_eq(data.vn, 'DATETIME')
    if dtcnt != 0:
        dtest=['0']                                  # initializing dtest array for EXTRACT_DATA call
        dtest = data.extract_data(dti[0], dtest, infowrite)  # get the dataset
        dtchk=len(dtest)
        dtvals=np.zeros((3,dtchk))-mint
        errorfound=0
    else:
        dti, dtcnt = where_eq(data.vn, 'DATETIME.START')
        dtsi, dtscnt = where_eq(data.vn, 'DATETIME.STOP')
        if dtcnt != 0 and dtscnt != 0:
            dtest=['0']                                 # initializing dtest array for EXTRACT_DATA call
            dtest = data.extract_data(dti[0],dtest,infowrite)  # get the dataset
            dtchk=len(dtest)
            dtest=['0']                                  # initializing dtest array for EXTRACT_DATA call
            dtest = data.extract_data(dtsi[0],dtest,infowrite) # return the dataset
            if dtchk == len(dtest):
                dtvals=np.zeros((3,dtchk))-mint
                errorfound=0

    vn = data.vn
    for vc in range(data.nvn):
        vnspl=vn[vc].split('_')
        vnspl=[vnspl,'x','x']                                      # ensures vnspl has a minimum of 3 values
        if vnspl[1] == 'AVK' or vnspl[2] == 'AVK':
            vavk[vc]=1 # identifies a dataset containing AVK data
        if (vn[vc] == 'DATETIME' or vn[vc] == 'DATETIME.START' or vn[vc] == 'DATETIME.STOP'):
            dtest=['0']                             # initializing dtest array for EXTRACT_DATA call
            dtest = data.extract_data(vc, dtest, infowrite) # return the dataset
            if errorfound == 0:
                dtnel=len(dtest)
                if vn[vc] == 'DATETIME.START':
                    if dtnel == dtchk:
                        dtvals[0,:]=np.float64(dtest)
                    elif dtnel == 1:
                        dtvals[0,0]=np.float64(dtest)
                    else:
                        errorfound=1
                        dtfv[0]=md.mv_dbl[2,vc]

                elif vn[vc] == 'DATETIME':
                    dtvals[1,:]=np.float64(dtest)
                    dtfv[1]=md.mv_dbl[2,vc]

                else:
                    if dtnel == dtchk:
                        dtvals[2,:] = np.float64(dtest)
                    elif dtnel == 1:
                        dtvals[2,dtchk-1] = np.float64(dtest)
                    else:
                        errorfound=1
                        dtfv[2]=md.mv_dbl[2,vc]



            dtest_64 = np.array([np.float64(k) for k in dtest])
            # gi, gcnt = where_eq[np.float64(k) for k in dtest, mv_dbl[2,vc]) # edit out fill values
            # IF gcnt != len(dtest):
            #minth=np.array( MIN(DOUBLE(dtest[gi]),MAX=maxth)
            minth = dtest_64[dtest_64 != md.mv_dbl[2,vc]].min()
            maxth = dtest_64[dtest_64 != md.mv_dbl[2,vc]].max()
            if minth < mint:
                mint = minth
            if maxth > maxt:
                maxt = maxth
            valfound = valfound+1

    if errorfound == 0 and valfound > 1:
        # Make all missing values equal to fill values
        for i in (0,1,2):
            mi, mcnt = where_eq(dtvals[i,:], np.float64(-90000.0))
            if mcnt != 0:
                dtvals[i,mi]=dtfv[i]
        #Check that DATETIME.START LE DATETIME LE DATETIME.STOP
        i=0
        while errorfound == 0 and i <= dtchk-1:
            test1=(dtvals[0,i] != dtfv[0]) and (dtvals[1,i] != dtfv[1])
            test2=(dtvals[1,i] != dtfv[1]) and (dtvals[2,i] != dtfv[2])
            test3=(dtvals[0,i] != dtfv[0]) and (dtvals[2,i] != dtfv[2])
            if (test1) and (dtvals[0,i] > dtvals[1,i]):
                errorfound=1
            if (test2) and (dtvals[1,i] > dtvals[2,i]):
                errorfound=2
            if (test3) and (dtvals[0,i] > dtvals[2,i]):
                errorfound=3
            i=i+1
            
        if errorfound != 0:
            v0="{:18.9f}".format(dtvals[0,i-1]).strip()
            v1="{:18.9f}".format(dtvals[0,i-1]).strip()
            v2="{:18.9f}".format(dtvals[0,i-1]).strip()
        
            if errorfound == 1:
                itxt=['DATETIME.START='+v0,'DATETIME='+v1]
            elif errorfound == 2:
                itxt=['DATETIME='+v1,'DATETIME.STOP='+v2]
            else:
                itxt=['DATETIME.START='+v0,'DATETIME.STOP='+v2]
            infotxt='3 '+itxt[0]+' is greater than '+itxt[1]
            infotxt_output(infotxt)

    valfound=0  # Boolean to indicate whether missing VAR_VALID_MIN/MAX values have been added

    # Add minimum and maximum values to Metadata as required
    di, dcnt = where_eq([i.upper() for i in data.vu],'MJD2K')
    if dcnt != 0:
        for vc in range(dcnt):
            mi, _ = where_eq(md.meta_arr, 'VAR_NAME='+vn[di[vc]])
            for i in range(nvchk):
                if md.meta_arr[mi[0]+mmi[i]:len(vchk[i])] != vchk[i]:
                    acti=mmi[i]-1
                else:
                    acti=mmi[i] # depends if VAR_NOTES is present or not
                res=md.meta_arr[mi[0]+acti].split('=')
                if len(res) == 1: # no value so add Min or Max value to this attribute
                    if np.mod(i, 2) == 0:
                        if mint != np.float64(90000.0):
                            md.meta_arr[mi[0]+acti]=vchk[i]+'=' + '{:18.9f}'.format(mint).strip()
                            valfound=1
                        md.mv_dbl[i,di[vc]]=mint
                    else:
                        if maxt != np.float64(-90000.0):
                            md.meta_arr[mi[0]+acti]=vchk[i]+'=' + '{:18.9f}'.format(mint).strip()
                            valfound=1
                        mdmv_dbl[i,di[vc]]=maxt
                    if valfound == 1:
                        infotxt=[]
                        infotxt.append('2 Missing '+vchk[i]+' value for '+md.meta_arr[mi[0]]+'| added based on available')
                        infotxt.append('    DATETIME[.START][.STOP] values: '+meta_arr[mi[0]+acti])
                        infotxt_output(infotxt)


def pre_defined_att_checks(ct, nav, vav=None, vnx=None, vdtv=None, verror=None):
    """
    Procedure to perform checks on the HDF4 pre-defined attributes, as well as check that
    the numeric variable attribute values are of the same data type as that given by the
    corresponding VAR_DATA_TYPE value (if input is via session memory)

    """

    if ct == -1:
        ncsa_str=['long_name','format','coordsys']
        ncsa_cal=['scale_factor','scale_factor_err','add_offset','add_offset_err','calibrated_nt']
        for i in range(len(ncsa_str)):
            ni, ncnt = where_eq([j.lower() for j in nav], ncsa_str[i])
            if ncnt != 0:
                if global_qa_yes:
                    itxt=' present in'
                else:
                    itxt=' will not be written to'
                infotxt='0 HDF4 pre-defined attribute '+ncsa_str[i]+itxt+' the HDF file'
                infotxt_output(infotxt)

        calfound=0
        for i in range(len(ncsa_cal)):
            if calfound == 0:
                ni, ncnt = where_eq([j.lower() for j in nav], ncsa_cal[i])
                if ncnt != 0:
                    calfound=1
                    infotxt='3 Datasets that have been scaled or had an offset applied are not permitted'
                    infotxt_output(infotxt)


def check_min_max_fill(data):
    """

    Procedure to check that the data, VAR_VALID_MIN/MAX, and VAR_FILL_VALUE fit within the given data
    type. That the VAR_FILL_VALUE falls within the GEOMS definition and that the data values are within
    the valid minimum and maximum values (or the fill value). The VAR_VALID_MIN value is less than or
    equal to the VAR_VALID_MAX value. The data values are returned in the data type given by VAR_DATA_TYPE.


    """

    proname='check_min_max_fill: '
    errtxt2=['']*(7)
    errtxt2[0]=' not valid, or does not match VAR_DATA_TYPE'
    errtxt2[1]=' data value outside VAR_VALID_'
    errtxt2[2]='Fill data value not outside VAR_VALID_MIN or VAR_VALID_MAX: '
    errtxt2[3]=' outside the data type range: '
    errtxt2[4]='DATETIME value cannot be a VAR_FILL_VALUE. '
    errtxt2[5]='DATETIME values not in chronological order'
    errtxt2[6]='Type conversion error.  Entry is not a valid value'

    # Loop over variables
    for i, (vn, val) in enumerate(data.sdss_d.items()):
        # Make sure values are correct data type
        f_np = dtype_to_np(data.ndm[i, 8])

        # Check if datatype is string
        if f_np == np.string_:
            check_string_datatype(vn, val)
            continue

        mnv = f_np(val.va_d['VAR_VALID_MIN'])
        mxv = f_np(val.va_d['VAR_VALID_MAX'])
        fv = f_np(val.va_d['VAR_FILL_VALUE'])

        # Check if VAR_FILL_VALUE falls within VAR_VALID_MIN/MAX - if so it is defined
        # as a default, rather than missing/error value
        if fv >= mnv and fv <= mxv:
            infotxt='0 Fill value for VAR_NAME=' + vn + ' falls within the data'
            infotxt=infotxt+' range and is defined as a ''Default'' value'
            infotxt_output(infotxt)

        # Do checks on VAR_VALID_MIN and MAX values for Latitude, Longitide, Zenith, Azimuth and Wind Direction datasets
        if 'LATITUDE' in vn:
            if mnv < -90.0 or mxv > 90.0:
                itxt='-90.0 (South) and +90.0 (North) degrees'
                infotxt='3 Valid Minimum and Maximum range values for '+ vn +' are ' + itxt
                infotxt_output(infotxt)

        elif 'LONGITUDE' in vn:
            if mnv < -180.0 or mxv > 180.0:
                itxt='-180.0 (West) and +180.0 (East) degrees'
                infotxt='3 Valid Minimum and Maximum range values for '+vn+' are '+itxt
                infotxt_output(infotxt)

        elif 'ZENITH' in vn and 'ANGLE' in vn:
            if mnv < 0.0 or mxv > 180.0:
                itxt='0.0 and 180.0 degrees'
                infotxt='3 Valid Minimum and Maximum range values for '+vn+' are '+itxt
                infotxt_output(infotxt)

        elif 'AZIMUTH' in vn and 'ANGLE' in vn:
            if mnv < 0.0 or mxv > 360.0:
                itxt='0.0 (North) and 360.0 degrees (East = 90 degrees and so on)'
                infotxt='3 Valid Minimum and Maximum range values for '+vn+' are '+itxt
                infotxt_output(infotxt)

        elif 'WIND.DIRECTION' in vn:
            if mnv != 0.0 or mxv != 360.0:
                itxt='0.0 (Calm) and 360.0 degrees (use WMO definition)'
                infotxt='3 Valid Minimum and Maximum range values for '+vn+' are '+itxt
                infotxt_output(infotxt)


        # Determine whether minimum and maximum data values are within VAR_VALID_MIN/MAX
        dtest = np.array(val.data, dtype=f_np)
        if np.where(dtest != fv)[0].shape[0] > 0: # i.e. there are non-fill values in the data
            minv = dtest[dtest != fv].min()
            maxv = dtest[dtest != fv].max()
            if minv < mnv:  # Minimum data value is less than VAR_VALID_MIN
                itxt=str(minv)
                infotxt='3 Minimum data value of ' + itxt + ' is outside VAR_VALID_MIN=' + str(val.va_d['VAR_VALID_MIN'])
                infotxt=infotxt+' for dataset ' + vn
                infotxt_output(infotxt)

            elif maxv > mxv: # Maximum data value is greater than VAR_VALID_MAX
                itxt=str(maxv)
                infotxt='3 Maximum data value of '+itxt+' is outside VAR_VALID_MAX=' + str(val.va_d['VAR_VALID_MAX'])
                infotxt=infotxt+' for dataset '+vn
                infotxt_output(infotxt)

        else:              # dataset consists of fill values so check that VAR_VALID_MIN LE VAR_VALID_MAX
            if mnv > mxv:
                infotxt='3 VAR_VALID_MIN value is greater than the VAR_VALID_MAX value for dataset '+vn
                infotxt_output(infotxt)

        # Do Checks with DATETIME attributes
        if 'DATETIME' in vn:   # i.e. DATETIME, DATETIME.START or DATETIME.STOP
        # Does DATETIME contain any fill values?
            dcnt = np.where(dtest == fv)[0].shape[0]
            if vn == 'DATETIME' and dcnt != 0:
                                # 'DTFVOK' = DATETIME fill value is OK
                pass
                #IF (o3[5] EQ 'DTFVOK') AND (dcnt NE ndl) THEN itxt='1 ' ELSE itxt='3 '
                #infotxt=itxt+'DATETIME contains fill value(s)'
                #INFOTXT_OUTPUT,infotxt

            # di=WHERE(dtest NE fv,dcnt)
            # IF dcnt GT 1 THEN BEGIN
            # Are DATETIME values in chronological order?
            if np.where(dtest != fv)[0].shape[0] > 0:
                dsort = np.sort(dtest)

                if not all(dsort == dtest):

        #    IF o3[5] EQ 'DTFVOK' THEN itxt='1 ' ELSE itxt='3 '
                    itxt='3 '
                    infotxt=itxt+vn+' values are not in chronological order'
                    infotxt_output(infotxt)




def check_string_datatype(var, data_val):
    """
    Procedure to check that the data, and Variable Attributes are correctly configured when the
    VAR_DATA_TYPE=STRING, including VAR_UNITS, VAR_VALID_MIN/MAX, and VAR_FILL_VALUE. The data
    values are returned in the string datatype.
    """

    # Check if text is left-justified
    # dtestx = [i.rstrip() for i in data_val.data]
    # dtest = [i.strip() for i in data_val.data]

    if 'MaskedArray' in type(data_val.data).__name__:
        if len(data_val.data.shape) == 1: # A single string
            data_val.data = data_val.data.tostring()
        else: # A list of strings
            data_val.data = [i.tostring() for i in data_val.data]

    # Text may be of type np.string_ or bytes
    try:
        dtestx = [i.rstrip() for i in data_val.data.decode()]
        dtest = [i.strip() for i in data_val.data.decode()]
    except AttributeError:
        dtestx = [i.rstrip() for i in data_val.data]
        dtest = [i.strip() for i in data_val.data]


    # format of the string dataset has been changed
    if dtestx != dtest:
        if global_qa_yes:
            infotxt='2 Dataset ' + var + ' incorrectly formatted (text must be left-justified)'
        else:
            infotxt='2 Dataset '+ var +' reformatted so that text is left-justified'
        infotxt_output(infotxt)

    label=['VAR_UNITS','VAR_SI_CONVERSION','VAR_VALID_MIN','VAR_VALID_MAX','VAR_FILL_VALUE']




def extract_and_test(dentry, cpos, nel, ta1, ta2, taname, aname, ti):
    """
    Procedure called by Check_Metadata to extract the relevant parts of tab_arr and test against
    a corresponding Metadata value.




      Inputs: dentry - the Metadata entry to be checked (everything after the '=')
              cpos - a pattern used by StrSplit on DEntry to split it into its component parts e.g. ';' or '_'
              nel - the number of sub-values in the set of tab_arr entries being checked (ta1 only)
              ta1 - the set of tab_arr values used to check against the Metadata contents
              ta2 - a second set of tab_arr values used for DATA_SOURCE checks, o/w set to ['']
              taname - a combination of the Metadata Attribute name and Table Field/Label used for the
                       error messages
              aname - the Metadata attribute name to be used for the error message
              ti - the index value of the ta2 values to be checked. Used for DATA_SOURCE checks only

      Output: ti - the index value(s) of tab_arr which match the Metadata value (used in Check_Metadata
                   to do further checks on the ORIGINATOR values)

      Called by: CHECK_METADATA

      Subroutines Called: STOP_WITH_ERROR (if error state detected)
        Possible Conditions for STOP_WITH_ERROR call (plus [line number] where called):
          1. Incorrect number of (sub-)values for a particular Metadata attribute
          2. Metadata value does not match any values in the Table Attribute Values file

    """

    proname='Check_Metadata/Extract_And_Test procedures: '
    errtxt2=['']*2
    errtxt2[0]='Number of (sub-)values for this attribute should be '
    errtxt2[1]='Doesn\'t match any Table Attribute Values under FIELD: '

    infoerr=['COUNTRY','AFFILIATION','FILE_ACCESS','FILE_META_VERSION','DATA_LOCATION',
             'DATA_DISCIPLINE','DATA_GROUP']  #do not stop the program for errors in these attributes



    # Check number of Metadata entry components is as expected
    achk=dentry.split(cpos)
    n_achk = len(achk)
    # Note: DATA_SOURCE can have 2 or 3 sub-values (nel=[2,3] from 4.0b53)
    # Is integer or list of 2 elements
    if np.where(np.array(nel) == n_achk)[0].size == 0:
        stop_with_error(proname+taname+'='+dentry+': ', errtxt2[0]+nel +'.')

    # Allow fix to be made to the affiliation if the acronym is correct but institute name is wrong
    n_ta1 = len(ta1)
    if 'AFFILIATION' in taname and cpos != '_':
        affchk=1
        acroval=achk[1].strip()
        t_acrovals=['']*n_ta1
    else:
        affchk=0

    # Check Metadata entry with relevant TAV entry or entries
    if cpos == '_':     # DATA_SOURCE entry so need to treat components separately
        #achk=''.join(achk.split())
        achk=[i.strip() for i in achk]
        nel=1
        achkh='   '+achk[1]

        # Strip the last three characters (instrument ID) off achk(1)
        achk[1]=achk[1][0:len(achk[1])-3]
        # Check that the instrument ID is valid
        instidchk=achkh[len(achkh)-3 : len(achkh)]

        if not instidchk.isnumeric():
            infotxt='3 DATA_SOURCE 3-digit instrument identifier missing or invalid'
            infotxt_output(infotxt)
    else:
        achk=''.join(dentry.split())

    for j in range(n_ta1):
        # Extract valid portion of the entries (depending on ATTRIBUTE name)
        tchk=ta1[j].split(';')
        ta1[j]=''.join(tchk[0].split())
        if nel > 1:
            for k in range(1,nel):
                ta1[j]=ta1[j]+';'+''.join(tchk[k].split())
        if affchk == 1:
            t_acrovals[j]=tchk[1]

    if ta2[0] != '':
        for j in range(len(ta2)):
            # Extract valid portion of the entries (depending on ATTRIBUTE name)
            tchk=ta2[j].split(';')
            ta2[j]=''.join(tchk[ti].split())

    if cpos == '_':
        gi, gcnt = where_eq(ta1, achk[0])
        if gcnt != 0:
            gi, gcnt = where_eq(ta2, achk[1])
        # Check for Case Sensitive or incorrect apostrophe problem
    else:
        gi, gcnt = where_eq(ta1, achk)
        if gcnt == 0:
#Check for Case Sensitivity and/or incorrect apostrophe usage or correct affiliation acronym only
#Will be fixed in CHECK_METADATA
            gi, gcnt = where_eq([i.lower for i in ta1], achk.lower())
            if gcnt != 0:
                dentry=ta1[gi[0]]
            else:
#Check for "`" instead of "'" in metadata entry
                if '`' in achk:
                    achkh=achk
                    achkh.replace('`', '\'' )
                    gi, gcnt = where_eq([i.lower for i in ta1], achkh.lower())
                    if gcnt != 0:
                        dentry=ta1[gi[0]]

                # If AFFILIATION check to see if the acronym only is valid
                if affchk == 1:
                    #gi, gcnt =where_eq([i.lower() for i in acroval], t_acrovals.lower())
                    gi, gcnt =where_eq([i.lower() for i in t_acrovals], acroval.lower())
                    if gcnt != 0:
                        dentry=ta1[gi[0]]


    if gcnt == 0:
        ei = -1
        for i in range(len(infoerr)):
            if infoerr[i] in taname:
                ei=i

        if ei != -1:     # Log message instead of STOP_WITH_ERROR
            if ei <= 1:
                metalabel=taname.split('/')[1]
            else:
                metalabel=infoerr[ei]
            infotxt='3 '+metalabel+'='+aname+' doesn\'t match any '
            infotxt=infotxt+'Table Attribute Values under '+infoerr[ei]+' field'
            infotxt_output(infotxt)
        else:
            stop_with_error(proname+aname+': ',errtxt2[1]+taname)
    ti=gi   # ti returned to the Check_MetaData procedure to check AFFILIATION, ADDRESS and EMAIL values
    return ti



def check_metadata(md, td, data):
    """
    Procedure to check the validity of Metadata values by checking them against the corresponding
    entries in the Table Attribute Values File

    """

    #meta_arr = md.meta_arr

    # Possible error messages for this procedure
    proname='Check_Metadata procedure: '
    errtxt=[]
    errtxt.append('No or Invalid value for this Attribute: ')
    errtxt.append('Number of (sub-)values for this Attribute should be ')
    errtxt.append('AFFILIATION value doesn\'t match NAME value: ')
    errtxt.append(' field not found in the input Table Attribute Values file')

    # Originator values to be checked
    orig_attr=['PI_','DO_','DS_','NAME','AFFILIATION','ADDRESS','EMAIL']
    # Special cases for initial capitalization of the COUNTRY value
    countrysp1=['Congo, The Democratic Republic of the','Cote d''Ivoire']
    # Special cases for initial capitalization of the COUNTRY value (individual words)
    countrysp2=['the','and','of']

    # initialize variables for VAR_UNITS/VAR_SI_CONVERSION checks
    firstvu=0
    tchk=''
    rd=1
    writeonce=0
    xval=[2,2,3,1]              # expected number of sub-values for the four originator attributes
    new_name=['']               # Entry which will contain any ORIGINATOR name values not present in the TAV file


    for i, mv in enumerate(md.meta_arr):
        res = md.meta_arr[i].split('=')
        nres = len(res)
        if res[0] == 'VAR_NAME':
            vnval=md.meta_arr[i]
        for j in range(3):
            for k in range(3,7):
                # need to change ORIGINATOR NAME ATTRIBUTE names to match TAV File
                if res[0] == orig_attr[j]+orig_attr[k]:
                    res[0]='ORIGINATOR'
                    jh=j
                    kh=k

        # Test whether metadata label is also a TAV field
        # First test checks for exact match between metadata label and TAV field
        # (This avoids possible problem using STRPOS (e.g. DATA_TYPE won't also pick up VAR_DATA_TYPE))
        #m_in_tav = res[0] in td.tab_arr.keys()


        # Back-up test in the event that label has more than one TAV field (e.g. DATA_VARIABLES)
        if True:
            # Store keys that matches search
            mi = []
            #for i, j in enumerate(sorted(list(td.tab_arr.keys()))):
            for j in td.tab_arr:
                re_ = re.search(res[0], j)
                if re_ is not None:
                    if re_.start() == 0:
                        mi.append(j)
            mcnt = len(mi)
        #

        if res[0] == 'ORIGINATOR':
            # Do checks on _NAME, _AFFILIATION, _ADDRESS, and _EMAIL sub-values
            resx = re.split('=|;', md.meta_arr[i])
            nel=len(resx)-1


            if kh == 4 and nel == xval[1]: # check on AFFILIATION value
                test_key = 'AFFILIATION'
                resy=md.meta_arr[i].split('=')
                if test_key in td.tab_arr.keys():
                    ta1=td.tab_arr[test_key] # subset of relevant FIELD ENTRIES
                    taname=test_key+'/'+resy[0]      # name combination for error message
                    dentry=resy[1] ; ti=-1
                    ti=extract_and_test(dentry,';',xval[1],ta1,[''],taname,resy[1],ti)
                    if len(data.rerr[0]) > 2:
                        return
                    if dentry != resy[1]:
# Case and/or apostrophe differences between Metadata entry and TAV entry
                        achk2=td.tab_arr[test_key][ti[0]].split(';')
                        achk=achk2[0].strip()
                        for m in range(1, xval[1]):
                            achk=achk+';' + achk2[m].strip()
                        if data.qa_yes:
                            itxt=' should be '
                        else:
                            itxt=' replaced with '
                        infotxt='2 ' + md.meta_arr[i]+itxt+resy[0]+'='+achk+' based on the TAV entry'
                        infotxt_output(infotxt)
                        md.meta_arr[i]=resy[0]+'='+achk


            if kh == 5 and nel == xval[2]: # check on COUNTRY value
                test_key = 'COUNTRY'
                if test_key in td.tab_arr.keys():
# Check the COUNTRY value against the TAV values (note: can be all uppercase or initial capitalization)
                    aname=res[1]
                    ti = -1
                    ta1=td.tab_arr[test_key]      # subset of relevant FIELD ENTRIES
                    taname=test_key+'/'+orig_attr[jh]+'ADDRESS' # name combination for error message
                    ti=extract_and_test(resx[nel].upper(),';',1,ta1,[''],taname,aname,ti)
                    if len(data.rerr[0]) > 2:
                        return
                    #Test for upper case or initial capitalization
                    if resx[nel] != resx[nel].upper():
                        counterr=0     # error code to detect problem with initial capitalization

                        sci, spcnt = where_eq(countrysp1, resx[nel].upper())
                        if spcnt != 0: # Test special case countries e.g. Cote d'Ivoire
                            if resx[nel] != countrysp1[sci[0]]:
                                counterr=1
                            else: # Test individual words in the country value
                                sepcount=re.split((' |,'), resx[nel])
                                spcnt = len(sepcount)
                                for j in range(spcnt):
                                    slen=len(sepcount[j])
                                    sci=where_eq(countrysp2, sepcount[j])
                                    if len(sci)!= 0:
                                        # special case lowercase word e.g. the, and, of
                                        test=countrysp2[sci[0]]
                                    elif slen == 1:
                                        test=sepcount[j].upper() # only one letter in the word so make uppercase
                                    else:
                                        test=sepcount[j].capitalize() # capitalize the first letter
                                    if sepcount[j] != test:
                                        counterr=1

                        if counterr == 1:
                            uccountry=resx[nel].upper()
                            if data.qa_yes:
                                itxt=' should be '
                            else:
                                itxt=' replaced with '
                            infotxt='2 COUNTRY value for ' + md.meta_arr[i]+itxt+uccountry
                            infotxt_output(infotxt)
                            pdb.set_trace()
                            md.meta_arr[i] = ';'.join(md.meta_arr[i].split(';')[:-1]) + ';'+resx[nel]

            if kh == 6 and nel == xval[3]: # check @ present in the e-mail value
                amppos=re.search('@', res[1])
                if amppos.start() == 0 or amppos.end() >= amppos.endpos:
                    infotxt='3 Invalid format for '+orig_attr[jh]+'EMAIL value'
                    infotxt_output(infotxt)

            if nel != xval[kh-3] or (md.meta_arr[i][-1] == ';'):
                if xval[kh-3] > 1:
                    itxt='sub-'
                else:
                    itxt=''
                infotxt='3 Number of '+itxt+'values for attribute ' + md.meta_arr[i]+' should be '+ str(xval[kh-3])
                infotxt_output(infotxt)
            mcnt=0


        if mcnt != 0:    # check Metadata value(s) against TAV file entries
            if nres != 2 and rd != 0:
                stop_with_error(proname,errtxt[0] + md.meta_arr[i])
            elif res[0] == 'DATA_SOURCE':

                #need to test against 'DATA_SOURCE' and 'AFFILIATION' (AVDC) or 'DATA_SOURCE_02'
                #(original Envisat) entries
                ta1=td.tab_arr[mi[0]] # DATA_SOURCE ENTRIES

                if td.tab_type == 0:
                    atxt='AFFILIATION'
                    atx=''
                    ti=1
                else:
                    atxt='DATA_SOURCE_02'
                    atx='_01'
                    ti=0

                #ai=WHERE(STRPOS(tab_arr[*,0],atxt) NE -1,acnt)
                ai=atxt in td.tab_arr
                if not ai:
                    stop_with_error(proname,atxt+errtxt[3])

                ta2=td.tab_arr[atxt] # AFFILIATION/DATA_SOURCE_02 ENTRIES
                taname=res[0]+atx+'/'+atxt  # name combination for error message
                n_sv=[2,3]                  # possible number of DATA_SOURCE sub-values
                ti=extract_and_test(res[1],'_',n_sv,ta1,ta2,taname,res[1],ti)
                if len(data.rerr[0]) > 2:
                    return


                achk=res[1].split('_') # COUNT=n_achk
                if len(achk) == 3:
                    # Check DATA_VERSION_NAME if there are 3 sub-values in DATA_SOURCE (introduced v4.0b53)
                    geoms_rule_changes_11(achk[2])

            elif res[0] == 'DATA_VARIABLES':
                # make into 2-D array and separate out the components of the VAR_NAMES
                achk=res[1].split(';')
                nel=len(achk)
                achk2=np.zeros((mcnt,nel),dtype=object)
                achk2[:,:] = ''
                for j in range(nel):
                    dv=achk[j].split('_')
                    n_dv = len(dv)
                    if n_dv > mcnt: # invalid VAR_NAME composition
                        stop_with_error(proname,errtxt[0]+res[0]+'='+achk[j])
                    achk2[0:n_dv, j]=dv


                for j in range(nel):
                    dvcnt = np.where(achk2[:,j] != '')[0].size
                    for k in range(mcnt):
                        if achk2[k,j] != '':
                            taname=mi[k]
                            # subset of TAV DATA_VARIABLES entries
                            ta1=td.tab_arr[mi[k]]
                            if (k != 0) and (dvcnt == 2):
                                # Need to check both variable mode and descriptor options
                                for m in range(k+1, mcnt):
                                    taname=taname+'/'+mi[m]
                                    ta1=ta1 + td.tab_arr[mi[m]]
                            ti=extract_and_test(achk2[k,j],';',1,ta1,[''],taname,achk[j],0)
                            if len(data.rerr[0]) > 2:
                                pdb.set_trace()
                                return


            elif res[0] == 'FILE_ACCESS':
                # need to check each Metadata entry individually
                achk=res[1].split(';')
                nel=len(achk)
                ta1=td.tab_arr[mi[0]] # FILE_ACCESS entries
                for j in range(nel):
                    ti=extract_and_test(achk[j],';',1,ta1,[''],res[0],achk[j],0)
                    # IF STRLEN(rerr[0]) GT 2 THEN RETURN

            elif res[0] == 'FILE_META_VERSION':
                # only need to look at the second value
                achk=res[1].split(';')
                if len(achk) != 2:
                    achk=[achk,'x','x']  # ensure there are at least 2 values
                    achk=[achk[1].strip()]
                    ta1=td.tab_arr[mi[0]] # FILE_META_VERSION entries
                    ti=extract_and_test(achk[0],';',1,ta1,[''],res[0],achk[0],0)
                    # IF STRLEN(rerr[0]) GT 2 THEN RETURN

            elif res[0] == 'VAR_UNITS':
                # VAR_UNITS and VAR_SI_CONVERSION entries (not needed for STRING data type)
                if rd != 0:
                    if res[1].upper() == 'MJD2000': # change to MJD2K
                        in1=md.meta_arr[i]
                        in2=res[1]
                        geoms_rule_changes_3(in1,in2,writeonce)
                        md.meta_arr[i]=in1
                        res[1]=in2
                        writeonce=1

                    datetimechk='DATETIME' in vnval
                    #tchk, errstate = var_units_test(res[1], rd, td.tab_type, tchk, errstate)
                    tchk, errstate = var_units_test(data, res[1], rd, td.tab_type)

                    if errstate != '':
                        # VAR_UNIT value not valid
                        if not datetimechk:
                            itxt='. VAR_SI_CONVERSION value not checked'
                        else:
                            itxt=''
                        infotxt='3 '+errstate+itxt
                        infotxt_output(infotxt)


                    # Find VAR_SI_CONVERSION attribute
                    k=1
                    while md.meta_arr[i+k][0:17] != 'VAR_SI_CONVERSION':
                        k += 1
                    achk=md.meta_arr[i+k].split('=') # VAR_SI_CONVERSION values
                    if len(achk) == 1:
                        achk.append(' ')
                    if datetimechk:
                    # Check if VAR_SI_CONVERSION values are 0.0;86400.0;s regardless of VAR_UNITS.
                    # If so, then don't change
                        achk1=re.split(' |;', achk[1])
                        achk1.extend(['','','']) # Ensure there are at least 3 values extracted
                        #pdb.set_trace()
                        if is_number(achk1[0]) and is_number(achk1[1]):
                            if int(float(achk1[0])) == 0 and int(float(achk1[1])) == 86400 and \
                               achk1[2].lower() == 's':
                                tchk='0.0;86400.0;s'
                        elif errstate != '':
                            tchk='0.0;86400.0;s'

                    # Check calculated VAR_SI_CONVERSION value against existing value in the metadata
                    if errstate == '' or datetimechk:
                        if achk[1] != tchk:
                            if data.qa_yes:
                                itxt=' should be '
                            else:
                                itxt=' replaced with '
                            infotxt='2 ' + md.meta_arr[i+k]+itxt+'VAR_SI_CONVERSION='+tchk+' for '+vnval
                            infotxt_output(infotxt)
                            #pdb.set_trace()
                        md.meta_arr[i+k]=achk[0]+'='+tchk

                    # Perform checks on the remaining ATTRIBUTE_NAMES
            else:
                achk=[st.strip() for st in res[1].split(';')]
                nel=len(achk) ; ta2=['']
                if res[0] == 'VAR_DATA_TYPE':
# To convert VAR_SI_CONVERSION values to floating point if necessary
                    if res[1].upper() == 'REAL' or res[1].upper() == 'DOUBLE':
                        rd=-1
                    elif res[1].upper() == 'STRING':
                        rd=0
                    else:
                        rd=1
                    if achk[0].upper() == 'LONG':
                        achk[0]='INTEGER'
                    #Changed to INTEGER for EXTRACT_AND_TEST only - actual change occurs in SET_UP_STRUCTURE

                if nel != mcnt or md.meta_arr[i][-1] == ';':
                    if mcnt > 1:
                        itxt='sub-'
                    else:
                        itxt=''
                    infotxt='3 Number of '+itxt+'values for attribute '+md.meta_arr[i]+' should be ' \
                        + str(mcnt)
                    infotxt_output(infotxt)

                nel=1
                for j in range(mcnt):
#                    pdb.set_trace()
                    ta1=td.tab_arr[mi[j]] # subset of relevant ENTRIES
                    dentry=achk[j]
                    ti=mi[j]
                    ti=extract_and_test(dentry, ';',nel,ta1,ta2,mi[j],achk[j],ti)
                    # if len(data.rerr[0]) > 2:
                    #     return
                    if mcnt == 1 and dentry != achk[j]:
                    #Case and/or apostrophe differences between Metadata entry and TAV entry
                        achk2=td.tab_arr[mi[j]][ti[0]+2].split(';')
                        achk=achk2[0].strip()
                        if data.qa_yes:
                            itxt=' should be '
                        else:
                            itxt=' replaced with '
                        infotxt='2 '+md.meta_arr[i]+itxt+res[0]+'='+achk+' based on the TAV entry'
                        infotxt_output(infotxt)
                        md.meta_arr[i]=res[0]+'='+achk




def avdc_hdf_write(hdffilename, md, data):
    """
    TODO

    """

    pass


def dtype_to_np(dt):
    """
    Return numpy class corresponding to the data type

    """

    if dt == 0:
        return np.uint8
    elif dt == 1:
        return np.int16
    elif dt == 2:
        return np.int32
    elif dt == 3:
        return np.int32
    elif dt == 4:
        return np.float32
    elif dt == 5:
        return np.float64
    elif dt == 6:
        return np.string_

    # if dt == 1:
    #     return np.uint8
    # elif dt == 2:
    #     return np.int16
    # elif dt == 3:
    #     return np.int32
    # elif dt == 4:
    #     return np.float32
    # elif dt == 5:
    #     return np.float64
    # elif dt == 7:
    #     return np.string_


    else:
        print("ERROR: wrong input in function.\nExiting.")
        sys.exit()


def avdc_hdf5_write(hdffilename, md, data):
    """
    Write global attributes & datasets to the Root
    Group in an HDF5 file.

      Inputs: hdffilename - a string holding the name of the HDF file being created
                            & written to
              md - Metadata object containing the Global & Variable Attributes
              data - Data object

      Outputs: An HDF5 file with the name hdffilename

    """

    import h5py

    nvn = data.nvn

    # Create an HDF5 file and its root group
    sd_id = h5py.File(os.path.join(data.out_dir, hdffilename), "w")

    # Write the Global Attributes to file
    # Section 4.1 (Bojkov et al., 2002): Originator attributes
    # Section 4.2 (Bojkov et al., 2002): Dataset attributes
    # Section 4.3 (Bojkov et al., 2002): File attributes
    for i in range(len(md.attr_arr_glob)):
        # separate out the Meta Attribute entry
        eq_split=md.meta_arr[i].split('=', maxsplit=1)
        attribute_value=eq_split[1] if eq_split[1] != '' else ' '
        #attribute_value=eq_split[1] if eq_split[1] != '' else ''

        # Use ASCII encoding. Without this instruction it will be stored as UTF-8 instead
        attribute_value = attribute_value.encode('ascii')

        #sd_id.attrs.create(md.attr_arr_glob[i], attribute_value)

        sd_id.attrs.create(eq_split[0], attribute_value)


    # Note: code below follows IDL routine avdc_hdf_write, which performs some common tasks for
    # different output formats
    natts = len(md.attr_arr_data)

    # List of dim. attributes to be assigned to the SDS dimensions.
    da_list=['VAR_NAME', 'VAR_UNITS']
    da_att=[str(j) for i,j in enumerate(da_list)]
    av=np.zeros((nvn, natts), dtype=object)       # array containing attribute and NCSA values for each dataset
    var_depend=['']*nvn     #holding array for the VAR_DEPEND attributes, used to assign DIM information
    var_size= ['']*nvn       #holding array for the VAR_SIZE attributes, used to assign netCDF DIM information


    # Write all Dataset Attribute values to an array
    for i in range(nvn):
        mi, _ = where_eq(md.meta_arr, md.attr_arr_data[0] + '=' + data.vn[i])
        ninc=0
        j=0
        while j <= natts-1:
            ma=mi[0]+ninc
            # separate out the Meta entry into two components
            # res[0]: attribute name; res[1]: attribute value
            res=md.meta_arr[ma].split('=', maxsplit=1)
            while res[0] != md.attr_arr_data[j]: # in case optional variable attributes are not included
                j=j+1

            av[i,j] = ' ' if len(res[1]) == 0 else res[1]
            #av[i,j] = '' if len(res[1]) == 0 else res[1]

            # Reverse Values for VAR_DEPEND and VAR_SIZE to match order values will be saved in HDF
            if res[0] == 'VAR_DEPEND' or res[0] == 'VAR_SIZE':
                atval=re.split(' |;', av[i,j])
                avcnt = len(atval)
                if avcnt > 1:
                    atval = atval[::-1]
                    for k in range(avcnt):
                        if k == 0:
                            ahold = atval[k]
                        else:
                            ahold = ahold + ';' + atval[k]

                    av[i,j]=ahold
                if res == 'VAR_DEPEND':
                    var_depend[i]=av[i,j]
                else:
                    var_size[i]=av[i,j]

            ninc=ninc+1
            j=j+1


    # Write the Datasets (SDS) to file
    # Section 5.1 (Bojkov et al, 2002): Variable description attributes.

    # Loop over datasets.
    #for i in range(nvn):
    i=-1
    for var_name, var_val in data.data.items():
        i += 1

        if data.ndm[i,8] == 0: #identifies the VAR_DATA_TYPE
            valid_range=[np.uint8(md.mv_lng[0,i]), np.uint8(md.mv_lng[1,i])]
            fill_value=np.uint8(md.mv_lng[2,i])
            dtype_id = np.dtype(np.uint8)
            var_val_file = np.array([np.uint8(i) for i in var_val])
        elif data.ndm[i,8] == 1:
            valid_range=[int(md.mv_lng[0,i]), int(md.mv_lng[1,i])]
            fill_value=md.mv_lng[2,i]
            var_val_file = np.array([np.int16(i) for i in var_val])
            dtype_id = np.dtype(np.int16)
        elif data.ndm[i,8] == 2:
            # valid_range=[int(md.mv_lng[0,i]), int(md.mv_lng[1,i])]
            # fill_value=np.int(mv_lng[2,i])
            # var_val_file = np.array([int(i) for i in var_val])
            # dtype_id = np.dtype(int)
            valid_range=[np.int32(md.mv_lng[0,i]), np.int32(md.mv_lng[1,i])]
            fill_value=np.int32(md.mv_lng[2,i])
            var_val_file = np.array([np.int32(i) for i in var_val])
            dtype_id = np.dtype(np.int32)
        elif data.ndm[i,8] == 3:
            valid_range=[md.mv_lng[0,i], md.mv_lng[1,i]]
            fill_value=md.mv_lng[2,i]
            # var_val_file = np.array([np.int64(i) for i in var_val])
            # dtype_id = np.dtype(int64)
            var_val_file = np.array([np.int32(i) for i in var_val])
            dtype_id = np.dtype(np.int32)
        elif data.ndm[i,8] == 4:
            valid_range=[np.float32(md.mv_dbl[0,i]), np.float32(md.mv_dbl[1,i])]
            fill_value=np.float32(md.mv_dbl[2,i])
            var_val_file = np.array([np.float32(i) for i in var_val])
            dtype_id = np.dtype(np.float32)
        elif data.ndm[i,8] == 5:
            valid_range=[md.mv_dbl[0,i],md.mv_dbl[1,i]]
            fill_value=md.mv_dbl[2,i]
            var_val_file = np.array([np.float64(i) for i in var_val])
            dtype_id = np.dtype(np.float64)
        elif data.ndm[i,8] == 6:
            valid_range=[' ',' ']
            fill_value=' '
            var_val_file = np.array([i.strip() for i in var_val], dtype='S')
            #var_val_file = var_val[0].encode('ascii')
            dtype_id=var_val_file.dtype
            #dtype_id=h5py.string_dtype('ascii')


        # gi=WHERE(ndm[i,0:7] NE 0L) & vs=TRANSPOSE(ndm[i,gi])
        # If more than one dimension, reshape data array accordingly
        if data.ndm[i, 1] != 0:
            dims = [i for i in data.ndm[i, 0:8] if i != 0]
            # Reshape
            var_val_file = var_val_file.reshape(dims)



        # Store 1-element arrays as scalars
        if len(var_val_file) == 1:
            dset_id = sd_id.create_dataset(var_name, data=var_val_file[0], dtype=dtype_id)
        else:
            dset_id = sd_id.create_dataset(var_name, data=var_val_file, dtype=dtype_id)


        #         #Make up data array for writing to HDF5 file
        #         data=*ds[i].data
        #         #may have lost array VAR_SIZE information if the last value is '1' e.g. 41,1
        #         gi=WHERE(ndm[i,0:7] NE 0L)
        #         vs=TRANSPOSE(ndm[i,gi])
        #         #transpose to get array values in a vector
        #         #get dataset type & space, needed to create the dataset
        #         dtype_id=H5T_IDL_CREATE(data)
        #         dspace_id=H5S_CREATE_SIMPLE(vs)
        #         #create dataset in the output file
        #         IF comp_fac EQ 0 THEN dset_id=H5D_CREATE(sd_id,vn[i],dtype_id,dspace_id)
        #         ELSE BEGIN
        #             #Perform compression & shuffling of the dataset
        #             dset_id=H5D_CREATE(sd_id,vn[i],dtype_id,dspace_id,CHUNK_DIMENSIONS=vs,GZIP=comp_fac,/SHUFFLE)
        #         ENDELSE
        #         #write data to dataset - for string datasets remove unwanted white space
        #         IF ndm[i,8] EQ 6 THEN H5D_WRITE,dset_id,STRTRIM(data,2) ELSE H5D_WRITE,dset_id,data

        # Write out variable attributes
        ninc=0
        j=0
        mi, _ = where_eq(md.meta_arr, md.attr_arr_data[0]+'=' + data.vn[i])
        while j <= natts-1:
            ma=mi[0]+ninc

            #Extract Attribute Label
            res=md.meta_arr[ma].split('=')
            while res[0] != md.attr_arr_data[j]: #al[j]:
                j=j+1 #in case of missing optional variable attributes

            if md.attr_arr_data[j] == 'VAR_VALID_MIN':
                athold=valid_range[0]
            elif md.attr_arr_data[j] == 'VAR_VALID_MAX':
                athold=valid_range[1]
            elif md.attr_arr_data[j] == 'VAR_FILL_VALUE':
                athold=fill_value
            else:
                athold=av[i,j]


            # create attribute in the output file
            # use ASCII encoding for strings
            if type(athold).__name__ == 'str':
                athold = athold.encode('ascii')

            dset_id.attrs.update({md.attr_arr_data[j] : athold})
            j=j+1
            ninc += 1


    sd_id.close()


def avdc_nc4_write(hdffilename, md, data):
    """
    Write to netCDF4 format

    """

    from netCDF4 import Dataset

    nvn = data.nvn

    sd_id = Dataset(os.path.join(data.out_dir, hdffilename), "w")

    # Write the Global Attributes to file
    # Section 4.1 (Bojkov et al., 2002): Originator attributes
    # Section 4.2 (Bojkov et al., 2002): Dataset attributes
    # Section 4.3 (Bojkov et al., 2002): File attributes
    for i in range(len(md.attr_arr_glob)):
        # separate out the Meta Attribute entry
        eq_split=md.meta_arr[i].split('=', maxsplit=1)
        attribute_value=eq_split[1] if eq_split[1] != '' else ' '

        # Use ASCII encoding.
        attribute_value = attribute_value.encode('ascii')

        setattr(sd_id, eq_split[0], attribute_value)


    # Note: code below follows IDL routine avdc_hdf_write, which performs some common tasks for
    # different output formats
    natts = len(md.attr_arr_data)

    # List of dim. attributes to be assigned to the SDS dimensions.
    da_list=['VAR_NAME', 'VAR_UNITS']
    da_att=[str(j) for j in da_list]
    av=np.zeros((nvn, natts), dtype=object)       # array containing attribute and NCSA values for each dataset
    var_depend=['']*nvn     #holding array for the VAR_DEPEND attributes, used to assign DIM information
    var_size= ['']*nvn       #holding array for the VAR_SIZE attributes, used to assign netCDF DIM information


    # Write all Dataset Attribute values to an array
    for i in range(nvn):
        mi, _ = where_eq(md.meta_arr, md.attr_arr_data[0] + '=' + data.vn[i])
        ninc=0
        j=0
        while j <= natts-1:
            ma=mi[0]+ninc
            # separate out the Meta entry into two components
            # res[0]: attribute name; res[1]: attribute value
            res=md.meta_arr[ma].split('=', maxsplit=1)
            while res[0] != md.attr_arr_data[j]: # in case optional variable attributes are not included
                j=j+1

            av[i,j] = ' ' if len(res[1]) == 0 else res[1]
            #av[i,j] = '' if len(res[1]) == 0 else res[1]

            # Reverse Values for VAR_DEPEND and VAR_SIZE to match order values will be saved in HDF
            if res[0] == 'VAR_DEPEND' or res[0] == 'VAR_SIZE':
                atval=re.split(' |;', av[i,j])
                avcnt = len(atval)
                if avcnt > 1:
                    atval = atval[::-1]
                    for k in range(avcnt):
                        if k == 0:
                            ahold = atval[k]
                        else:
                            ahold = ahold + ';' + atval[k]

                    av[i,j]=ahold
                if res == 'VAR_DEPEND':
                    var_depend[i]=av[i,j]
                else:
                    var_size[i]=av[i,j]

            ninc=ninc+1
            j=j+1


    # Write the Datasets (SDS) to file
    # Section 5.1 (Bojkov et al, 2002): Variable description attributes.


    # Create dimensions id needed and determine dimensions for the dataset
    for i, (var_name, var_val) in enumerate(data.sdss_d.items()):
        current_dim = None
        # Variable depends on itself so it is a dimension
        # pdb.set_trace()
        #if k == v.va_d['VAR_DEPEND']:
        # Some cases can have ALTITUDE with VAR_DEPEND 'DATETIME;ALTITUDE' for example
        #if k in re.split(' |;', v.va_d['VAR_DEPEND']):
        if data.ndm[i,8] == 6: # need to include string length dimension variable
            mxstrlen = len(max(var_val.data, key = len))
            sd_id.createDimension(var_name + '_STRLEN', mxstrlen)
        dim_idx, n_dim = where_eq(re.split(' |;', var_val.va_d['VAR_DEPEND']), var_name)
        if n_dim > 0:
            # Variable depends on itself so it is an axis variable
            # sd_id.createDimension(k, int(v.va_d['VAR_SIZE']))
            dim_len = int(re.split(' |;', var_val.va_d['VAR_SIZE'])[dim_idx[0]])
            sd_id.createDimension(var_name, dim_len)

        current_dim = []

        # Add to list of dimensions for this variable
        if var_val.va_d['VAR_DEPEND'] != 'CONSTANT' and var_val.va_d['VAR_DEPEND'] != 'INDEPENDENT':
            for dim in re.split(' |;', var_val.va_d['VAR_DEPEND']):
                current_dim.append(dim)

        # Add string length to list of dimensions
        if data.ndm[i,8] == 6:
            current_dim.append(var_name + '_STRLEN')





    #pdb.set_trace()

    # Loop over datasets.
    #i=-1
    #for var_name, var_val in data.data.items():
    #for var_name, var_val in data.sdss_d.items():
        # i += 1

        # TODO: all data are stored as lists of string, change this somewhere else.
        # For now converting here
        if data.ndm[i,8] == 0: #identifies the VAR_DATA_TYPE
            valid_range=[np.uint8(md.mv_lng[0,i]), np.uint8(md.mv_lng[1,i])]
            fill_value=np.uint8(md.mv_lng[2,i])
            dtype_id = np.dtype(np.uint8)
            var_val_file = np.array([np.uint8(i) for i in var_val.data])
        elif data.ndm[i,8] == 1:
            valid_range=[int(md.mv_lng[0,i]), int(md.mv_lng[1,i])]
            fill_value=md.mv_lng[2,i]
            var_val_file = np.array([np.int16(i) for i in var_val.data])
            dtype_id = np.dtype(np.int16)
        elif data.ndm[i,8] == 2:
            # valid_range=[int(md.mv_lng[0,i]), int(md.mv_lng[1,i])]
            # fill_value=np.int(mv_lng[2,i])
            # var_val_file = np.array([int(i) for i in var_val.data])
            # dtype_id = np.dtype(int)
            valid_range=[np.int32(md.mv_lng[0,i]), np.int32(md.mv_lng[1,i])]
            fill_value=np.int32(md.mv_lng[2,i])
            var_val_file = np.array([np.int32(i) for i in var_val.data])
            dtype_id = np.dtype(np.int32)
        elif data.ndm[i,8] == 3:
            valid_range=[md.mv_lng[0,i], md.mv_lng[1,i]]
            fill_value=md.mv_lng[2,i]
            # var_val_file = np.array([np.int64(i) for i in var_val.data])
            # dtype_id = np.dtype(int64)
            var_val_file = np.array([np.int32(i) for i in var_val.data])
            dtype_id = np.dtype(np.int32)
        elif data.ndm[i,8] == 4:
            valid_range=[np.float32(md.mv_dbl[0,i]), np.float32(md.mv_dbl[1,i])]
            fill_value=np.float32(md.mv_dbl[2,i])
            var_val_file = np.array([np.float32(i) for i in var_val.data])
            dtype_id = np.dtype(np.float32)
        elif data.ndm[i,8] == 5:
            valid_range=[md.mv_dbl[0,i],md.mv_dbl[1,i]]
            fill_value=md.mv_dbl[2,i]
            var_val_file = np.array([np.float64(i) for i in var_val.data])
            dtype_id = np.dtype(np.float64)
        elif data.ndm[i,8] == 6:
            valid_range=[' ',' ']
            fill_value=' '
            var_val_file = np.array([i.strip() for i in var_val.data], dtype='S')
            #var_val_file = var_val[0].encode('ascii')
            dtype_id=var_val_file.dtype
            #dtype_id=h5py.string_dtype('ascii')






        # gi=WHERE(ndm[i,0:7] NE 0L) & vs=TRANSPOSE(ndm[i,gi])
        # If more than one dimension, reshape data array accordingly
        if data.ndm[i, 1] != 0:
            dims = [i for i in data.ndm[i, 0:8] if i != 0]
            # Reshape
            var_val_file = var_val_file.reshape(dims)

        # Create variable (name, datatype, dimensions)
        #pdb.set_trace()
        #dset_id = sd_id.createVariable(var_name, dtype_id, tuple([sd_id.dimensions[key] for key in current_dim]))

        if data.ndm[i,8] == 6:
            # String datatype
            str_out = netCDF4.stringtochar(np.array(var_val_file, dtype_id))
            dset_id = sd_id.createVariable(var_name, 'S1', tuple([key for key in current_dim]))
            dset_id[:] = str_out

        else:
            dset_id = sd_id.createVariable(var_name, dtype_id, tuple([key for key in current_dim]))
            dset_id[:] = var_val_file



        # Store 1-element arrays as scalars
        # if len(var_val_file) == 1:
        #     dset_id = sd_id.createDataset(var_name, dtype_id, var_val_file[0])
        # else:
        #     dset_id = sd_id.createDataset(var_name, dtype_id, var_val_file)

        # Write out variable attributes
        ninc=0
        j=0
        mi, _ = where_eq(md.meta_arr, md.attr_arr_data[0]+'=' + data.vn[i])

        #pdb.set_trace()

        for mn, mv in var_val.va_d.items():
            if mn ==  'VAR_VALID_MIN':
                athold = valid_range[0]
            elif mn == 'VAR_VALID_MAX':
                athold = valid_range[1]
            elif mn == 'VAR_FILL_VALUE':
                athold = fill_value
            else:
                athold = mv

            setattr(dset_id, mn, athold)

        #     # create attribute in the output file
        #     # use ASCII encoding for strings
        #     if type(athold).__name__ == 'str':
        #         athold = athold.encode('ascii')

        #     #dset_id.attrs.update({md.attr_arr_data[j] : athold})
        #     j=j+1
        #     ninc += 1

    sd_id.close()


def find_hdf_filename(meta_data, hdf_filename):
    """
    Compare filename created by the program from the DATASET_ATTRIBUTES with that under the
    FILE_NAME entry (if present). If necessary create/change the FILE_NAME entry to match the
    created filename

    """

    errtxt=['']*4
    proname='find_hdf_filename procedure: '
    errtxt[0]='Number of sub-values under DATA_'
    errtxt[1]=' should be '
    errtxt[2]=' ISO8601 format not valid: '
    errtxt[3]='Type conversion error.  MJD2K entry is not valid: '


    fn_val = -1
    fni, _ = where_eq(meta_data.attr_arr_glob, 'FILE_NAME')
    res = re.split(' |=', meta_data.meta_arr[fni[0]])
    fn_val = res[1]

    # check to see if the resulting filename is the same as that under FILE_NAME
#    if hdf_filename != fn_val:
    if hdf_filename.name != fn_val:
        if fn_val != -1:
            infotxt =['']*3
            infotxt[0]='2 FILE_NAME entry under File Attributes does not match the filename'
            infotxt[0]=infotxt[0]+' determined from the Global Attributes\n'
            infotxt[1]='    '+fn_val+' -> '+hdf_filename.name+'|'
            infotxt[2]='\n    Filename determined from Global Attributes used\n'
        else:
            if data.qa_yes:
                qtxt='should be '+hdf_filename.name
            else:
                qtxt=hdf_filename.name + ' added'
            infotxt=['']*2
            infotxt[0]='2 Missing FILE_NAME value '+qtxt + '\n'
            infotxt[1]='    based on Global Attribute values\n'
        infotxt_output(infotxt)
        meta_data.meta_arr[fni[0]]='FILE_NAME='+hdf_filename.name # change FILE_NAME entry

    return hdf_filename



def sds_to_meta(geoms_obj):
    """
    Using the sds structure from pycr8ascii, create metafile (metadata stored in list)


    For running QA.

    PRO idlcr8hdf performs additional checks, not implemented yet

    """

    # list of text for the meta data
    meta_file = []

    # Add the global attributes
    for key, value in geoms_obj.ga_d.items():
        meta_file.append(key + '=' + value)

    # Add the variable attribute labels/values
    # in the same order as DATA_VARIABLES
    for d_var in geoms_obj.ga_d['DATA_VARIABLES'].split(';'):
        d_val =  geoms_obj.sdss_d[d_var]
        for att_key, att_val in d_val.va_d.items():
            meta_file.append(att_key + '=' + str(att_val))


    return meta_file

def run_qa(tav_file, geoms_obj, out_dir, global_log_file=None):
    """

    Run pycr8hdf from qeoms_qa

    """

    global global_qa_yes

    # Get meta_file as a list of text
    tav_data = Tabledata()
    meta_data = Metadata()
    data = Data()
    data.qa_yes = True
    global_qa_yes = True
    data.inf = 0 # Indicate that data is a structure
    data.set_global_log_file_name(global_log_file)
    data.geoms = geoms_obj # Attach the geoms object to data
    meta_file = sds_to_meta(geoms_obj)

    # Read TAV file
    read_tablefile(tav_file, meta_data, tav_data)

    # meta_file is a list
    inf=0

    # Read metadata
    #metadata = read_metadata(meta_file, tav_data, meta_data, data, inf, qa_yes=True)
    metadata = read_metadata(meta_file, tav_data, meta_data, data, inf)

    # Check metadata
    check_metadata(meta_data, tav_data, data)

    # Call the procedure to set up the data structure and assign arrays for VAR_NAME and VAR_UNITS
    set_up_structure(meta_data, data)

    # Call the procedure to read the data
    read_data(meta_data, data)

    # Copy from geoms
    data.sdss_d = geoms_obj.sdss_d

    # Check min/max/fill
    check_min_max_fill(data)

    # Geoms rule changes
    #geoms_rule_changes_1(metadata, data)



if __name__ == "__main__":
    """
    Test, write a h5 file

    """

    # Implemented output formats
    output_formats = ('h5', 'nc4')

    # TODO: cmd line options
    parser = argparse.ArgumentParser(
        prog="pycr8hdf.py",
        description="Create a GEOMS file in hdf5 format from .meta and .data files",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('-m', '--meta_file',
                        dest='meta_file',
                        help="GEOMS meta data file",
                        default=None)

    parser.add_argument('-d', '--data_file',
                        dest='data_file',
                        help="GEOMS data file",
                        default=None)

    parser.add_argument('-t', '--tav_file',
                        dest='tav_file',
                        help="GEOMS TAV file",
                        default="./test_data/tableattrvalue_04R072.dat")

    parser.add_argument('-o', '--outdir',
                        dest='out_dir',
                        help="Directory to write output. Same as input directory if not given.",
                        default=None)

    parser.add_argument('-f', '--format',
                        dest='file_format',
                        help="File output format {}".format(output_formats),
                        default='h5')

    parser.add_argument('-l', '--log_file',
                        dest='log_file',
                        action='store_true',
                        help="If true, write output to log file pycr8hdf.log in output directory",
                        default=False)    


    args = parser.parse_args()
    meta_file = args.meta_file
    data_file = args.data_file
    tav_file =  args.tav_file
    out_dir = args.out_dir
    file_format = args.file_format
    log_file = args.log_file

    # Check input
    if len(sys.argv) == 1:
        parser.print_help()
        parser.exit()

    if file_format not in output_formats:
        print("Error: invalid output format. Valid formats are {}".format(output_formats))
        sys.exit()

    # meta_file is a file name
    inf=1

    if out_dir is None:
        # Set to input directory
        out_dir = os.path.dirname(os.path.abspath(data_file))
    else:
        out_dir = os.path.abspath(out_dir)


    tabledata = Tabledata()
    metadata = Metadata()
    data = Data()

    print(' File Input/Output Log - Program Started on {}\n'.format(datetime.datetime.utcnow().
                                                                  strftime("%a %b %d %H:%M:%S %Y")))

    # Set log file
    if args.log_file:
        log_file = os.path.join(out_dir, 'pycr8hdf.log')
        data.set_global_log_file_name(log_file)
        # Delete old log file on program start
        if os.path.isfile(log_file):
            os.remove(log_file)

        with open(log_file, 'a') as f:
            f.write(' File Input/Output Log - Program Started on {}\n\n'.format(datetime.datetime.utcnow().
                                                                  strftime("%a %b %d %H:%M:%S %Y")))
        
    data.qa_yes = False
    data.datafile = data_file
    data.inf = 1 # Indicate that data is read from text file
    data.out_dir = out_dir


    # Read TAV file
    read_tablefile(tav_file, metadata, tabledata)

    # Write files beoing read to log
    if inf == 1 and args.log_file:
        with open(log_file, 'a') as f:
            f.write('Input Metadata File: {}\n'.format(os.path.abspath(meta_file)))
            f.write('Input Data File: {}\n'.format(os.path.abspath(data_file)))


    # Read metadata
    metadata = read_metadata(meta_file, tabledata, metadata, data, inf)

    # Check metadata
    check_metadata(metadata, tabledata, data)

    # Call the procedure to set up the data structure and assign arrays for VAR_NAME and VAR_UNITS
    set_up_structure(metadata, data)

    # Call the procedure to read the data
    read_data(metadata, data)

    # Place data in a structure
    data.meta_data_to_geoms(metadata)

    # Check min/max/fill values
    check_min_max_fill(data)

    # Write output file
    if file_format == 'h5':
        hdf_filename = Path(os.path.basename(data_file)).with_suffix('.h5')
        hdf_filename = find_hdf_filename(metadata, hdf_filename)
    elif file_format == 'nc4':
        hdf_filename = Path(os.path.basename(data_file)).with_suffix('.nc')
        hdf_filename = find_hdf_filename(metadata, hdf_filename)        

    # No errors detected:
    if global_o34_idl == '0':
        # Write output file
        if file_format == 'h5':
            avdc_hdf5_write(hdf_filename, metadata, data)
        elif file_format == 'nc4':
            avdc_nc4_write(hdf_filename, metadata, data)

        if args.log_file:
            with open(log_file, 'a') as f:
                f.write('{} successfully created!\n\n'.format(os.path.join(out_dir, hdf_filename.name)))
                f.write('File creation completed - Program Ended on {}\n'.format(
                    datetime.datetime.utcnow().strftime("%a %b %d %H:%M:%S %Y")))
        else: # If no log file, write to standard output
            print('{} successfully created!\n\n'.format(os.path.join(out_dir, hdf_filename.name)))
            print('File creation completed - Program Ended on {}\n'.format(
                datetime.datetime.utcnow().strftime("%a %b %d %H:%M:%S %Y")))

    elif global_o34_idl == 'NOHDF':
        if args.log_file:
            with open(log_file, 'a') as f:
                f.write('GEOMS file testing completed - Program Ended on {}\n'.format(
                        datetime.datetime.utcnow().strftime("%a %b %d %H:%M:%S %Y")))
        else:
            # If no log file, write to standard output                
            print('GEOMS file testing completed - Program Ended on {}\n'.format(
                datetime.datetime.utcnow().strftime("%a %b %d %H:%M:%S %Y")))
